Module build_excl
  ! SETUP MODULES

  Use angles,          Only: angles_type
  Use bonds,           Only: bonds_type
  Use comms,           Only: comms_type,&
                             gcheck,&
                             gmax
  ! CONFIG MODULE
  Use configuration,   Only: configuration_type
  ! INTERACTION MODULES
  Use constraints,     Only: constraints_type
  Use core_shell,      Only: core_shell_type
  Use dihedrals,       Only: dihedrals_type
  Use errors_warnings, Only: error,&
                             warning
  Use inversions,      Only: inversions_type
  Use kinds,           Only: wi,&
                             wp
  Use neighbours,      Only: neighbours_type
  Use numerics,        Only: local_index,&
                             shellsort
  Use rigid_bodies,    Only: rigid_bodies_type

  Implicit None

  Private
  Public :: build_excl_intra, add_exclusion

Contains

  Subroutine build_excl_intra(lecx, cshell, cons, bond, angle, dihedral, inversion, neigh, rigid, config, comm)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !
    ! dl_poly_4 subroutine for constructing the excluded pair interaction
    ! of the part of the MD system mapped onto this node
    !
    ! keyinteraction < 0 exclusion does not apply
    !
    ! copyright - daresbury laboratory
    ! author    - w.smith february 1999
    ! amended   - i.t.todorov july 2016
    ! refactoring:
    !           - a.m.elena march-october 2018
    !           - j.madge march-october 2018
    !           - a.b.g.chalk march-october 2018
    !           - i.scivetti march-october 2018
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    Logical,                  Intent(In   ) :: lecx
    Type(core_shell_type),    Intent(InOut) :: cshell
    Type(constraints_type),   Intent(In   ) :: cons
    Type(bonds_type),         Intent(In   ) :: bond
    Type(angles_type),        Intent(In   ) :: angle
    Type(dihedrals_type),     Intent(In   ) :: dihedral
    Type(inversions_type),    Intent(In   ) :: inversion
    Type(neighbours_type),    Intent(InOut) :: neigh
    Type(rigid_bodies_type),  Intent(In   ) :: rigid
    Type(configuration_type), Intent(InOut) :: config
    Type(comms_type),         Intent(InOut) :: comm

    Character(Len=256)                    :: message
    Integer                               :: fail(1:2), i, ia, ia0, ib, ib0, ibig, ic, ic0, id, &
                                             id0, j, ja, ja0, jb, jb0, jc, jc0, jd, jd0, k, ka, &
                                             ka0, kb, kb0, kc, kc0, kd, kd0, kk, l
    Integer, Allocatable, Dimension(:, :) :: irgd, irgd0, jrgd, jrgd0
    Logical                               :: safe

    If (rigid%max_rigid > 0) Then
      fail = 0
      Allocate (irgd(1:rigid%max_list, 1:rigid%max_rigid), &
                irgd0(1:rigid%max_list, 1:rigid%max_rigid), Stat=fail(1))
      Allocate (jrgd(1:rigid%max_list, 1:rigid%max_rigid), &
                jrgd0(1:rigid%max_list, 1:rigid%max_rigid), Stat=fail(2))
      If (Any(fail > 0)) Then
        Write (message, '(a)') 'build_excl_intra allocation failure'
        Call error(0, message)
      End If
    End If

    ! variables for array bound checking

    ibig = 0
    safe = .true.

    ! exclude sites on basis of bond constraints

    Do i = 1, cons%ntcons
      ia = cons%listcon(1, i)
      ib = cons%listcon(2, i)

      ia0 = local_index(ia, config%nlast, config%lsi, config%lsa)
      ib0 = local_index(ib, config%nlast, config%lsi, config%lsa)

      If (ia0 > config%natms) ia0 = 0
      If (ib0 > config%natms) ib0 = 0

      ! add atoms to exclusion list

      If (ia0 > 0) Call add_exclusion(safe, ia0, ib, ibig, neigh%list_excl)
      If (ib0 > 0) Call add_exclusion(safe, ib0, ia, ibig, neigh%list_excl)
    End Do

    ! exclude sites on basis of RBs

    Do i = 1, rigid%n_types
      Do j = 1, rigid%list(-1, i)
        irgd(j, i) = rigid%list(j, i)
        irgd0(j, i) = local_index(rigid%list(j, i), config%nlast, config%lsi, config%lsa)
        If (irgd0(j, i) > config%natms) irgd0(j, i) = 0
      End Do

      ! add atoms to exclusion list

      Do j = 1, rigid%list(-1, i)
        If (irgd0(j, i) > 0) Then
          Do k = 1, rigid%list(-1, i)
            If (k /= j) Call add_exclusion(safe, irgd0(j, i), irgd(k, i), ibig, neigh%list_excl)
          End Do
        End If
      End Do
    End Do

    ! exclude sites on basis of chemical bonds

    Do i = 1, bond%n_types
      If (.not. bond%restrained(bond%list(0, i))) Then
        ia = bond%list(1, i)
        ib = bond%list(2, i)

        ia0 = local_index(ia, config%nlast, config%lsi, config%lsa)
        ib0 = local_index(ib, config%nlast, config%lsi, config%lsa)

        If (ia0 > config%natms) ia0 = 0
        If (ib0 > config%natms) ib0 = 0

        ! add atoms to exclusion list

        If (ia0 > 0) Call add_exclusion(safe, ia0, ib, ibig, neigh%list_excl)
        If (ib0 > 0) Call add_exclusion(safe, ib0, ia, ibig, neigh%list_excl)
      End If
    End Do

    ! exclude sites on basis of bond angles

    Do i = 1, angle%n_types
      If (.not. angle%restrained(angle%list(0, i))) Then
        ia = angle%list(1, i)
        ib = angle%list(2, i)
        ic = angle%list(3, i)

        ia0 = local_index(ia, config%nlast, config%lsi, config%lsa)
        ib0 = local_index(ib, config%nlast, config%lsi, config%lsa)
        ic0 = local_index(ic, config%nlast, config%lsi, config%lsa)

        If (ia0 > config%natms) ia0 = 0
        If (ib0 > config%natms) ib0 = 0
        If (ic0 > config%natms) ic0 = 0

        ! add atoms to exclusion list

        If (ia0 > 0) Then ! ia : ib - ic interactions
          Call add_exclusion(safe, ia0, ib, ibig, neigh%list_excl)
          Call add_exclusion(safe, ia0, ic, ibig, neigh%list_excl)
        End If

        If (ib0 > 0) Then ! ib : ia - ic interactions
          Call add_exclusion(safe, ib0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, ib0, ic, ibig, neigh%list_excl)
        End If

        If (ic0 > 0) Then ! ic : ia - ib interactions
          Call add_exclusion(safe, ic0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, ic0, ib, ibig, neigh%list_excl)
        End If
      End If
    End Do

    ! exclude sites on basis of dihedral angles

    Do i = 1, dihedral%n_types
      If (.not. dihedral%restrained(dihedral%list(0, i))) Then
        ia = dihedral%list(1, i)
        ib = dihedral%list(2, i)
        ic = dihedral%list(3, i)
        id = dihedral%list(4, i)

        ia0 = local_index(ia, config%nlast, config%lsi, config%lsa)
        ib0 = local_index(ib, config%nlast, config%lsi, config%lsa)
        ic0 = local_index(ic, config%nlast, config%lsi, config%lsa)
        id0 = local_index(id, config%nlast, config%lsi, config%lsa)

        If (ia0 > config%natms) ia0 = 0
        If (ib0 > config%natms) ib0 = 0
        If (ic0 > config%natms) ic0 = 0
        If (id0 > config%natms) id0 = 0

        ! add atoms to exclusion list

        If (ia0 > 0) Then ! ia : ib - ic - id interactions
          Call add_exclusion(safe, ia0, ib, ibig, neigh%list_excl)
          Call add_exclusion(safe, ia0, ic, ibig, neigh%list_excl)
          Call add_exclusion(safe, ia0, id, ibig, neigh%list_excl)
        End If

        If (ib0 > 0) Then ! ib : ia - ic - id interactions
          Call add_exclusion(safe, ib0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, ib0, ic, ibig, neigh%list_excl)
          Call add_exclusion(safe, ib0, id, ibig, neigh%list_excl)
        End If

        If (ic0 > 0) Then ! ic : ia - ib - id interactions
          Call add_exclusion(safe, ic0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, ic0, ib, ibig, neigh%list_excl)
          Call add_exclusion(safe, ic0, id, ibig, neigh%list_excl)
        End If

        If (id0 > 0) Then ! id : ia - ib - ic interactions
          Call add_exclusion(safe, id0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, id0, ib, ibig, neigh%list_excl)
          Call add_exclusion(safe, id0, ic, ibig, neigh%list_excl)
        End If
      End If
    End Do

    ! exclude sites on basis of inversion potentials

    Do i = 1, inversion%n_types
      If (.not. inversion%restrained(inversion%list(0, i))) Then
        ia = inversion%list(1, i)
        ib = inversion%list(2, i)
        ic = inversion%list(3, i)
        id = inversion%list(4, i)

        ia0 = local_index(ia, config%nlast, config%lsi, config%lsa)
        ib0 = local_index(ib, config%nlast, config%lsi, config%lsa)
        ic0 = local_index(ic, config%nlast, config%lsi, config%lsa)
        id0 = local_index(id, config%nlast, config%lsi, config%lsa)

        If (ia0 > config%natms) ia0 = 0
        If (ib0 > config%natms) ib0 = 0
        If (ic0 > config%natms) ic0 = 0
        If (id0 > config%natms) id0 = 0

        ! add atoms to exclusion list

        If (ia0 > 0) Then ! ia : ib - ic - id interactions
          Call add_exclusion(safe, ia0, ib, ibig, neigh%list_excl)
          Call add_exclusion(safe, ia0, ic, ibig, neigh%list_excl)
          Call add_exclusion(safe, ia0, id, ibig, neigh%list_excl)
        End If

        If (ib0 > 0) Then ! ib : ia - ic - id interactions
          Call add_exclusion(safe, ib0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, ib0, ic, ibig, neigh%list_excl)
          Call add_exclusion(safe, ib0, id, ibig, neigh%list_excl)
        End If

        If (ic0 > 0) Then ! ic : ia - ib - id interactions
          Call add_exclusion(safe, ic0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, ic0, ib, ibig, neigh%list_excl)
          Call add_exclusion(safe, ic0, id, ibig, neigh%list_excl)
        End If

        If (id0 > 0) Then ! id : ia - ib - ic interactions
          Call add_exclusion(safe, id0, ia, ibig, neigh%list_excl)
          Call add_exclusion(safe, id0, ib, ibig, neigh%list_excl)
          Call add_exclusion(safe, id0, ic, ibig, neigh%list_excl)
        End If
      End If
    End Do

    ! exclude sites on basis of core-shell units

    Do i = 1, cshell%ntshl2
      ia = cshell%listshl(1, i) ! This is the core
      ib = cshell%listshl(2, i) ! This is the shell

      ia0 = local_index(ia, config%nlast, config%lsi, config%lsa)
      ib0 = local_index(ib, config%nlast, config%lsi, config%lsa)

      If (ia0 > config%natms) ia0 = 0
      If (ib0 > config%natms) ib0 = 0

      ! add atoms to exclusion list

      If (ia0 > 0) Call add_exclusion(safe, ia0, ib, ibig, neigh%list_excl)
      If (ib0 > 0) Call add_exclusion(safe, ib0, ia, ibig, neigh%list_excl)

      ! exclude sites on basis of constraint bonds to core-shell units

      Do kk = 1, cons%ntcons1
        ja = cons%listcon(1, kk)
        jb = cons%listcon(2, kk)

        ja0 = local_index(ja, config%nlast, config%lsi, config%lsa)
        jb0 = local_index(jb, config%nlast, config%lsi, config%lsa)

        If (ja0 > config%natms) ja0 = 0
        If (jb0 > config%natms) jb0 = 0

        ! Check for adjacent core or shell so that all possible, even between
        ! adjacent particles in any intra-unit, interactions are excluded.
        ! The checks commented out are not needed due to preventative checks
        ! in read_field by which shells are not allowed to be frozen,
        ! constrained bonded, RBed or tethered.

        ka = 0; ka0 = 0
        kb = 0; kb0 = 0

        If (lecx) Then
          Do j = 1, cshell%ntshl2
            If (cshell%listshl(1, j) == ja) Then
              ka = cshell%listshl(2, j)
              ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
              If (ka0 > config%natms) ka0 = 0
              !              Else If (cshell%listshl(2,j) == ja) Then
              !                 ka=cshell%listshl(1,j)
              !                 ka0=local_index(ka,nlast,lsi,lsa)
              !                 If (ka0 > natms) ka0=0
            End If

            If (cshell%listshl(1, j) == jb) Then
              kb = cshell%listshl(2, j)
              kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
              If (kb0 > config%natms) kb0 = 0
              !              Else If (cshell%listshl(2,j) == jb) Then
              !                 kb=cshell%listshl(1,j)
              !                 kb0=local_index(kb,nlast,lsi,lsa)
              !                 If (kb0 > natms) kb0=0
            End If
          End Do
        End If

        If (ia == ja) Then
          If (ib0 > 0) Then
            Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
            If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)
          End If
          If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
          If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)
        End If

        If (ia == jb) Then
          If (ib0 > 0) Then
            Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
            If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)
          End If
          If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
          If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)
        End If

        !        If (ib == ja) Then
        !           If (ia0 > 0) Then
        !              Call add_exclusion(safe,ia0,jb,ibig,neigh%list_excl)
        !              If (kb > 0) Call add_exclusion(safe,ia0,kb,ibig,neigh%list_excl)
        !           End If
        !           If (jb0 > 0) Call add_exclusion(safe,jb0,ia,ibig,neigh%list_excl)
        !           If (kb0 > 0) Call add_exclusion(safe,kb0,ia,ibig,neigh%list_excl)
        !        End If
        !
        !        If (ib == jb) Then
        !           If (ia0 > 0) Then
        !              Call add_exclusion(safe,ia0,ja,ibig,neigh%list_excl)
        !              If (ka > 0) Call add_exclusion(safe,ia0,ka,ibig,neigh%list_excl)
        !           End If
        !           If (ja0 > 0) Call add_exclusion(safe,ja0,ia,ibig,neigh%list_excl)
        !           If (ka0 > 0) Call add_exclusion(safe,ka0,ia,ibig,neigh%list_excl)
        !        End If
      End Do

      ! exclude sites on basis of RBs to core-shell units

      Do j = 1, rigid%n_types_book
        Do k = 1, rigid%list(-1, j)
          irgd(k, j) = rigid%list(k, j)
          irgd0(k, j) = local_index(rigid%list(k, j), config%nlast, config%lsi, config%lsa)
          If (irgd0(k, j) > config%natms) irgd0(k, j) = 0
        End Do

        ! Check for adjacent core or shell so that all possible, even between
        ! adjacent particles in any intra-unit, interactions are excluded
        ! The checks commented out are not needed due to preventative checks
        ! in read_field by which shells are not allowed to be frozen,
        ! constrained bonded, RBed or tethered.

        jrgd = 0; jrgd0 = 0

        If (lecx) Then
          Do kk = 1, cshell%ntshl2
            Do k = 1, rigid%list(-1, j)
              If (irgd(k, j) == cshell%listshl(1, kk)) Then
                jrgd(k, j) = cshell%listshl(2, kk)
                jrgd0(k, j) = local_index(jrgd(k, j), config%nlast, config%lsi, config%lsa)
                If (jrgd0(k, j) > config%natms) jrgd0(k, j) = 0
                !                 Else If (irgd(k,j) == cshell%listshl(2,kk)) Then
                !                    jrgd(k,j)=cshell%listshl(2,kk)
                !                    jrgd0(k,j)=local_index(jrgd(k,j),nlast,lsi,lsa)
                !                    If (jrgd0(k,j) > natms) jrgd0(k,j)=0
              End If
            End Do
          End Do
        End If

        Do k = 1, rigid%list(-1, j)
          If (irgd(k, j) == ia) Then
            Do l = 1, rigid%list(-1, j)
              If (l /= k) Then
                If (ib0 > 0) Then
                  Call add_exclusion(safe, ib0, irgd(l, j), ibig, neigh%list_excl)
                  If (jrgd(l, j) > 0) Call add_exclusion(safe, ib0, jrgd(l, j), ibig, neigh%list_excl)
                End If
                If (irgd0(l, j) > 0) Call add_exclusion(safe, irgd0(l, j), ib, ibig, neigh%list_excl)
                If (jrgd0(l, j) > 0) Call add_exclusion(safe, jrgd0(l, j), ib, ibig, neigh%list_excl)
              End If
            End Do
          End If

          !           If (irgd(k,j) == ib) Then
          !              Do l=1,rigid%list(-1,j)
          !                 If (l /= k) Then
          !                    If (ia0 > 0) Then
          !                       Call add_exclusion(safe,ia0,irgd(l,j),ibig,neigh%list_excl)
          !                       If (jrgd(l,j) > 0) Call add_exclusion(safe,ia0,jrgd(l,j),ibig,neigh%list_excl)
          !                    End If
          !                    If (irgd0(l,j) > 0) Call add_exclusion(safe,irgd0(l,j),ia,ibig,neigh%list_excl)
          !                    If (jrgd0(l,j) > 0) Call add_exclusion(safe,jrgd0(l,j),ia,ibig,neigh%list_excl)
          !                 End If
          !              End Do
          !           End If
        End Do
      End Do

      ! exclude sites on basis of bonds to core-shell units

      Do kk = 1, bond%n_types1
        If (.not. bond%restrained(bond%list(0, kk))) Then
          ja = bond%list(1, kk)
          jb = bond%list(2, kk)

          ja0 = local_index(ja, config%nlast, config%lsi, config%lsa)
          jb0 = local_index(jb, config%nlast, config%lsi, config%lsa)

          If (ja0 > config%natms) ja0 = 0
          If (jb0 > config%natms) jb0 = 0

          ! Check for adjacent core or shell so that all possible, even between
          ! adjacent particles in any intra-unit, interactions are excluded

          ka = 0; ka0 = 0
          kb = 0; kb0 = 0

          If (lecx) Then
            Do j = 1, cshell%ntshl2
              If (cshell%listshl(1, j) == ja) Then
                ka = cshell%listshl(2, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              Else If (cshell%listshl(2, j) == ja) Then
                ka = cshell%listshl(1, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              End If

              If (cshell%listshl(1, j) == jb) Then
                kb = cshell%listshl(2, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              Else If (cshell%listshl(2, j) == jb) Then
                kb = cshell%listshl(1, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              End If
            End Do
          End If

          If (ia == ja) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)
            End If
            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jb) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)
            End If
            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)
          End If

          If (ib == ja) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)
            End If
            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jb) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)
            End If
            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)
          End If
        End If
      End Do

      ! exclude sites on basis of valence angles to core-shell units

      Do kk = 1, angle%n_types1
        If (.not. angle%restrained(angle%list(0, kk))) Then
          ja = angle%list(1, kk)
          jb = angle%list(2, kk)
          jc = angle%list(3, kk)

          ja0 = local_index(ja, config%nlast, config%lsi, config%lsa)
          jb0 = local_index(jb, config%nlast, config%lsi, config%lsa)
          jc0 = local_index(jc, config%nlast, config%lsi, config%lsa)

          If (ja0 > config%natms) ja0 = 0
          If (jb0 > config%natms) jb0 = 0
          If (jc0 > config%natms) jc0 = 0

          ! Check for adjacent core or shell so that all possible, even between
          ! adjacent particles in any intra-unit, interactions are excluded

          ka = 0; ka0 = 0
          kb = 0; kb0 = 0
          kc = 0; kc0 = 0

          If (lecx) Then
            Do j = 1, cshell%ntshl2
              If (cshell%listshl(1, j) == ja) Then
                ka = cshell%listshl(2, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              Else If (cshell%listshl(2, j) == ja) Then
                ka = cshell%listshl(1, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              End If

              If (cshell%listshl(1, j) == jb) Then
                kb = cshell%listshl(2, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              Else If (cshell%listshl(2, j) == jb) Then
                kb = cshell%listshl(1, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              End If

              If (cshell%listshl(1, j) == jc) Then
                kc = cshell%listshl(2, j)
                kc0 = local_index(kc, config%nlast, config%lsi, config%lsa)
                If (kc0 > config%natms) kc0 = 0
              Else If (cshell%listshl(2, j) == jc) Then
                kc = cshell%listshl(1, j)
                kc0 = local_index(kc, config%nlast, config%lsi, config%lsa)
                If (kc0 > config%natms) kc0 = 0
              End If
            End Do
          End If

          If (ia == ja) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)
            End If

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jb) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jc) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)
          End If

          If (ib == ja) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)
            End If

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jb) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jc) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)
          End If
        End If
      End Do

      ! exclude sites on basis of dihedral angles to core-shell units

      Do kk = 1, dihedral%n_types1
        If (.not. dihedral%restrained(dihedral%list(0, kk))) Then
          ja = dihedral%list(1, kk)
          jb = dihedral%list(2, kk)
          jc = dihedral%list(3, kk)
          jd = dihedral%list(4, kk)

          ja0 = local_index(ja, config%nlast, config%lsi, config%lsa)
          jb0 = local_index(jb, config%nlast, config%lsi, config%lsa)
          jc0 = local_index(jc, config%nlast, config%lsi, config%lsa)
          jd0 = local_index(jd, config%nlast, config%lsi, config%lsa)

          If (ja0 > config%natms) ja0 = 0
          If (jb0 > config%natms) jb0 = 0
          If (jc0 > config%natms) jc0 = 0
          If (jd0 > config%natms) jd0 = 0

          ! Check for adjacent core or shell so that all possible, even between
          ! adjacent particles in any intra-unit, interactions are excluded

          ka = 0; ka0 = 0
          kb = 0; kb0 = 0
          kc = 0; kc0 = 0
          kd = 0; kd0 = 0

          If (lecx) Then
            Do j = 1, cshell%ntshl2
              If (cshell%listshl(1, j) == ja) Then
                ka = cshell%listshl(2, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              Else If (cshell%listshl(2, j) == ja) Then
                ka = cshell%listshl(1, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              End If

              If (cshell%listshl(1, j) == jb) Then
                kb = cshell%listshl(2, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              Else If (cshell%listshl(2, j) == jb) Then
                kb = cshell%listshl(1, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              End If

              If (cshell%listshl(1, j) == jc) Then
                kc = cshell%listshl(2, j)
                kc0 = local_index(kc, config%nlast, config%lsi, config%lsa)
                If (kc0 > config%natms) kc0 = 0
              Else If (cshell%listshl(2, j) == jc) Then
                kc = cshell%listshl(1, j)
                kc0 = local_index(kc, config%nlast, config%lsi, config%lsa)
                If (kc0 > config%natms) kc0 = 0
              End If

              If (cshell%listshl(1, j) == jd) Then
                kd = cshell%listshl(2, j)
                kd0 = local_index(kd, config%nlast, config%lsi, config%lsa)
                If (kd0 > config%natms) kd0 = 0
              Else If (cshell%listshl(2, j) == jd) Then
                kd = cshell%listshl(1, j)
                kd0 = local_index(kd, config%nlast, config%lsi, config%lsa)
                If (kd0 > config%natms) kd0 = 0
              End If
            End Do
          End If

          If (ia == ja) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ib0, kd, ibig, neigh%list_excl)
            End If

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ib, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jb) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ib0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ib, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jc) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ib0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ib, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jd) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)
          End If

          If (ib == ja) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ia0, kd, ibig, neigh%list_excl)
            End If

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ia, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jb) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ia0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ia, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jc) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ia0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ia, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jd) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)
          End If
        End If
      End Do

      ! exclude sites on basis of inversion angles to core-shell units

      Do kk = 1, inversion%n_types1
        If (.not. inversion%restrained(inversion%list(0, kk))) Then
          ja = inversion%list(1, kk)
          jb = inversion%list(2, kk)
          jc = inversion%list(3, kk)
          jd = inversion%list(4, kk)

          ja0 = local_index(ja, config%nlast, config%lsi, config%lsa)
          jb0 = local_index(jb, config%nlast, config%lsi, config%lsa)
          jc0 = local_index(jc, config%nlast, config%lsi, config%lsa)
          jd0 = local_index(jd, config%nlast, config%lsi, config%lsa)

          If (ja0 > config%natms) ja0 = 0
          If (jb0 > config%natms) jb0 = 0
          If (jc0 > config%natms) jc0 = 0
          If (jd0 > config%natms) jd0 = 0

          ! Check for adjacent core or shell so that all possible, even between
          ! adjacent particles in any intra-unit, interactions are excluded

          ka = 0; ka0 = 0
          kb = 0; kb0 = 0
          kc = 0; kc0 = 0
          kd = 0; kd0 = 0

          If (lecx) Then
            Do j = 1, cshell%ntshl2
              If (cshell%listshl(1, j) == ja) Then
                ka = cshell%listshl(2, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              Else If (cshell%listshl(2, j) == ja) Then
                ka = cshell%listshl(1, j)
                ka0 = local_index(ka, config%nlast, config%lsi, config%lsa)
                If (ka0 > config%natms) ka0 = 0
              End If

              If (cshell%listshl(1, j) == jb) Then
                kb = cshell%listshl(2, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              Else If (cshell%listshl(2, j) == jb) Then
                kb = cshell%listshl(1, j)
                kb0 = local_index(kb, config%nlast, config%lsi, config%lsa)
                If (kb0 > config%natms) kb0 = 0
              End If

              If (cshell%listshl(1, j) == jc) Then
                kc = cshell%listshl(2, j)
                kc0 = local_index(kc, config%nlast, config%lsi, config%lsa)
                If (kc0 > config%natms) kc0 = 0
              Else If (cshell%listshl(2, j) == jc) Then
                kc = cshell%listshl(1, j)
                kc0 = local_index(kc, config%nlast, config%lsi, config%lsa)
                If (kc0 > config%natms) kc0 = 0
              End If

              If (cshell%listshl(1, j) == jd) Then
                kd = cshell%listshl(2, j)
                kd0 = local_index(kd, config%nlast, config%lsi, config%lsa)
                If (kd0 > config%natms) kd0 = 0
              Else If (cshell%listshl(2, j) == jd) Then
                kd = cshell%listshl(1, j)
                kd0 = local_index(kd, config%nlast, config%lsi, config%lsa)
                If (kd0 > config%natms) kd0 = 0
              End If
            End Do
          End If

          If (ia == ja) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ib0, kd, ibig, neigh%list_excl)
            End If

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ib, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jb) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ib0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ib, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jc) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ib0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ib, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ib, ibig, neigh%list_excl)
          End If

          If (ia == jd) Then
            If (ib0 > 0) Then
              Call add_exclusion(safe, ib0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ib0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ib0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ib0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ib0, kc, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ib, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ib, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ib, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ib, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ib, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ib, ibig, neigh%list_excl)
          End If

          If (ib == ja) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ia0, kd, ibig, neigh%list_excl)
            End If

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ia, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jb) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ia0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ia, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jc) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jd, ibig, neigh%list_excl)
              If (kd > 0) Call add_exclusion(safe, ia0, kd, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)

            If (jd0 > 0) Call add_exclusion(safe, jd0, ia, ibig, neigh%list_excl)
            If (kd0 > 0) Call add_exclusion(safe, kd0, ia, ibig, neigh%list_excl)
          End If

          If (ib == jd) Then
            If (ia0 > 0) Then
              Call add_exclusion(safe, ia0, ja, ibig, neigh%list_excl)
              If (ka > 0) Call add_exclusion(safe, ia0, ka, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jb, ibig, neigh%list_excl)
              If (kb > 0) Call add_exclusion(safe, ia0, kb, ibig, neigh%list_excl)

              Call add_exclusion(safe, ia0, jc, ibig, neigh%list_excl)
              If (kc > 0) Call add_exclusion(safe, ia0, kc, ibig, neigh%list_excl)
            End If

            If (ja0 > 0) Call add_exclusion(safe, ja0, ia, ibig, neigh%list_excl)
            If (ka0 > 0) Call add_exclusion(safe, ka0, ia, ibig, neigh%list_excl)

            If (jb0 > 0) Call add_exclusion(safe, jb0, ia, ibig, neigh%list_excl)
            If (kb0 > 0) Call add_exclusion(safe, kb0, ia, ibig, neigh%list_excl)

            If (jc0 > 0) Call add_exclusion(safe, jc0, ia, ibig, neigh%list_excl)
            If (kc0 > 0) Call add_exclusion(safe, kc0, ia, ibig, neigh%list_excl)
          End If
        End If
      End Do
    End Do

    ! check for exceeded array bounds

    Call gcheck(comm, safe)
    If (.not. safe) Then
      Call gmax(comm, ibig)
      Call warning(250, Real(ibig, wp), Real(neigh%max_exclude, wp), 0.0_wp)
      Call error(65)
    End If

    ! sort neigh%list_excl

    Do i = 1, config%natms
      j = neigh%list_excl(0, i)
      If (j > 0) Call shellsort(j, neigh%list_excl(1:j, i))
    End Do

    If (rigid%max_rigid > 0) Then
      Deallocate (irgd, irgd0, Stat=fail(1))
      Deallocate (jrgd, jrgd0, Stat=fail(2))
      If (Any(fail > 0)) Then
        Write (message, '(a)') 'build_excl_intra deallocation failure'
        Call error(0, message)
      End If
    End If
  End Subroutine build_excl_intra

  Subroutine add_exclusion(safe, ia0, ib, ibig, list)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !
    ! dl_poly_4 subroutine to add excluded atoms to the excluded atom list
    ! provided they are not already excluded
    !
    ! copyright - daresbury laboratory
    ! author    - w.smith march 1999
    ! amended   - i.t.todorov july 2016
    ! refactoring:
    !           - a.m.elena march-october 2018
    !           - j.madge march-october 2018
    !           - a.b.g.chalk march-october 2018
    !           - i.scivetti march-october 2018
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    Logical,                    Intent(InOut) :: safe
    Integer,                    Intent(In   ) :: ia0, ib
    Integer,                    Intent(InOut) :: ibig
    Integer, Dimension(0:, 1:), Intent(InOut) :: list

    Integer          :: last
    Integer(Kind=wi) :: max_exclude
    Logical          :: l_excluded, safe_local

    ! Get upper bound of list
    max_exclude = Ubound(list, 1)

    ! Get current length
    last = list(0, ia0)

    ! Determine possible exclusion
    l_excluded = Any(list(1:last, ia0) == ib)
    If (.not. l_excluded) Then

      ! Get local safety no array overflow
      safe_local = (last < max_exclude - 1)

      ! Determine global safety
      safe = safe .and. safe_local

      If (safe_local) Then
        ! Increase length of the ia0 exclusion list and tag ib in it
        last = last + 1
        list(0, ia0) = last
        list(last, ia0) = ib

        ibig = Max(ibig, last)
      Else
        ! Collect number of offences
        list(max_exclude, ia0) = list(max_exclude, ia0) + 1
        ibig = Max(ibig, last + list(max_exclude, ia0))
      End If

    End If
  End Subroutine add_exclusion
End Module build_excl
