Module coul_mpole

  !!------------------------------------------------------------------------!
  !!
  !! dl_poly_4 module declaring configuration variables and arrays for
  !! multipoles
  !!
  !! copyright - daresbury laboratory
  !! author    - h.a.boateng & i.t.todorov january 2017
  !! contrib   - a.v.brukhno & m.a.seaton august 2020
  !! refactoring:
  !!           - a.m.elena march-october 2018
  !!           - j.madge march-october 2018
  !!           - a.b.g.chalk march-october 2018
  !!           - i.scivetti march-october 2018
  !!
  !!------------------------------------------------------------------------!

  Use configuration,    Only: configuration_type
  Use constants,        Only: r4pie0,&
                              sqrpi,&
                              zero_plus
  Use electrostatic,    Only: ELECTROSTATIC_COULOMB,&
                              ELECTROSTATIC_COULOMB_FORCE_SHIFT,&
                              ELECTROSTATIC_COULOMB_REACTION_FIELD,&
                              ELECTROSTATIC_DDDP,&
                              ELECTROSTATIC_EWALD,&
                              electrostatic_type
  Use kinds,            Only: wp
  Use mpole,            Only: mpole_type
  Use mpoles_container, Only: coul_deriv,&
                              ewald_deriv,&
                              explicit_ewald_real_loops,&
                              explicit_fscp_rfp_loops
  Use neighbours,       Only: neighbours_type
  Use numerics,         Only: images_s

  Implicit None

  Private

  Public :: intra_mcoul, coul_fscp_mforces, coul_rfp_mforces, coul_cp_mforces, coul_dddp_mforces, &
            coul_chrm_forces, d_ene_trq_mpoles

Contains

  Subroutine intra_mcoul(rcut, iatm, jatm, scale, &
                         rrr, xdf, ydf, zdf, coul, virele, fx, fy, fz, safe, mpoles, electro, config)

    !!------------------------------------------------------------------------!
    !!
    !! dl_poly_4 subroutine for calculating bond's or 1-4 dihedral
    !! electrostatics: adjusted by a config%weighting factor
    !!
    !! copyright - daresbury laboratory
    !! amended   - i.t.todorov & h.a.boateng february 2016
    !! refactoring:
    !!           - a.m.elena march-october 2018
    !!           - j.madge march-october 2018
    !!           - a.b.g.chalk march-october 2018
    !!           - i.scivetti march-october 2018
    !!
    !!------------------------------------------------------------------------!

    Real(Kind=wp),            Intent(In   ) :: rcut
    Integer,                  Intent(In   ) :: iatm, jatm
    Real(Kind=wp),            Intent(In   ) :: scale, rrr, xdf, ydf, zdf
    Real(Kind=wp),            Intent(  Out) :: coul, virele, fx, fy, fz
    Logical,                  Intent(InOut) :: safe
    Type(mpole_type),         Intent(InOut) :: mpoles
    Type(electrostatic_type), Intent(InOut) :: electro
    Type(configuration_type), Intent(In   ) :: config

    Real(Kind=wp), Parameter :: aa1 = 0.254829592_wp, aa2 = -0.284496736_wp, aa3 = 1.421413741_wp, &
                                aa4 = -1.453152027_wp, aa5 = 1.061405429_wp, pp = 0.3275911_wp

    Integer       :: ii, jj, k1, k2, k3, ks1, ks11, ks2, ks21, ks3, ks31, n, s1, s2, s3
    Logical, Save :: newjob = .true.
    Real(Kind=wp) :: alphan, b0, erc, exp1, fer, kx, ky, kz, sx, sy, sz, t1, t2, talpha, tix, tiy, &
                     tiz, tjx, tjy, tjz, tmp, tmpi, tmpj, tt, txyz, jmpz(1:mpoles%max_mpoles), &
                     jmpy(1:mpoles%max_mpoles), jmpx(1:mpoles%max_mpoles), &
                     jmp(1:mpoles%max_mpoles), impz(1:mpoles%max_mpoles), &
                     impy(1:mpoles%max_mpoles), impx(1:mpoles%max_mpoles), &
                     imp(1:mpoles%max_mpoles), d1(-2:2 * mpoles%max_order + 1, -2:2 * mpoles% &
                     max_order + 1, -2:2 * mpoles%max_order + 1), b1(-2:2 * mpoles%max_order + 1, &
                     -2:2 * mpoles%max_order + 1, -2:2 * mpoles%max_order + 1), a1(-2:2 * mpoles% &
                     max_order + 1, -2:2 * mpoles%max_order + 1, -2:2 * mpoles%max_order + 1)

    If (newjob) Then
      newjob = .false.

      ! Check for damped force-shifted coulombic and reaction field interactions
      ! and set force and potential shifting parameters dependingly

      electro%damp = electro%damping > zero_plus
      If (electro%damp) Then

        exp1 = Exp(-(electro%damping * rcut)**2)
        tt = 1.0_wp / (1.0_wp + pp * electro%damping * rcut)

        erc = tt * (aa1 + tt * (aa2 + tt * (aa3 + tt * (aa4 + tt * aa5)))) * exp1 / rcut
        fer = (erc + 2.0_wp * (electro%damping / sqrpi) * exp1) / rcut**2

        electro%force_shift = fer * rcut
        electro%energy_shift = -(erc + electro%force_shift * rcut)
      Else If (electro%key == ELECTROSTATIC_COULOMB_FORCE_SHIFT) Then
        electro%force_shift = 1.0_wp / rcut**2
        electro%energy_shift = -2.0_wp / rcut ! = -(1.0_wp/rcut+aa*rcut)
      End If

      ! set reaction field terms for RFC

      If (electro%key == ELECTROSTATIC_COULOMB_REACTION_FIELD) Then
        b0 = 2.0_wp * (electro%eps - 1.0_wp) / (2.0_wp * electro%eps + 1.0_wp)
        electro%reaction_field(0) = b0 / rcut**3
        electro%reaction_field(1) = (1.0_wp + 0.5_wp * b0) / rcut
        electro%reaction_field(2) = 0.5_wp * electro%reaction_field(0)
      End If
    End If

    ! initialise defaults for coulombic energy and force contributions

    coul = 0.0_wp; virele = 0.0_wp
    fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp
    tix = 0.0_wp; tiy = 0.0_wp; tiz = 0.0_wp
    tjx = 0.0_wp; tjy = 0.0_wp; tjz = 0.0_wp

    ! get the multipoles for sites i and j

    imp = mpoles%global_frame(:, iatm)
    jmp = mpoles%global_frame(:, jatm)

    ! ignore interaction if the charge is zero

    If (Maxval(Abs(imp)) <= zero_plus .or. Maxval(Abs(jmp)) <= zero_plus) Return

    ! get the components for site i and j infinitesimal rotations

    impx = mpoles%rotation_x(:, iatm)
    impy = mpoles%rotation_y(:, iatm)
    impz = mpoles%rotation_z(:, iatm)

    jmpx = mpoles%rotation_x(:, jatm)
    jmpy = mpoles%rotation_y(:, jatm)
    jmpz = mpoles%rotation_z(:, jatm)

    ! default convergence factor and derivative of 'r'

    talpha = 1.0_wp
    a1 = 0.0_wp

    ! scale imp multipoles

    imp = scale * imp * r4pie0 / electro%eps

    ! Electrostatics by ewald sum = direct coulombic

    If (Any([ELECTROSTATIC_EWALD, ELECTROSTATIC_COULOMB] == electro%key)) Then

      ! compute derivatives of 1/r kernel

      Call coul_deriv(1, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, d1)

      ! distance dependent dielectric

    Else If (electro%key == ELECTROSTATIC_DDDP) Then

      ! Compute derivatives of 1/r^2 kernel

      Call coul_deriv(2, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, d1)

      ! force shifted coulombic and reaction field

    Else If (Any([ELECTROSTATIC_COULOMB_FORCE_SHIFT, ELECTROSTATIC_COULOMB_REACTION_FIELD] == electro%key)) Then

      If (electro%damp) Then ! calculate damping contributions

        ! compute derivatives of 'r'

        Call coul_deriv(-1, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, a1)

        ! scale the derivatives of 'r'

        a1 = electro%force_shift * a1

        exp1 = Exp(-(electro%damping * rrr)**2)
        tt = 1.0_wp / (1.0_wp + pp * electro%damping * rrr)

        fer = erc / electro%damping

        ! compute derivatives of the ewald real space kernel

        Call ewald_deriv(-2, 2 * mpoles%max_order + 1, 1, fer, electro%damping * xdf, electro%damping * ydf, &
                         electro%damping * zdf, electro%damping * rrr, mpoles%max_order, d1)

        ! scale the derivatives into the right form

        d1 = 2.0_wp * electro%damping * d1 / sqrpi

      End If

      If (electro%key == ELECTROSTATIC_COULOMB_FORCE_SHIFT) Then ! force shifted coulombic
        If (.not. electro%damp) Then ! pure

          ! compute derivatives of '1/r'

          Call coul_deriv(1, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, d1)

          ! compute derivatives of 'r'

          Call coul_deriv(-1, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, a1)

          ! scale the derivatives of 'r' and add to d1

          d1 = d1 + electro%force_shift * a1
          a1 = 0.0_wp

        Else ! damped

          talpha = electro%damping

        End If
      Else If (electro%key == ELECTROSTATIC_COULOMB_REACTION_FIELD) Then ! reaction field
        If (.not. electro%damp) Then ! pure

          ! compute derivatives of '1/r'

          Call coul_deriv(1, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, d1)

          ! compute derivatives of 'r^2'

          Call coul_deriv(-2, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, a1)

          ! scale the derivatives of 'r' and add to d1

          d1 = d1 + electro%reaction_field(2) * a1
          a1 = 0.0_wp

        Else

          ! compute derivatives of 'r^2'

          Call coul_deriv(-2, 2 * mpoles%max_order + 1, xdf, ydf, zdf, rrr, b1)

          ! scale the derivatives of 'r^2' and add to scaled derivatives of 'r'

          a1 = a1 + electro%reaction_field(2) * b1

          talpha = electro%damping

        End If
      End If

    Else

      safe = .false.

    End If

    If (safe) Then

      kz = 1.0_wp
      Do k3 = 0, mpoles%max_order

        ky = kz
        Do k2 = 0, mpoles%max_order - k3

          kx = ky
          Do k1 = 0, mpoles%max_order - k3 - k2

            jj = mpoles%map(k1, k2, k3)

            If (Abs(jmp(jj)) > zero_plus) Then

              txyz = kx * jmp(jj)

              sz = 1.0_wp
              Do s3 = 0, mpoles%max_order
                ks3 = k3 + s3; ks31 = ks3 + 1

                sy = sz
                Do s2 = 0, mpoles%max_order - s3
                  ks2 = k2 + s2; ks21 = ks2 + 1

                  sx = sy
                  Do s1 = 0, mpoles%max_order - s3 - s2
                    ks1 = k1 + s1; ks11 = ks1 + 1

                    n = ks1 + ks2 + ks3
                    alphan = talpha**n

                    ii = mpoles%map(s1, s2, s3)

                    tmp = alphan * d1(ks1, ks2, ks3) + a1(ks1, ks2, ks3)

                    tmpi = txyz * tmp
                    tmpj = sx * imp(ii) * tmp

                    t2 = txyz * imp(ii)
                    t1 = alphan * t2

                    !  energy

                    coul = coul + t1 * d1(ks1, ks2, ks3) + t2 * a1(ks1, ks2, ks3)

                    !  force
                    t1 = t1 * talpha

                    fx = fx - t1 * d1(ks11, ks2, ks3) + t2 * a1(ks11, ks2, ks3)
                    fy = fy - t1 * d1(ks1, ks21, ks3) + t2 * a1(ks1, ks21, ks3)
                    fz = fz - t1 * d1(ks1, ks2, ks31) + t2 * a1(ks1, ks2, ks31)

                    !  torque on iatm

                    tix = tix + impx(ii) * tmpi
                    tiy = tiy + impy(ii) * tmpi
                    tiz = tiz + impz(ii) * tmpi

                    !  torque on jatm

                    tjx = tjx + jmpx(jj) * tmpj
                    tjy = tjy + jmpy(jj) * tmpj
                    tjz = tjz + jmpz(jj) * tmpj

                    sx = -sx
                  End Do

                  sy = -sy
                End Do

                sz = -sz
              End Do

            End If

            kx = -kx

          End Do

          ky = -ky

        End Do

        kz = -kz

      End Do

      If (Any([ELECTROSTATIC_EWALD, ELECTROSTATIC_DDDP, ELECTROSTATIC_COULOMB] == electro%key)) Then

        virele = -coul

      Else If (electro%key == ELECTROSTATIC_COULOMB_FORCE_SHIFT) Then ! force shifted coulombic

        virele = -(fx * xdf + fy * ydf + fz * zdf)

        ! shift potential

        tmp = electro%force_shift * rrr + electro%energy_shift
        coul = coul + tmp * imp(1) * jmp(1)

        ! shift torque

        tmpi = tmp * jmp(1)
        tix = tix + impx(1) * tmpi
        tiy = tiy + impy(1) * tmpi
        tiz = tiz + impz(1) * tmpi

        tmpj = tmp * imp(1)
        tjx = tjx + jmpx(1) * tmpj
        tjy = tjy + jmpy(1) * tmpj
        tjz = tjz + jmpz(1) * tmpj

      Else If (electro%key == ELECTROSTATIC_COULOMB_REACTION_FIELD) Then ! reaction field

        virele = -(fx * xdf + fy * ydf + fz * zdf)

        ! shift potential

        coul = coul - electro%reaction_field(1) * imp(1) * jmp(1)

        ! shift torque

        tmpi = -electro%reaction_field(1) * jmp(1)
        tix = tix + impx(1) * tmpi
        tiy = tiy + impy(1) * tmpi
        tiz = tiz + impz(1) * tmpi

        tmpj = -electro%reaction_field(1) * imp(1)
        tjx = tjx + jmpx(1) * tmpj
        tjy = tjy + jmpy(1) * tmpj
        tjz = tjz + jmpz(1) * tmpj

      End If

      tix = tix * r4pie0 / electro%eps
      tiy = tiy * r4pie0 / electro%eps
      tiz = tiz * r4pie0 / electro%eps

      If (iatm <= config%natms) Then

        mpoles%torque_x(iatm) = mpoles%torque_x(iatm) + tix
        mpoles%torque_y(iatm) = mpoles%torque_y(iatm) + tiy
        mpoles%torque_z(iatm) = mpoles%torque_z(iatm) + tiz

      End If

      If (jatm <= config%natms) Then

        mpoles%torque_x(jatm) = mpoles%torque_x(jatm) + tjx
        mpoles%torque_y(jatm) = mpoles%torque_y(jatm) + tjy
        mpoles%torque_z(jatm) = mpoles%torque_z(jatm) + tjz

      End If

    End If
  End Subroutine intra_mcoul

  Subroutine coul_fscp_mforces(iatm, xxt, yyt, zzt, rrt, engcpe, vircpe, stress, neigh, &
                               mpoles, electro, config)

    !!------------------------------------------------------------------------!
    !!
    !! dl_poly_4 subroutine for calculating coulombic energy and force terms
    !! in a periodic system using multipoles assuming a force shifted
    !! coulomb potential kernel
    !!
    !! U is proportional to ( 1/r + aa*r  + bb ) such that dU(neigh%cutoff)/dr = 0
    !! therefore aa = 1/(neigh%cutoff)**2 and U(neigh%cutoff) = 0 therefore bb = -2/(neigh%cutoff)
    !!
    !! Note: FS potential can be generalised (R1) by using a damping function
    !! as used for damping the real space coulombic interaction in the
    !! standard Ewald summation.  This generalisation applies when electro%damping > 0.
    !!
    !! R1: C.J. Fennell and J.D. Gezelter J. Chem. Phys. 124, 234104 (2006)
    !!
    !! copyright - daresbury laboratory
    !! author    - h.a.boateng february 2016
    !! amended   - i.t.todorov february 2016
    !! contrib   - a.v.brukhno & m.a.seaton august 2020 - 'half-halo' VNL
    !! refactoring:
    !!           - a.m.elena march-october 2018
    !!           - j.madge march-october 2018
    !!           - a.b.g.chalk march-october 2018
    !!           - i.scivetti march-october 2018
    !!
    !!------------------------------------------------------------------------!

    Integer,                                    Intent(In   ) :: iatm
    Type(neighbours_type),                      Intent(In   ) :: neigh
    Real(Kind=wp), Dimension(1:9),              Intent(InOut) :: stress
    Real(Kind=wp),                              Intent(  Out) :: vircpe, engcpe
    Real(Kind=wp), Dimension(1:neigh%max_list), Intent(In   ) :: rrt, zzt, yyt, xxt
    Type(mpole_type),                           Intent(InOut) :: mpoles
    Type(electrostatic_type),                   Intent(InOut) :: electro
    Type(configuration_type),                   Intent(InOut) :: config

    Integer       :: idi, ii, jatm, jj, k1, k2, k3, ks1, ks11, ks2, ks21, ks3, ks31, m, n, s1, s2, &
                     s3
    Logical, Save :: newjob = .true.
    Real(Kind=wp) :: alphan, engmpl, erfcr, fix, fiy, fiz, fx, fy, fz, kx, ky, kz, rrr, scl, &
                     strs1, strs2, strs3, strs5, strs6, strs9, sx, sy, sz, t1, t2, tix, tiy, tiz, &
                     tjx, tjy, tjz, tmp, tmpi, tmpj, txyz, jmpz(1:mpoles%max_mpoles), &
                     jmpy(1:mpoles%max_mpoles), jmpx(1:mpoles%max_mpoles), &
                     jmp(1:mpoles%max_mpoles), impz(1:mpoles%max_mpoles), &
                     impy(1:mpoles%max_mpoles), impx(1:mpoles%max_mpoles), &
                     imp(1:mpoles%max_mpoles), d1(-2:2 * mpoles%max_order + 1, -2:2 * mpoles% &
                     max_order + 1, -2:2 * mpoles%max_order + 1), a1(-2:2 * mpoles%max_order + 1, &
                     -2:2 * mpoles%max_order + 1, -2:2 * mpoles%max_order + 1)

    If (newjob) Then
      newjob = .false.

      If (electro%damp) Then

        Call electro%erfcgen(neigh%cutoff, electro%damping)

        electro%force_shift = electro%erfc_deriv%end_sample * neigh%cutoff
        electro%energy_shift = -(electro%erfc%end_sample + electro%force_shift * neigh%cutoff)

      Else

        ! set force and potential shifting parameters (screened terms)

        electro%force_shift = 1.0_wp / neigh%cutoff**2
        electro%energy_shift = -2.0_wp / neigh%cutoff ! = -(1.0_wp/neigh%cutoff+aa*neigh%cutoff)

      End If
    End If

    ! initialise potential energy and virial

    engcpe = 0.0_wp
    vircpe = 0.0_wp

    ! initialise stress tensor accumulators

    strs1 = 0.0_wp
    strs2 = 0.0_wp
    strs3 = 0.0_wp
    strs5 = 0.0_wp
    strs6 = 0.0_wp
    strs9 = 0.0_wp

    ! global identity of iatm

    idi = config%ltg(iatm)

    ! get the multipoles for site i

    imp = mpoles%global_frame(:, iatm)

    ! ignore interaction if the charge is zero

    If (Maxval(Abs(imp)) > zero_plus) Then

      ! get the components for site i infinitesimal rotations

      impx = mpoles%rotation_x(:, iatm)
      impy = mpoles%rotation_y(:, iatm)
      impz = mpoles%rotation_z(:, iatm)

      ! multipole scaler

      scl = r4pie0 / electro%eps

      ! scale imp multipoles

      imp = imp * scl

      ! load forces

      fix = config%parts(iatm)%fxx
      fiy = config%parts(iatm)%fyy
      fiz = config%parts(iatm)%fzz

      ! initialize torques for atom i (temporary)

      tix = 0.0_wp; tiy = 0.0_wp; tiz = 0.0_wp

      ! start of primary loop for forces evaluation

      Do m = 1, neigh%list(0, iatm)

        ! atomic index

        jatm = neigh%list(m, iatm)

        ! get the multipoles for site j

        jmp = mpoles%global_frame(:, jatm)

        ! interatomic distance

        rrr = rrt(m)

        ! truncation of potential

        If (Maxval(Abs(jmp)) > zero_plus .and. rrr < neigh%cutoff) Then

          ! get the components for site j infinitesimal rotations

          jmpx = mpoles%rotation_x(:, jatm)
          jmpy = mpoles%rotation_y(:, jatm)
          jmpz = mpoles%rotation_z(:, jatm)

          ! compute derivatives of kernel

          If (electro%damp) Then

            ! compute derivatives of 'r'

            Call coul_deriv(-1, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrr, a1)

            ! scale the derivatives of 'r'

            a1 = electro%force_shift * a1

            ! get the value of the ewald real space kernel using 3pt interpolation

            ! erfcr = (t1 + (t2-t1)*ppp*0.5_wp)/electro%damping

            erfcr = electro%erfc%calc(rrr) / electro%damping
            ! compute derivatives of the ewald real space kernel

            Call ewald_deriv(-2, 2 * mpoles%max_order + 1, 1, erfcr, electro%damping * xxt(m), &
                             electro%damping * yyt(m), electro%damping * zzt(m), electro%damping * rrr, mpoles%max_order, d1)

            ! scale the derivatives into the right form

            d1 = 2.0_wp * electro%damping * d1 / sqrpi

            ! calculate forces

            engmpl = 0.0_wp
            fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp
            tjx = 0.0_wp; tjy = 0.0_wp; tjz = 0.0_wp

            If (mpoles%max_order < 5) Then

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then
                      Call explicit_fscp_rfp_loops &
                        (2 * mpoles%max_order + 1, k1, k2, k3, electro%damping, d1, a1, &
                         imp, impx, impy, impz, tix, tiy, tiz, &
                         kx * jmp(jj), jmpx(jj), jmpy(jj), jmpz(jj), tjx, tjy, tjz, &
                         engmpl, fx, fy, fz, mpoles)
                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            Else

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then

                      txyz = kx * jmp(jj)

                      sz = 1.0_wp
                      Do s3 = 0, mpoles%max_order
                        ks3 = k3 + s3; ks31 = ks3 + 1

                        sy = sz
                        Do s2 = 0, mpoles%max_order - s3
                          ks2 = k2 + s2; ks21 = ks2 + 1

                          sx = sy
                          Do s1 = 0, mpoles%max_order - s3 - s2
                            ks1 = k1 + s1; ks11 = ks1 + 1

                            n = ks1 + ks2 + ks3
                            alphan = electro%damping**n

                            ii = mpoles%map(s1, s2, s3)

                            tmp = alphan * d1(ks1, ks2, ks3) + a1(ks1, ks2, ks3)

                            tmpi = txyz * tmp
                            tmpj = sx * imp(ii) * tmp

                            t2 = txyz * imp(ii)
                            t1 = alphan * t2

                            ! energy

                            engmpl = engmpl + t1 * d1(ks1, ks2, ks3) + t2 * a1(ks1, ks2, ks3)

                            ! force
                            t1 = t1 * electro%damping

                            fx = fx - t1 * d1(ks11, ks2, ks3) + t2 * a1(ks11, ks2, ks3)
                            fy = fy - t1 * d1(ks1, ks21, ks3) + t2 * a1(ks1, ks21, ks3)
                            fz = fz - t1 * d1(ks1, ks2, ks31) + t2 * a1(ks1, ks2, ks31)

                            ! torque on iatm

                            tix = tix + impx(ii) * tmpi
                            tiy = tiy + impy(ii) * tmpi
                            tiz = tiz + impz(ii) * tmpi

                            ! torque on jatm

                            tjx = tjx + jmpx(jj) * tmpj
                            tjy = tjy + jmpy(jj) * tmpj
                            tjz = tjz + jmpz(jj) * tmpj

                            sx = -sx
                          End Do

                          sy = -sy
                        End Do

                        sz = -sz
                      End Do

                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            End If

          Else

            ! compute derivatives of '1/r'

            Call coul_deriv(1, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrr, d1)

            ! compute derivatives of 'r'

            Call coul_deriv(-1, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrr, a1)

            ! scale the derivatives of 'r' and add to d1

            d1 = d1 + a1 / neigh%cutoff**2

            ! calculate potential forces

            engmpl = 0.0_wp
            fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp
            tjx = 0.0_wp; tjy = 0.0_wp; tjz = 0.0_wp

            If (mpoles%max_order < 5) Then

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then
                      Call explicit_ewald_real_loops &
                        (-2, 2 * mpoles%max_order + 1, k1, k2, k3, 1.0_wp, d1, &
                         imp, impx, impy, impz, tix, tiy, tiz, &
                         kx * jmp(jj), jmpx(jj), jmpy(jj), jmpz(jj), tjx, tjy, tjz, &
                         engmpl, fx, fy, fz, mpoles)
                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            Else

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then

                      txyz = kx * jmp(jj)

                      sz = 1.0_wp
                      Do s3 = 0, mpoles%max_order
                        ks3 = k3 + s3; ks31 = ks3 + 1

                        sy = sz
                        Do s2 = 0, mpoles%max_order - s3
                          ks2 = k2 + s2; ks21 = ks2 + 1

                          sx = sy
                          Do s1 = 0, mpoles%max_order - s3 - s2
                            ks1 = k1 + s1; ks11 = ks1 + 1

                            n = ks1 + ks2 + ks3

                            ii = mpoles%map(s1, s2, s3)

                            tmp = d1(ks1, ks2, ks3)

                            tmpi = txyz * tmp
                            tmpj = sx * imp(ii) * tmp

                            t1 = txyz * imp(ii)

                            ! energy

                            engmpl = engmpl + t1 * tmp

                            ! force

                            fx = fx - t1 * d1(ks11, ks2, ks3)
                            fy = fy - t1 * d1(ks1, ks21, ks3)
                            fz = fz - t1 * d1(ks1, ks2, ks31)

                            ! torque on iatm

                            tix = tix + impx(ii) * tmpi
                            tiy = tiy + impy(ii) * tmpi
                            tiz = tiz + impz(ii) * tmpi

                            ! torque on jatm

                            tjx = tjx + jmpx(jj) * tmpj
                            tjy = tjy + jmpy(jj) * tmpj
                            tjz = tjz + jmpz(jj) * tmpj

                            sx = -sx
                          End Do

                          sy = -sy
                        End Do

                        sz = -sz
                      End Do

                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            End If

          End If

          ! shift potential

          tmp = electro%force_shift * rrr + electro%energy_shift
          engmpl = engmpl + tmp * imp(1) * jmp(1)

          ! shift torque

          tmpi = tmp * jmp(1)
          tix = tix + impx(1) * tmpi
          tiy = tiy + impy(1) * tmpi
          tiz = tiz + impz(1) * tmpi

          tmpj = tmp * imp(1)
          tjx = tjx + jmpx(1) * tmpj
          tjy = tjy + jmpy(1) * tmpj
          tjz = tjz + jmpz(1) * tmpj

          fix = fix + fx
          fiy = fiy + fy
          fiz = fiz + fz

#ifndef HALF_HALO
          If (jatm <= config%natms) Then
#endif /* HALF_HALO */

            config%parts(jatm)%fxx = config%parts(jatm)%fxx - fx
            config%parts(jatm)%fyy = config%parts(jatm)%fyy - fy
            config%parts(jatm)%fzz = config%parts(jatm)%fzz - fz

            mpoles%torque_x(jatm) = mpoles%torque_x(jatm) + tjx
            mpoles%torque_y(jatm) = mpoles%torque_y(jatm) + tjy
            mpoles%torque_z(jatm) = mpoles%torque_z(jatm) + tjz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

#ifndef HALF_HALO
          If (jatm <= config%natms .or. idi < config%ltg(jatm)) Then
#endif /* HALF_HALO */

            ! accumulate potential energy

            engcpe = engcpe + engmpl

            ! calculate virial

            vircpe = vircpe - (fx * xxt(m) + fy * yyt(m) + fz * zzt(m))

            ! calculate stress tensor

            strs1 = strs1 + xxt(m) * fx
            strs2 = strs2 + xxt(m) * fy
            strs3 = strs3 + xxt(m) * fz
            strs5 = strs5 + yyt(m) * fy
            strs6 = strs6 + yyt(m) * fz
            strs9 = strs9 + zzt(m) * fz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

        End If

      End Do

      ! load back forces

      config%parts(iatm)%fxx = fix
      config%parts(iatm)%fyy = fiy
      config%parts(iatm)%fzz = fiz

      ! and torques due to multipoles

      mpoles%torque_x(iatm) = mpoles%torque_x(iatm) + scl * tix
      mpoles%torque_y(iatm) = mpoles%torque_y(iatm) + scl * tiy
      mpoles%torque_z(iatm) = mpoles%torque_z(iatm) + scl * tiz

      ! complete stress tensor

      stress(1) = stress(1) + strs1
      stress(2) = stress(2) + strs2
      stress(3) = stress(3) + strs3
      stress(4) = stress(4) + strs2
      stress(5) = stress(5) + strs5
      stress(6) = stress(6) + strs6
      stress(7) = stress(7) + strs3
      stress(8) = stress(8) + strs6
      stress(9) = stress(9) + strs9

    End If

  End Subroutine coul_fscp_mforces

  Subroutine coul_rfp_mforces(iatm, xxt, yyt, zzt, rrt, engcpe, vircpe, stress, neigh, &
                              mpoles, electro, config)

    !!------------------------------------------------------------------------!
    !!
    !! dl_poly_4 subroutine for calculating coulombic energy and force terms
    !! in a periodic system using multipoles assuming a reaction field
    !! potential (corrected for the existence of a dipole moment outside neigh%cutoff)
    !!
    !! Note: RF potential can be generalised (R1) by using a damping function
    !! as used for damping the real space coulombic interaction in the
    !! standard Ewald summation.  This generalisation applies when electro%damping > 0.
    !!
    !! R1: C.J. Fennell and J.D. Gezelter J. Chem. Phys. 124, 234104 (2006)
    !! R2: M Neumann, J Chem Phys, 82 (12), 5663, (1985)
    !!
    !! copyright - daresbury laboratory
    !! author    - h.a.boateng february 2016
    !! amended   - i.t.todorov february 2016
    !! contrib   - a.v.brukhno & m.a.seaton august 2020 - 'half-halo' VNL
    !!
    !!------------------------------------------------------------------------!

    Integer,                                    Intent(In   ) :: iatm
    Type(neighbours_type),                      Intent(In   ) :: neigh
    Real(Kind=wp), Dimension(1:9),              Intent(InOut) :: stress
    Real(Kind=wp),                              Intent(  Out) :: vircpe, engcpe
    Real(Kind=wp), Dimension(1:neigh%max_list), Intent(In   ) :: rrt, zzt, yyt, xxt
    Type(mpole_type),                           Intent(InOut) :: mpoles
    Type(electrostatic_type),                   Intent(InOut) :: electro
    Type(configuration_type),                   Intent(InOut) :: config

    Integer       :: idi, ii, jatm, jj, k1, k2, k3, ks1, ks11, ks2, ks21, ks3, ks31, m, n, s1, s2, &
                     s3
    Logical, Save :: newjob = .true.
    Real(Kind=wp) :: alphan, b0, engmpl, erfcr, fix, fiy, fiz, fx, fy, fz, kx, ky, kz, rrr, scl, &
                     strs1, strs2, strs3, strs5, strs6, strs9, sx, sy, sz, t1, t2, tix, tiy, tiz, &
                     tjx, tjy, tjz, tmp, tmpi, tmpj, txyz, jmpz(1:mpoles%max_mpoles), &
                     jmpy(1:mpoles%max_mpoles), jmpx(1:mpoles%max_mpoles), &
                     jmp(1:mpoles%max_mpoles), impz(1:mpoles%max_mpoles), &
                     impy(1:mpoles%max_mpoles), impx(1:mpoles%max_mpoles), &
                     imp(1:mpoles%max_mpoles), d1(-2:2 * mpoles%max_order + 1, -2:2 * mpoles% &
                     max_order + 1, -2:2 * mpoles%max_order + 1), a1(-2:2 * mpoles%max_order + 1, &
                     -2:2 * mpoles%max_order + 1, -2:2 * mpoles%max_order + 1)

    If (newjob) Then
      newjob = .false.

      ! reaction field terms

      b0 = 2.0_wp * (electro%eps - 1.0_wp) / (2.0_wp * electro%eps + 1.0_wp)
      electro%reaction_field(0) = b0 / neigh%cutoff**3
      electro%reaction_field(1) = (1.0_wp + 0.5_wp * b0) / neigh%cutoff
      electro%reaction_field(2) = 0.5_wp * electro%reaction_field(0)

      If (electro%damp) Then

        Call electro%erfcgen(neigh%cutoff, electro%damping)

        electro%force_shift = electro%erfc_deriv%end_sample * neigh%cutoff
        electro%energy_shift = -(electro%erfc%end_sample + electro%force_shift * neigh%cutoff)

      Else

        ! set force and potential shifting parameters (screened terms)

        electro%force_shift = 1.0_wp / neigh%cutoff**2
        electro%energy_shift = -2.0_wp / neigh%cutoff ! = -(1.0_wp/neigh%cutoff+aa*neigh%cutoff)

      End If
    End If

    ! initialise potential energy and virial

    engcpe = 0.0_wp
    vircpe = 0.0_wp

    ! initialise stress tensor accumulators

    strs1 = 0.0_wp
    strs2 = 0.0_wp
    strs3 = 0.0_wp
    strs5 = 0.0_wp
    strs6 = 0.0_wp
    strs9 = 0.0_wp

    ! global identity of iatm

    idi = config%ltg(iatm)

    ! get the multipoles for site i

    imp = mpoles%global_frame(:, iatm)

    ! ignore interaction if the charge is zero

    If (Maxval(Abs(imp)) > zero_plus) Then

      ! get the components for site i infinitesimal rotations

      impx = mpoles%rotation_x(:, iatm)
      impy = mpoles%rotation_y(:, iatm)
      impz = mpoles%rotation_z(:, iatm)

      ! multipole scaler

      scl = r4pie0 / electro%eps

      ! scale imp multipoles

      imp = imp * scl

      ! load forces

      fix = config%parts(iatm)%fxx
      fiy = config%parts(iatm)%fyy
      fiz = config%parts(iatm)%fzz

      ! initialize torques for atom i (temporary)

      tix = 0.0_wp; tiy = 0.0_wp; tiz = 0.0_wp

      Do m = 1, neigh%list(0, iatm)

        ! atomic index

        jatm = neigh%list(m, iatm)

        ! get the multipoles for site j

        jmp = mpoles%global_frame(:, jatm)

        ! interatomic distance

        rrr = rrt(m)

        ! truncation of potential

        If (Maxval(Abs(jmp)) > zero_plus .and. rrr < neigh%cutoff) Then

          ! get the components for site j infinitesimal rotations

          jmpx = mpoles%rotation_x(:, jatm)
          jmpy = mpoles%rotation_y(:, jatm)
          jmpz = mpoles%rotation_z(:, jatm)

          ! compute derivatives of kernel

          If (electro%damp) Then

            ! compute derivatives of 'r^2'

            Call coul_deriv(-2, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrr, a1)

            ! scale the derivatives of 'r^2'

            a1 = electro%reaction_field(2) * a1

            ! compute derivatives of 'r'

            Call coul_deriv(-1, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrr, d1)

            ! scale the derivatives of 'r' and add to a1

            a1 = a1 + electro%force_shift * d1

            ! get the value of the ewald real space kernel using 3pt interpolation

            ! erfcr = (t1 + (t2-t1)*ppp*0.5_wp)/electro%damping
            erfcr = electro%erfc%calc(rrr) / electro%damping

            ! compute derivatives of the ewald real space kernel

            Call ewald_deriv(-2, 2 * mpoles%max_order + 1, 1, erfcr, electro%damping * xxt(m), &
                             electro%damping * yyt(m), electro%damping * zzt(m), electro%damping * rrr, mpoles%max_order, d1)

            ! scale the derivatives into the right form

            d1 = 2.0_wp * electro%damping * d1 / sqrpi

            ! calculate forces

            engmpl = 0.0_wp
            fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp
            tjx = 0.0_wp; tjy = 0.0_wp; tjz = 0.0_wp

            If (mpoles%max_order < 5) Then

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then
                      Call explicit_fscp_rfp_loops &
                        (2 * mpoles%max_order + 1, k1, k2, k3, electro%damping, d1, a1, &
                         imp, impx, impy, impz, tix, tiy, tiz, &
                         kx * jmp(jj), jmpx(jj), jmpy(jj), jmpz(jj), tjx, tjy, tjz, &
                         engmpl, fx, fy, fz, mpoles)
                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            Else

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then

                      txyz = kx * jmp(jj)

                      sz = 1.0_wp
                      Do s3 = 0, mpoles%max_order
                        ks3 = k3 + s3; ks31 = ks3 + 1

                        sy = sz
                        Do s2 = 0, mpoles%max_order - s3
                          ks2 = k2 + s2; ks21 = ks2 + 1

                          sx = sy
                          Do s1 = 0, mpoles%max_order - s3 - s2
                            ks1 = k1 + s1; ks11 = ks1 + 1

                            n = ks1 + ks2 + ks3
                            alphan = electro%damping**n

                            ii = mpoles%map(s1, s2, s3)

                            tmp = alphan * d1(ks1, ks2, ks3) + a1(ks1, ks2, ks3)

                            tmpi = txyz * tmp
                            tmpj = sx * imp(ii) * tmp

                            t2 = txyz * imp(ii)
                            t1 = alphan * t2

                            ! energy

                            engmpl = engmpl + t1 * d1(ks1, ks2, ks3) + t2 * a1(ks1, ks2, ks3)

                            ! force
                            t1 = t1 * electro%damping

                            fx = fx - t1 * d1(ks11, ks2, ks3) + t2 * a1(ks11, ks2, ks3)
                            fy = fy - t1 * d1(ks1, ks21, ks3) + t2 * a1(ks1, ks21, ks3)
                            fz = fz - t1 * d1(ks1, ks2, ks31) + t2 * a1(ks1, ks2, ks31)

                            ! torque on iatm

                            tix = tix + impx(ii) * tmpi
                            tiy = tiy + impy(ii) * tmpi
                            tiz = tiz + impz(ii) * tmpi

                            ! torque on jatm

                            tjx = tjx + jmpx(jj) * tmpj
                            tjy = tjy + jmpy(jj) * tmpj
                            tjz = tjz + jmpz(jj) * tmpj

                            sx = -sx
                          End Do

                          sy = -sy
                        End Do

                        sz = -sz
                      End Do

                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            End If

            ! shift potential

            tmp = electro%energy_shift - 0.5_wp * b0 / neigh%cutoff
            engmpl = engmpl + tmp * imp(1) * jmp(1)

            ! shift torque

            tmpi = tmp * jmp(1)
            tix = tix + impx(1) * tmpi
            tiy = tiy + impy(1) * tmpi
            tiz = tiz + impz(1) * tmpi

            tmpj = tmp * imp(1)
            tjx = tjx + jmpx(1) * tmpj
            tjy = tjy + jmpy(1) * tmpj
            tjz = tjz + jmpz(1) * tmpj

          Else

            ! compute derivatives of '1/r'

            Call coul_deriv(1, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrr, d1)

            ! compute derivatives of 'r^2'

            Call coul_deriv(-2, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrr, a1)

            ! scale the derivatives of 'r' and add to d1

            d1 = d1 + electro%reaction_field(2) * a1

            ! calculate potential forces

            engmpl = 0.0_wp
            fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp
            tjx = 0.0_wp; tjy = 0.0_wp; tjz = 0.0_wp

            If (mpoles%max_order < 5) Then

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then
                      Call explicit_ewald_real_loops &
                        (-2, 2 * mpoles%max_order + 1, k1, k2, k3, 1.0_wp, d1, &
                         imp, impx, impy, impz, tix, tiy, tiz, &
                         kx * jmp(jj), jmpx(jj), jmpy(jj), jmpz(jj), tjx, tjy, tjz, &
                         engmpl, fx, fy, fz, mpoles)
                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            Else

              kz = 1.0_wp
              Do k3 = 0, mpoles%max_order

                ky = kz
                Do k2 = 0, mpoles%max_order - k3

                  kx = ky
                  Do k1 = 0, mpoles%max_order - k3 - k2

                    jj = mpoles%map(k1, k2, k3)

                    If (Abs(jmp(jj)) > zero_plus) Then

                      txyz = kx * jmp(jj)

                      sz = 1.0_wp
                      Do s3 = 0, mpoles%max_order
                        ks3 = k3 + s3; ks31 = ks3 + 1

                        sy = sz
                        Do s2 = 0, mpoles%max_order - s3
                          ks2 = k2 + s2; ks21 = ks2 + 1

                          sx = sy
                          Do s1 = 0, mpoles%max_order - s3 - s2
                            ks1 = k1 + s1; ks11 = ks1 + 1

                            n = ks1 + ks2 + ks3

                            ii = mpoles%map(s1, s2, s3)

                            tmp = d1(ks1, ks2, ks3)

                            tmpi = txyz * tmp
                            tmpj = sx * imp(ii) * tmp

                            t1 = txyz * imp(ii)

                            ! energy

                            engmpl = engmpl + t1 * tmp

                            ! force

                            fx = fx - t1 * d1(ks11, ks2, ks3)
                            fy = fy - t1 * d1(ks1, ks21, ks3)
                            fz = fz - t1 * d1(ks1, ks2, ks31)

                            ! torque on iatm

                            tix = tix + impx(ii) * tmpi
                            tiy = tiy + impy(ii) * tmpi
                            tiz = tiz + impz(ii) * tmpi

                            ! torque on jatm

                            tjx = tjx + jmpx(jj) * tmpj
                            tjy = tjy + jmpy(jj) * tmpj
                            tjz = tjz + jmpz(jj) * tmpj

                            sx = -sx
                          End Do

                          sy = -sy
                        End Do

                        sz = -sz
                      End Do

                    End If

                    kx = -kx

                  End Do

                  ky = -ky

                End Do

                kz = -kz

              End Do

            End If

            ! shift potential

            engmpl = engmpl - electro%reaction_field(1) * imp(1) * jmp(1)

            ! shift torque

            tmpi = -electro%reaction_field(1) * jmp(1)
            tix = tix + impx(1) * tmpi
            tiy = tiy + impy(1) * tmpi
            tiz = tiz + impz(1) * tmpi

            tmpj = -electro%reaction_field(1) * imp(1)
            tjx = tjx + jmpx(1) * tmpj
            tjy = tjy + jmpy(1) * tmpj
            tjz = tjz + jmpz(1) * tmpj

          End If

          fix = fix + fx
          fiy = fiy + fy
          fiz = fiz + fz

#ifndef HALF_HALO
          If (jatm <= config%natms) Then
#endif /* HALF_HALO */

            config%parts(jatm)%fxx = config%parts(jatm)%fxx - fx
            config%parts(jatm)%fyy = config%parts(jatm)%fyy - fy
            config%parts(jatm)%fzz = config%parts(jatm)%fzz - fz

            mpoles%torque_x(jatm) = mpoles%torque_x(jatm) + tjx
            mpoles%torque_y(jatm) = mpoles%torque_y(jatm) + tjy
            mpoles%torque_z(jatm) = mpoles%torque_z(jatm) + tjz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

#ifndef HALF_HALO
          If (jatm <= config%natms .or. idi < config%ltg(jatm)) Then
#endif /* HALF_HALO */

            ! accumulate potential energy

            engcpe = engcpe + engmpl

            ! calculate virial

            vircpe = vircpe - (fx * xxt(m) + fy * yyt(m) + fz * zzt(m))

            ! calculate stress tensor

            strs1 = strs1 + xxt(m) * fx
            strs2 = strs2 + xxt(m) * fy
            strs3 = strs3 + xxt(m) * fz
            strs5 = strs5 + yyt(m) * fy
            strs6 = strs6 + yyt(m) * fz
            strs9 = strs9 + zzt(m) * fz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

        End If

      End Do

      ! load back forces

      config%parts(iatm)%fxx = fix
      config%parts(iatm)%fyy = fiy
      config%parts(iatm)%fzz = fiz

      ! and torques due to multipoles

      mpoles%torque_x(iatm) = mpoles%torque_x(iatm) + scl * tix
      mpoles%torque_y(iatm) = mpoles%torque_y(iatm) + scl * tiy
      mpoles%torque_z(iatm) = mpoles%torque_z(iatm) + scl * tiz

      ! complete stress tensor

      stress(1) = stress(1) + strs1
      stress(2) = stress(2) + strs2
      stress(3) = stress(3) + strs3
      stress(4) = stress(4) + strs2
      stress(5) = stress(5) + strs5
      stress(6) = stress(6) + strs6
      stress(7) = stress(7) + strs3
      stress(8) = stress(8) + strs6
      stress(9) = stress(9) + strs9

    End If

  End Subroutine coul_rfp_mforces

  Subroutine coul_cp_mforces &
    (iatm, eps, xxt, yyt, zzt, rrt, engcpe, vircpe, stress, neigh, mpoles, config)

    !!------------------------------------------------------------------------!
    !!
    !! dl_poly_4 subroutine for calculating coulombic energy and force terms
    !! in a periodic system using multipoles with 1/r kernel with no
    !! truncation or damping
    !!
    !! copyright - daresbury laboratory
    !! author    - h.a.boateng february 2016
    !! amended   - i.t.todorov february 2016
    !! contrib   - a.v.brukhno & m.a.seaton august 2020 - 'half-halo' VNL
    !! refactoring:
    !!           - a.m.elena march-october 2018
    !!           - j.madge march-october 2018
    !!           - a.b.g.chalk march-october 2018
    !!           - i.scivetti march-october 2018
    !!
    !!------------------------------------------------------------------------!

    Integer,                                    Intent(In   ) :: iatm
    Real(Kind=wp),                              Intent(In   ) :: eps
    Type(neighbours_type),                      Intent(In   ) :: neigh
    Real(Kind=wp), Dimension(1:9),              Intent(InOut) :: stress
    Real(Kind=wp),                              Intent(  Out) :: vircpe, engcpe
    Real(Kind=wp), Dimension(1:neigh%max_list), Intent(In   ) :: rrt, zzt, yyt, xxt
    Type(mpole_type),                           Intent(InOut) :: mpoles
    Type(configuration_type),                   Intent(InOut) :: config

    Integer       :: idi, ii, jatm, jj, k1, k2, k3, ks1, ks11, ks2, ks21, ks3, ks31, m, s1, s2, s3
    Real(Kind=wp) :: engmpl, fix, fiy, fiz, fx, fy, fz, kx, ky, kz, scl, strs1, strs2, strs3, &
                     strs5, strs6, strs9, sx, sy, sz, t1, tix, tiy, tiz, tjx, tjy, tjz, tmp, tmpi, &
                     tmpj, txyz, jmpz(1:mpoles%max_mpoles), jmpy(1:mpoles%max_mpoles), &
                     jmpx(1:mpoles%max_mpoles), jmp(1:mpoles%max_mpoles), &
                     impz(1:mpoles%max_mpoles), impy(1:mpoles%max_mpoles), &
                     impx(1:mpoles%max_mpoles), imp(1:mpoles%max_mpoles), d1(-2:2 * mpoles% &
                     max_order + 1, -2:2 * mpoles%max_order + 1, -2:2 * mpoles%max_order + 1)

    ! initialise potential energy and virial

    engcpe = 0.0_wp
    vircpe = 0.0_wp

    ! initialise stress tensor accumulators

    strs1 = 0.0_wp
    strs2 = 0.0_wp
    strs3 = 0.0_wp
    strs5 = 0.0_wp
    strs6 = 0.0_wp
    strs9 = 0.0_wp

    ! global identity of iatm

    idi = config%ltg(iatm)

    ! get the multipoles for site i

    imp = mpoles%global_frame(:, iatm)

    ! ignore interaction if the charge is zero

    If (Maxval(Abs(imp)) > zero_plus) Then

      ! get the components for site i infinitesimal rotations

      impx = mpoles%rotation_x(:, iatm)
      impy = mpoles%rotation_y(:, iatm)
      impz = mpoles%rotation_z(:, iatm)

      ! multipole scaler

      scl = r4pie0 / eps

      ! scale imp multipoles

      imp = imp * scl

      ! load forces

      fix = config%parts(iatm)%fxx
      fiy = config%parts(iatm)%fyy
      fiz = config%parts(iatm)%fzz

      ! initialize torques for atom i (temporary)

      tix = 0.0_wp; tiy = 0.0_wp; tiz = 0.0_wp

      ! start of primary loop for forces evaluation

      Do m = 1, neigh%list(0, iatm)

        ! atomic index

        jatm = neigh%list(m, iatm)

        ! get the multipoles for site j

        jmp = mpoles%global_frame(:, jatm)

        ! truncation of potential - rrt(m) is the interatomic distance

        If (Maxval(Abs(jmp)) > zero_plus .and. rrt(m) < neigh%cutoff) Then

          ! get the components for site j infinitesimal rotations

          jmpx = mpoles%rotation_x(:, jatm)
          jmpy = mpoles%rotation_y(:, jatm)
          jmpz = mpoles%rotation_z(:, jatm)

          ! compute derivatives of kernel

          Call coul_deriv(1, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrt(m), d1)

          ! calculate forces

          engmpl = 0.0_wp
          fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp
          tjx = 0.0_wp; tjy = 0.0_wp; tjz = 0.0_wp

          If (mpoles%max_order < 5) Then

            kz = 1.0_wp
            Do k3 = 0, mpoles%max_order

              ky = kz
              Do k2 = 0, mpoles%max_order - k3

                kx = ky
                Do k1 = 0, mpoles%max_order - k3 - k2

                  jj = mpoles%map(k1, k2, k3)

                  If (Abs(jmp(jj)) > zero_plus) Then
                    Call explicit_ewald_real_loops &
                      (-2, 2 * mpoles%max_order + 1, k1, k2, k3, 1.0_wp, d1, &
                       imp, impx, impy, impz, tix, tiy, tiz, &
                       kx * jmp(jj), jmpx(jj), jmpy(jj), jmpz(jj), tjx, tjy, tjz, &
                       engmpl, fx, fy, fz, mpoles)
                  End If

                  kx = -kx

                End Do

                ky = -ky

              End Do

              kz = -kz

            End Do

          Else

            kz = 1.0_wp
            Do k3 = 0, mpoles%max_order

              ky = kz
              Do k2 = 0, mpoles%max_order - k3

                kx = ky
                Do k1 = 0, mpoles%max_order - k3 - k2

                  jj = mpoles%map(k1, k2, k3)

                  If (Abs(jmp(jj)) > zero_plus) Then

                    txyz = kx * jmp(jj)

                    sz = 1.0_wp
                    Do s3 = 0, mpoles%max_order
                      ks3 = k3 + s3; ks31 = ks3 + 1

                      sy = sz
                      Do s2 = 0, mpoles%max_order - s3
                        ks2 = k2 + s2; ks21 = ks2 + 1

                        sx = sy
                        Do s1 = 0, mpoles%max_order - s3 - s2
                          ks1 = k1 + s1; ks11 = ks1 + 1

                          ii = mpoles%map(s1, s2, s3)

                          tmp = d1(ks1, ks2, ks3)

                          tmpi = txyz * tmp
                          tmpj = sx * imp(ii) * tmp

                          t1 = txyz * imp(ii)

                          ! energy

                          engmpl = engmpl + t1 * tmp

                          ! force

                          fx = fx - t1 * d1(ks11, ks2, ks3)
                          fy = fy - t1 * d1(ks1, ks21, ks3)
                          fz = fz - t1 * d1(ks1, ks2, ks31)

                          ! torque on iatm

                          tix = tix + impx(ii) * tmpi
                          tiy = tiy + impy(ii) * tmpi
                          tiz = tiz + impz(ii) * tmpi

                          ! torque on jatm

                          tjx = tjx + jmpx(jj) * tmpj
                          tjy = tjy + jmpy(jj) * tmpj
                          tjz = tjz + jmpz(jj) * tmpj

                          sx = -sx
                        End Do

                        sy = -sy
                      End Do

                      sz = -sz
                    End Do

                  End If

                  kx = -kx

                End Do

                ky = -ky

              End Do

              kz = -kz

            End Do

          End If

          fix = fix + fx
          fiy = fiy + fy
          fiz = fiz + fz

#ifndef HALF_HALO
          If (jatm <= config%natms) Then
#endif /* HALF_HALO */

            config%parts(jatm)%fxx = config%parts(jatm)%fxx - fx
            config%parts(jatm)%fyy = config%parts(jatm)%fyy - fy
            config%parts(jatm)%fzz = config%parts(jatm)%fzz - fz

            mpoles%torque_x(jatm) = mpoles%torque_x(jatm) + tjx
            mpoles%torque_y(jatm) = mpoles%torque_y(jatm) + tjy
            mpoles%torque_z(jatm) = mpoles%torque_z(jatm) + tjz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

#ifndef HALF_HALO
          If (jatm <= config%natms .or. idi < config%ltg(jatm)) Then
#endif /* HALF_HALO */

            ! accumulate potential energy

            engcpe = engcpe + engmpl

            ! calculate stress tensor

            strs1 = strs1 + xxt(m) * fx
            strs2 = strs2 + xxt(m) * fy
            strs3 = strs3 + xxt(m) * fz
            strs5 = strs5 + yyt(m) * fy
            strs6 = strs6 + yyt(m) * fz
            strs9 = strs9 + zzt(m) * fz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

        End If

      End Do

      ! load back forces

      config%parts(iatm)%fxx = fix
      config%parts(iatm)%fyy = fiy
      config%parts(iatm)%fzz = fiz

      ! and torques due to multipoles

      mpoles%torque_x(iatm) = mpoles%torque_x(iatm) + scl * tix
      mpoles%torque_y(iatm) = mpoles%torque_y(iatm) + scl * tiy
      mpoles%torque_z(iatm) = mpoles%torque_z(iatm) + scl * tiz

      ! virial

      vircpe = -engcpe

      ! complete stress tensor

      stress(1) = stress(1) + strs1
      stress(2) = stress(2) + strs2
      stress(3) = stress(3) + strs3
      stress(4) = stress(4) + strs2
      stress(5) = stress(5) + strs5
      stress(6) = stress(6) + strs6
      stress(7) = stress(7) + strs3
      stress(8) = stress(8) + strs6
      stress(9) = stress(9) + strs9

    End If

  End Subroutine coul_cp_mforces

  Subroutine coul_dddp_mforces &
    (iatm, eps, xxt, yyt, zzt, rrt, engcpe, vircpe, stress, neigh, mpoles, config)

    !!------------------------------------------------------------------------!
    !!
    !! dl_poly_4 subroutine for calculating coulombic energy and force terms
    !! in a periodic system using multipoles with 1/r kernel assuming a
    !! distance dependent dielectric 'constant'
    !!
    !! copyright - daresbury laboratory
    !! author    - h.a.boateng february 2016
    !! amended   - i.t.todorov february 2016
    !! contrib   - a.v.brukhno & m.a.seaton august 2020 - 'half-halo' VNL
    !! refactoring:
    !!           - a.m.elena march-october 2018
    !!           - j.madge march-october 2018
    !!           - a.b.g.chalk march-october 2018
    !!           - i.scivetti march-october 2018
    !!
    !!------------------------------------------------------------------------!

    Integer,                                    Intent(In   ) :: iatm
    Real(Kind=wp),                              Intent(In   ) :: eps
    Type(neighbours_type),                      Intent(In   ) :: neigh
    Real(Kind=wp), Dimension(1:9),              Intent(InOut) :: stress
    Real(Kind=wp),                              Intent(  Out) :: vircpe, engcpe
    Real(Kind=wp), Dimension(1:neigh%max_list), Intent(In   ) :: rrt, zzt, yyt, xxt
    Type(mpole_type),                           Intent(InOut) :: mpoles
    Type(configuration_type),                   Intent(InOut) :: config

    Integer       :: idi, ii, jatm, jj, k1, k2, k3, ks1, ks11, ks2, ks21, ks3, ks31, m, s1, s2, s3
    Real(Kind=wp) :: engmpl, fix, fiy, fiz, fx, fy, fz, kx, ky, kz, scl, strs1, strs2, strs3, &
                     strs5, strs6, strs9, sx, sy, sz, t1, tix, tiy, tiz, tjx, tjy, tjz, tmp, tmpi, &
                     tmpj, txyz, jmpz(1:mpoles%max_mpoles), jmpy(1:mpoles%max_mpoles), &
                     jmpx(1:mpoles%max_mpoles), jmp(1:mpoles%max_mpoles), &
                     impz(1:mpoles%max_mpoles), impy(1:mpoles%max_mpoles), &
                     impx(1:mpoles%max_mpoles), imp(1:mpoles%max_mpoles), d1(-2:2 * mpoles% &
                     max_order + 1, -2:2 * mpoles%max_order + 1, -2:2 * mpoles%max_order + 1)

    ! initialise potential energy and virial

    engcpe = 0.0_wp
    vircpe = 0.0_wp

    ! initialise stress tensor accumulators

    strs1 = 0.0_wp
    strs2 = 0.0_wp
    strs3 = 0.0_wp
    strs5 = 0.0_wp
    strs6 = 0.0_wp
    strs9 = 0.0_wp

    ! global identity of iatm

    idi = config%ltg(iatm)

    ! get the multipoles for site i

    imp = mpoles%global_frame(:, iatm)

    ! ignore interaction if the charge is zero

    If (Maxval(Abs(imp)) > zero_plus) Then

      ! get the components for site i infinitesimal rotations

      impx = mpoles%rotation_x(:, iatm)
      impy = mpoles%rotation_y(:, iatm)
      impz = mpoles%rotation_z(:, iatm)

      ! multipole scaler

      scl = r4pie0 / eps

      ! scale imp multipoles

      imp = imp * scl

      ! load forces

      fix = config%parts(iatm)%fxx
      fiy = config%parts(iatm)%fyy
      fiz = config%parts(iatm)%fzz

      ! initialize torques for atom i (temporary)

      tix = 0.0_wp; tiy = 0.0_wp; tiz = 0.0_wp

      ! start of primary loop for forces evaluation

      Do m = 1, neigh%list(0, iatm)

        ! atomic index

        jatm = neigh%list(m, iatm)

        ! get the multipoles for site j

        jmp = mpoles%global_frame(:, jatm)

        ! truncation of potential - rrt(m) is the interatomic distance

        If (Maxval(Abs(jmp)) > zero_plus .and. rrt(m) < neigh%cutoff) Then

          ! get the components for site j infinitesimal rotations

          jmpx = mpoles%rotation_x(:, jatm)
          jmpy = mpoles%rotation_y(:, jatm)
          jmpz = mpoles%rotation_z(:, jatm)

          ! compute derivatives of kernel

          Call coul_deriv(2, 2 * mpoles%max_order + 1, xxt(m), yyt(m), zzt(m), rrt(m), d1)

          ! calculate forces

          engmpl = 0.0_wp
          fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp
          tjx = 0.0_wp; tjy = 0.0_wp; tjz = 0.0_wp

          If (mpoles%max_order < 5) Then

            kz = 1.0_wp
            Do k3 = 0, mpoles%max_order

              ky = kz
              Do k2 = 0, mpoles%max_order - k3

                kx = ky
                Do k1 = 0, mpoles%max_order - k3 - k2

                  jj = mpoles%map(k1, k2, k3)

                  If (Abs(jmp(jj)) > zero_plus) Then
                    Call explicit_ewald_real_loops &
                      (-2, 2 * mpoles%max_order + 1, k1, k2, k3, 1.0_wp, d1, &
                       imp, impx, impy, impz, tix, tiy, tiz, &
                       kx * jmp(jj), jmpx(jj), jmpy(jj), jmpz(jj), tjx, tjy, tjz, &
                       engmpl, fx, fy, fz, mpoles)
                  End If

                  kx = -kx

                End Do

                ky = -ky

              End Do

              kz = -kz

            End Do

          Else

            kz = 1.0_wp
            Do k3 = 0, mpoles%max_order

              ky = kz
              Do k2 = 0, mpoles%max_order - k3

                kx = ky
                Do k1 = 0, mpoles%max_order - k3 - k2

                  jj = mpoles%map(k1, k2, k3)

                  If (Abs(jmp(jj)) > zero_plus) Then

                    txyz = kx * jmp(jj)

                    sz = 1.0_wp
                    Do s3 = 0, mpoles%max_order
                      ks3 = k3 + s3; ks31 = ks3 + 1

                      sy = sz
                      Do s2 = 0, mpoles%max_order - s3
                        ks2 = k2 + s2; ks21 = ks2 + 1

                        sx = sy
                        Do s1 = 0, mpoles%max_order - s3 - s2
                          ks1 = k1 + s1; ks11 = ks1 + 1

                          ii = mpoles%map(s1, s2, s3)

                          tmp = d1(ks1, ks2, ks3)

                          tmpi = txyz * tmp
                          tmpj = sx * imp(ii) * tmp

                          t1 = txyz * imp(ii)

                          ! energy

                          engmpl = engmpl + t1 * tmp

                          ! force

                          fx = fx - t1 * d1(ks11, ks2, ks3)
                          fy = fy - t1 * d1(ks1, ks21, ks3)
                          fz = fz - t1 * d1(ks1, ks2, ks31)

                          ! torque on iatm

                          tix = tix + impx(ii) * tmpi
                          tiy = tiy + impy(ii) * tmpi
                          tiz = tiz + impz(ii) * tmpi

                          ! torque on jatm

                          tjx = tjx + jmpx(jj) * tmpj
                          tjy = tjy + jmpy(jj) * tmpj
                          tjz = tjz + jmpz(jj) * tmpj

                          sx = -sx
                        End Do

                        sy = -sy
                      End Do

                      sz = -sz
                    End Do

                  End If

                  kx = -kx

                End Do

                ky = -ky

              End Do

              kz = -kz

            End Do

          End If

          fix = fix + fx
          fiy = fiy + fy
          fiz = fiz + fz

#ifndef HALF_HALO
          If (jatm <= config%natms) Then
#endif /* HALF_HALO */

            config%parts(jatm)%fxx = config%parts(jatm)%fxx - fx
            config%parts(jatm)%fyy = config%parts(jatm)%fyy - fy
            config%parts(jatm)%fzz = config%parts(jatm)%fzz - fz

            mpoles%torque_x(jatm) = mpoles%torque_x(jatm) + tjx
            mpoles%torque_y(jatm) = mpoles%torque_y(jatm) + tjy
            mpoles%torque_z(jatm) = mpoles%torque_z(jatm) + tjz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

#ifndef HALF_HALO
          If (jatm <= config%natms .or. idi < config%ltg(jatm)) Then
#endif /* HALF_HALO */

            ! accumulate potential energy

            engcpe = engcpe + engmpl

            ! calculate stress tensor

            strs1 = strs1 + xxt(m) * fx
            strs2 = strs2 + xxt(m) * fy
            strs3 = strs3 + xxt(m) * fz
            strs5 = strs5 + yyt(m) * fy
            strs6 = strs6 + yyt(m) * fz
            strs9 = strs9 + zzt(m) * fz

#ifndef HALF_HALO
          End If
#endif /* HALF_HALO */

        End If

      End Do

      ! load back forces

      config%parts(iatm)%fxx = fix
      config%parts(iatm)%fyy = fiy
      config%parts(iatm)%fzz = fiz

      ! and torques due to multipoles

      mpoles%torque_x(iatm) = mpoles%torque_x(iatm) + scl * tix
      mpoles%torque_y(iatm) = mpoles%torque_y(iatm) + scl * tiy
      mpoles%torque_z(iatm) = mpoles%torque_z(iatm) + scl * tiz

      ! virial

      vircpe = -engcpe

      ! complete stress tensor

      stress(1) = stress(1) + strs1
      stress(2) = stress(2) + strs2
      stress(3) = stress(3) + strs3
      stress(4) = stress(4) + strs2
      stress(5) = stress(5) + strs5
      stress(6) = stress(6) + strs6
      stress(7) = stress(7) + strs3
      stress(8) = stress(8) + strs6
      stress(9) = stress(9) + strs9

    End If

  End Subroutine coul_dddp_mforces

  Subroutine coul_chrm_forces(iatm, eps, xxt, yyt, zzt, rrt, engcpe_ch, vircpe_ch, stress, neigh, mpoles, config)

    !!------------------------------------------------------------------------!
    !!
    !! dl_poly_4 subroutine for calculating coulombic energy and force terms
    !! for CHARMM model intra-core-shell interactions
    !!
    !! S(r)=1-(1+u/2).exp(-u) ; u=u(r)=r_ij.(a_i+a_j)/(p_i.p_j)^(1/6)
    !!
    !! S'(r)=(1+u).u'.exp(-u)/2 ; |r|'=r/|r|
    !!
    !! Uchrm(r_ij) =  S(r_ij)*U(r_ij) ; F(r_ij) = -U'(r_ij)
    !! Fchrm(r_ij) = -Uchrm'(r_ij) = S(r_ij)*F(r_ij) - S'(r_ij)*U(r_ij)
    !!
    !! copyright - daresbury laboratory
    !! author    - i.t.todorov february 2017
    !! contrib   - a.v.brukhno & m.a.seaton august 2020 - 'half-halo' VNL
    !! refactoring:
    !!           - a.m.elena march-october 2018
    !!           - j.madge march-october 2018
    !!           - a.b.g.chalk march-october 2018
    !!           - i.scivetti march-october 2018
    !!
    !!------------------------------------------------------------------------!

    Integer,                                    Intent(In   ) :: iatm
    Real(Kind=wp),                              Intent(In   ) :: eps
    Type(neighbours_type),                      Intent(In   ) :: neigh
    Real(Kind=wp), Dimension(1:9),              Intent(InOut) :: stress
    Real(Kind=wp),                              Intent(  Out) :: vircpe_ch, engcpe_ch
    Real(Kind=wp), Dimension(1:neigh%max_list), Intent(In   ) :: rrt, zzt, yyt, xxt
    Type(mpole_type),                           Intent(InOut) :: mpoles
    Type(configuration_type),                   Intent(InOut) :: config

    Integer       :: idi, jatm, limit, m
    Real(Kind=wp) :: chgea, chgprd, coul, dmpa, dmpsum, fcoul, fix, fiy, fiz, fx, fy, fz, plra, &
                     plrprd, r_r, rrr, scr, spr, strs1, strs2, strs3, strs5, strs6, strs9, tmp, u

    ! initialise potential energy and virial

    engcpe_ch = 0.0_wp
    vircpe_ch = 0.0_wp

    ! initialise stress tensor accumulators

    strs1 = 0.0_wp
    strs2 = 0.0_wp
    strs3 = 0.0_wp
    strs5 = 0.0_wp
    strs6 = 0.0_wp
    strs9 = 0.0_wp

    ! global identity of iatm

    idi = config%ltg(iatm)

    ! charge, inverse polarisability and dumping

    chgea = config%parts(iatm)%chge
    plra = mpoles%polarisation_atom(iatm)
    dmpa = mpoles%dump_atom(iatm)

    ! scale main charge

    chgea = chgea * r4pie0 / eps

    ! load forces

    fix = config%parts(iatm)%fxx
    fiy = config%parts(iatm)%fyy
    fiz = config%parts(iatm)%fzz

    ! start of primary loop for forces evaluation

    ! Get neigh%list limit

    limit = neigh%list(-3, iatm) - neigh%list(0, iatm)

    Do m = 1, limit

      ! interatomic distance and derivatives

      rrr = rrt(m)
      r_r = 1.0_wp / rrr

      ! atomic index, charge & inverse polarisability products
      ! and total inter-atomic summed dumping

      jatm = neigh%list(neigh%list(0, iatm) + m, iatm)
      chgprd = chgea * config%parts(jatm)%chge
      plrprd = plra * mpoles%polarisation_atom(jatm)
      dmpsum = dmpa + mpoles%dump_atom(jatm)

      u = (dmpsum * plrprd**6) * rrr ! dimensionless

      tmp = Exp(-u)
      scr = 1.0_wp - (1.0_wp + 0.5_wp * u) * tmp ! S(r)
      spr = (1.0_wp + u) * tmp * u ! S'(r).r

      ! calculate forces

      coul = scr * chgprd * r_r
      tmp = (scr - spr) * chgprd ! used later for the virial
      fcoul = tmp * r_r**3

      fx = fcoul * xxt(m)
      fy = fcoul * yyt(m)
      fz = fcoul * zzt(m)

      fix = fix + fx
      fiy = fiy + fy
      fiz = fiz + fz

#ifndef HALF_HALO
      If (jatm <= config%natms) Then
#endif /* HALF_HALO */

        config%parts(jatm)%fxx = config%parts(jatm)%fxx - fx
        config%parts(jatm)%fyy = config%parts(jatm)%fyy - fy
        config%parts(jatm)%fzz = config%parts(jatm)%fzz - fz

#ifndef HALF_HALO
      End If
#endif /* HALF_HALO */

#ifndef HALF_HALO
      If (jatm <= config%natms .or. idi < config%ltg(jatm)) Then
#endif /* HALF_HALO */

        ! calculate potential energy and virial

        engcpe_ch = engcpe_ch + coul
        vircpe_ch = vircpe_ch - tmp * r_r

        ! calculate stress tensor

        strs1 = strs1 + xxt(m) * fx
        strs2 = strs2 + xxt(m) * fy
        strs3 = strs3 + xxt(m) * fz
        strs5 = strs5 + yyt(m) * fy
        strs6 = strs6 + yyt(m) * fz
        strs9 = strs9 + zzt(m) * fz

#ifndef HALF_HALO
      End If
#endif /* HALF_HALO */

    End Do

    ! load back forces

    config%parts(iatm)%fxx = fix
    config%parts(iatm)%fyy = fiy
    config%parts(iatm)%fzz = fiz

    ! complete stress tensor

    stress(1) = stress(1) + strs1
    stress(2) = stress(2) + strs2
    stress(3) = stress(3) + strs3
    stress(4) = stress(4) + strs2
    stress(5) = stress(5) + strs5
    stress(6) = stress(6) + strs6
    stress(7) = stress(7) + strs3
    stress(8) = stress(8) + strs6
    stress(9) = stress(9) + strs9

  End Subroutine coul_chrm_forces

  Subroutine d_ene_trq_mpoles(vircpe_dt, stress, mpoles, config)

    !!------------------------------------------------------------------------!
    !!
    !! dl_poly_4 subroutine for calculating the change in energy produced by
    !! an infinitesimal rotation of multipoles
    !!
    !! Reference : Sagui, Pedersen, Darden, J. Chem. Phys. 120, 73 (2004)
    !!             doi: 10.1063/1.1630791
    !!
    !! copyright - daresbury laboratory
    !! author    - h.a.boateng april 2015
    !! amended   - i.t.todorov february 2016
    !! contrib   - a.v.brukhno & m.a.seaton august 2020 - 'half-halo' VNL
    !! refactoring:
    !!           - a.m.elena march-october 2018
    !!           - j.madge march-october 2018
    !!           - a.b.g.chalk march-october 2018
    !!           - i.scivetti march-october 2018
    !!
    !!------------------------------------------------------------------------!

    Real(Kind=wp),                 Intent(  Out) :: vircpe_dt
    Real(Kind=wp), Dimension(1:9), Intent(InOut) :: stress
    Type(mpole_type),              Intent(InOut) :: mpoles
    Type(configuration_type),      Intent(InOut) :: config

    Integer       :: iatm, idi, j, jatm
    Real(Kind=wp) :: dedux, deduy, deduz, dedv, dedw, dux_dx, dux_dy, dux_dz, duy_dy, duy_dz, &
                     duz_dz, fix, fiy, fiz, fx, fy, fz, magp1, magp1perp2, magp2, magp2perp1, &
                     p1(1:3), p1perp2(1:3), p2(1:3), p2perp1(1:3), p2u(1:3), rmag, rmag3, rrr, &
                     rsq, strs1, strs2, strs3, strs4, strs5, strs6, strs7, strs8, strs9, stx(1:2), &
                     sty(1:2), stz(1:2), tmptx, tmpty, tmptz, tx, ty, tz, u(1:3), v(1:3), w(1:3), &
                     xdf, ydf, zdf

    ! initialise virial

    vircpe_dt = 0.0_wp

    ! initialise stress tensor accumulators

    strs1 = 0.0_wp
    strs2 = 0.0_wp
    strs3 = 0.0_wp
    strs4 = 0.0_wp
    strs5 = 0.0_wp
    strs6 = 0.0_wp
    strs7 = 0.0_wp
    strs8 = 0.0_wp
    strs9 = 0.0_wp

    Do iatm = 1, config%natms

      ! load forces

      fix = config%parts(iatm)%fxx
      fiy = config%parts(iatm)%fyy
      fiz = config%parts(iatm)%fzz

      ! global identity of iatm

      idi = config%ltg(iatm)

      If (mpoles%rotation(iatm)%flag == 1) Then

        ! p1 and p2 define the local frame

        p1 = mpoles%rotation(iatm)%p1
        p2 = mpoles%rotation(iatm)%p2

        magp1 = Sqrt(p1(1) * p1(1) + p1(2) * p1(2) + p1(3) * p1(3))
        magp2 = Sqrt(p2(1) * p2(1) + p2(2) * p2(2) + p2(3) * p2(3))

        p2u = p2 / magp2

        ! standard basis for coordinate system

        u(1) = mpoles%rotation(iatm)%mtrxa(1)
        u(2) = mpoles%rotation(iatm)%mtrxa(4)
        u(3) = mpoles%rotation(iatm)%mtrxa(7)

        v(1) = mpoles%rotation(iatm)%mtrxa(2)
        v(2) = mpoles%rotation(iatm)%mtrxa(5)
        v(3) = mpoles%rotation(iatm)%mtrxa(8)

        w(1) = mpoles%rotation(iatm)%mtrxa(3)
        w(2) = mpoles%rotation(iatm)%mtrxa(6)
        w(3) = mpoles%rotation(iatm)%mtrxa(9)

        ! change in energy (E) due to infinitesimal rotation (torque => \tau) dE_{\omega} = -\tau * d\omega
        ! we omit the negative here and introduce it in the final force computation, i.e., because force is
        ! the negative of the change in energy with respect to energy, we'll add the magnitude of the force
        ! computed instead of subtracting the magnitude

        tmptx = mpoles%torque_x(iatm)
        tmpty = mpoles%torque_y(iatm)
        tmptz = mpoles%torque_z(iatm)

        tx = tmptx * u(1) + tmpty * u(2) + tmptz * u(3)
        ty = (tmptx * p2(1) + tmpty * p2(2) + tmptz * p2(3)) / magp2
        tz = tmptx * w(1) + tmpty * w(2) + tmptz * w(3)

        ! component of p2 perpendicular to p1

        p2perp1 = p2 - (u(1) * p2(1) + u(2) * p2(2) + u(3) * p2(3)) * u
        magp2perp1 = Sqrt(p2perp1(1) * p2perp1(1) + p2perp1(2) * p2perp1(2) + p2perp1(3) * p2perp1(3))

        ! component of p1 perpendicular to p2

        p1perp2 = p1 - (p2u(1) * p1(1) + p2u(2) * p1(2) + p2u(3) * p1(3)) * p2u
        magp1perp2 = Sqrt(p1perp2(1) * p1perp2(1) + p1perp2(2) * p1perp2(2) + p1perp2(3) * p1perp2(3))

        ! For p1

        ! dedu = 0.0 since movement du along the u axis does not rotate the frame

        dedv = tz / magp1
        dedw = -ty / magp1perp2

        stx(1) = dedv * v(1) + dedw * w(1)
        sty(1) = dedv * v(2) + dedw * w(2)
        stz(1) = dedv * v(3) + dedw * w(3)

        ! for p2

        dedw = tx / magp2perp1

        stx(2) = dedw * w(1)
        sty(2) = dedw * w(2)
        stz(2) = dedw * w(3)

        ! now compute forces and virial

        fx = 0.0_wp; fy = 0.0_wp; fz = 0.0_wp

        Do j = 1, 2

          jatm = mpoles%rotation(iatm)%mbnd(j)

          If (jatm > 0) Then
            xdf = config%parts(jatm)%xxx - config%parts(iatm)%xxx
            ydf = config%parts(jatm)%yyy - config%parts(iatm)%yyy
            zdf = config%parts(jatm)%zzz - config%parts(iatm)%zzz

            Call images_s(config%imcon, config%cell, xdf, ydf, zdf)

            rsq = xdf * xdf + ydf * ydf + zdf * zdf
            rrr = Sqrt(rsq)
            rmag = 1.0_wp / rrr
            rmag3 = 1.0_wp / (rsq * rrr)

            dux_dx = rmag - xdf * xdf * rmag3
            dux_dy = -xdf * ydf * rmag3 ! duy_dx = dux_dy
            dux_dz = -xdf * zdf * rmag3 ! duz_dx = dux_dz
            duy_dy = rmag - ydf * ydf * rmag3
            duy_dz = -ydf * zdf * rmag3 ! duz_dy = duy_dz
            duz_dz = rmag - zdf * zdf * rmag3

            dedux = stx(j)
            deduy = sty(j)
            deduz = stz(j)

            ! Now to find the forces (derivatives of energy with respect to cartesian positions),
            ! i.e. fx=dedx, fy=dedy, fz=dedz

            fx = dedux * dux_dx + deduy * dux_dy + deduz * dux_dz
            fy = dedux * dux_dy + deduy * duy_dy + deduz * duy_dz
            fz = dedux * dux_dz + deduy * duy_dz + deduz * duz_dz

            fix = fix + fx
            fiy = fiy + fy
            fiz = fiz + fz

#ifndef HALF_HALO
            If (jatm <= config%natms) Then
#endif /* HALF_HALO */

              config%parts(jatm)%fxx = config%parts(jatm)%fxx - fx
              config%parts(jatm)%fyy = config%parts(jatm)%fyy - fy
              config%parts(jatm)%fzz = config%parts(jatm)%fzz - fz

#ifndef HALF_HALO
            End If
#endif /* HALF_HALO */

#ifndef HALF_HALO
            If (jatm <= config%natms .or. idi < config%ltg(jatm)) Then
#endif /* HALF_HALO */

              ! calculate virial

              vircpe_dt = vircpe_dt - (fx * xdf + fy * ydf + fz * zdf)

              ! calculate stress tensor

              strs1 = strs1 + xdf * fx
              strs2 = strs2 + xdf * fy
              strs3 = strs3 + xdf * fz
              strs4 = strs4 + ydf * fx
              strs5 = strs5 + ydf * fy
              strs6 = strs6 + ydf * fz
              strs7 = strs7 + zdf * fx
              strs8 = strs8 + zdf * fy
              strs9 = strs9 + zdf * fz

#ifndef HALF_HALO
            End If
#endif /* HALF_HALO */

          End If

        End Do

        ! load back forces

        config%parts(iatm)%fxx = fix
        config%parts(iatm)%fyy = fiy
        config%parts(iatm)%fzz = fiz

      End If

      ! complete stress tensor (and symmetrise)

      stress(1) = stress(1) + strs1
      stress(2) = stress(2) + 0.5_wp * (strs2 + strs4)
      stress(3) = stress(3) + 0.5_wp * (strs3 + strs7)
      stress(4) = stress(4) + 0.5_wp * (strs2 + strs4)
      stress(5) = stress(5) + strs5
      stress(6) = stress(6) + 0.5_wp * (strs6 + strs8)
      stress(7) = stress(7) + 0.5_wp * (strs3 + strs7)
      stress(8) = stress(8) + 0.5_wp * (strs6 + strs8)
      stress(9) = stress(9) + strs9

    End Do

  End Subroutine d_ene_trq_mpoles

End Module coul_mpole
