Module ttm

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !
  ! dl_poly_4 module for defining arrays and initial parameters for
  ! two-temperature model(ttm)
  !
  ! copyright - daresbury laboratory
  ! authors   - s.l.daraszewicz & m.a.seaton may 2012
  ! contrib   - g.khara may 2016
  ! contrib   - m.a.seaton september 2017
  ! contrib   - m.a.seaton february 2020
  ! refactoring:
  !           - a.m.elena march-october 2018
  !           - j.madge march-october 2018
  !           - a.b.g.chalk march-october 2018
  !           - i.scivetti march-october 2018
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  Use comms,           Only: comms_type,&
                             gcheck,&
                             gmax,&
                             gmin,&
                             grid1_tag,&
                             grid2_tag,&
                             gsum,&
                             gsync,&
                             wp_mpi
  Use configuration,   Only: configuration_type
  Use constants,       Only: boltz,&
                             eu_ev,&
                             pi,&
                             rt2,&
                             sqrpi,&
                             tenunt,&
                             zero_plus,&
                             JKms_to_kBAps,&
                             Jm3K_to_kBA3
  Use domains,         Only: domains_type,&
                             idcube
  Use errors_warnings, Only: error,&
                             info,&
                             warning
  Use flow_control,    Only: RESTART_KEY_CLEAN,&
                             RESTART_KEY_OLD
  Use kinds,           Only: wp
  Use parse,           Only: get_line,&
                             get_word,&
                             word_2_real
  Use numerics,        Only: invert

#ifdef SERIAL
  Use mpi_api
#else
  Use mpi
#endif

  Implicit None
  Private

  Type, Public :: ttm_type

    Real(Kind=wp), Allocatable :: eltemp(:, :, :, :), eltemp_adj(:, :, :, :)
    Real(Kind=wp), Allocatable :: act_ele_cell(:, :, :, :), old_ele_cell(:, :, :, :)
    Real(Kind=wp), Allocatable :: asource(:)
    Real(Kind=wp), Allocatable :: tempion(:)
    Real(Kind=wp), Allocatable :: ttmvom(:, :)
    Real(Kind=wp), Allocatable :: gsource(:)
    Logical, Allocatable :: adjust(:, :, :, :)

    Real(Kind=wp), Allocatable :: cetable(:, :)
    Real(Kind=wp), Allocatable :: gtable(:, :)
    Real(Kind=wp), Allocatable :: detable(:, :)
    Real(Kind=wp), Allocatable :: ketable(:, :)

    !> Ionic temperature grid size
    Integer :: ntsys(3) = [0, 0, 10]
    !> Electronic temperature grid size
    Integer :: eltsys(3) = 50
    Integer :: ntcell(3)
    Integer :: eltcell(3)
    Integer :: ntcelloff(3)
    Integer :: midI(3), midE(3)
    Integer :: zeroE(3)

    Integer :: tmpmsgx, tmpmsgy, tmpmsgz
    Integer :: nummsgx, nummsgy, nummsgz

    !> Temperature grid size
    Real(Kind=wp) :: delx, dely, delz
    !> Temperature reciprocal grid size
    Real(Kind=wp) :: delu, delv, delw
    !> Volume, recip volume
    Real(Kind=wp) :: volume, rvolume
    Real(Kind=wp) :: zerocell(3)
    Real(Kind=wp) :: grcell(9)
    !> Number of TTM cells
    Integer :: numcell
    Integer :: ttmbc(6)
    Integer :: ttmbcmap(6)

    !> TTM Active
    Logical :: l_ttm = .false.
    !> Modelling metal
    Logical :: isMetal = .false.
    !> Electron-phonon coupling
    Logical :: l_epcp = .false.
    !> Redistribute energy between cells
    Logical :: redistribute
    !> Remove CoM motion
    Logical :: ttmthvel = .true.
    !> Remove CoM motion in Z only
    Logical :: ttmthvelz = .false.
    !> One way e-p coupling
    Logical :: oneway = .false.
    !> Density calculated dynamically
    Logical :: ttmdyndens = .false.
    Logical :: findepo
    !> Heat capacity type
    Integer :: CeType = 0
    !> Condictivity type
    Integer :: KeType = 0
    !> Diffusivity type
    Integer :: DeType = 0
    !> E-P coupling type
    Integer :: gvar = 0
    !> Boundary conditions type
    Integer :: bcTypeE = 3
    !> Frequency of printing STATS file
    Integer :: ttmstats = 0
    !> Frequency of printing TRAJ file
    Integer :: ttmtraj = 0
    !> Temporal deposition type
    Integer :: tdepoType = 1
    !> Spatial deposition type
    Integer :: sdepoType = 0
    !> Permitted flux for Robin boundaries
    Real(Kind=wp) :: fluxout = 0.95_wp
    !> Time to start E-P Coupling
    Real(Kind=wp) :: ttmoffset = 0.0_wp
    Real(Kind=wp) :: depostart
    Real(Kind=wp) :: depoend
    !> Specific heat parameters
    Real(Kind=wp) :: sh_A = 0.0_wp, sh_B = 0.0_wp
    !> Thermal conductivity
    Real(Kind=wp) :: Ka0 = 0.0_wp
    !> Heat capacity
    Real(Kind=wp) :: Ce0 = 1.0_wp
    !> Heat capacity max
    Real(Kind=wp) :: Cemax = 0.0_wp
    !> Fermi temperature
    Real(Kind=wp) :: Tfermi = 0.0_wp
    !> Diffusivity
    Real(Kind=wp) :: Diff0 = 0.0_wp
    !> Energy
    Real(Kind=wp) :: dEdX = 0.0_wp
    !> Spatial distribution parameter
    Real(Kind=wp) :: sig = 1.0_wp
    !> Spatial distribution cutoff
    Real(Kind=wp) :: sigmax = 5.0_wp
    !> Temporal deposition parameter
    Real(Kind=wp) :: tdepo = 1.0e-3_wp
    !> Temporal deposition cutoff
    Real(Kind=wp) :: tcdepo = 5.0_wp

    ! DEBUG (TODO)
    Real(Kind=wp) :: epstart
    Integer :: keyres0 = RESTART_KEY_CLEAN
    Integer :: nstepcpl = 0

    Integer :: cel, gel, del, kel
    Integer :: acell
    Integer :: acell_old
    !> Minimum number of atoms for ionic cells
    Integer :: amin = 1
    !> Unit conversions
    Real(Kind=wp) :: Jm3K_to_kBA3 = 1.0e-7_wp / (boltz * tenunt) ! convert J m^-3 K^-1 to kB A^-3
    Real(Kind=wp) :: JKms_to_kBAps = 10.0_wp / (boltz * tenunt) ! convert W m^-1 K^-1 to kB A^-1 ps^-1
    Real(Kind=wp) :: kB_to_eV = boltz / eu_ev ! convert kB to eV
    Real(Kind=wp) :: eV_to_kB = eu_ev / boltz ! convert eV to kB
    Real(Kind=wp) :: mJcm2_to_eVA2 = 1.0e4_wp / (eu_ev * tenunt) ! convert mJ cm^-2 to eV A^-2
    Real(Kind=wp) :: epc_to_chi
    !> Cell density
    Real(Kind=wp) :: cellrho
    !> Reciprocal cell density
    Real(Kind=wp) :: rcellrho
    !> Number density
    Real(Kind=wp) :: sysrho
    !> Laser deposition fluence
    Real(Kind=wp) :: fluence = 0.0_wp
    !> Laser deposition penetration depth
    Real(Kind=wp) :: pdepth = 0.0_wp
    Real(Kind=wp) :: epthreshold = 1.1_wp
    Real(Kind=wp) :: tdiffw(6)

    Real(Kind=wp), Allocatable, Dimension(:, :, :) :: lat_U, lat_B, lat_I
    Real(Kind=wp) :: norm

    Logical :: trackInit = .false.
  End Type ttm_type

  Integer, Parameter, Public :: TTM_CE_CONST=0, TTM_CE_TANH=1, &
       TTM_CE_LINEAR=2, TTM_CE_TABULATED=3, &
       TTM_CE_CONST_DYN=4, TTM_CE_TANH_DYN=5, &
       TTM_CE_LINEAR_DYN=6, TTM_CE_TABULATED_DYN=7
  Integer, Parameter, Public :: TTM_KE_INFINITE=0, TTM_KE_CONST=1, &
       TTM_KE_DRUDE=2, TTM_KE_TABULATED=3
  Integer, Parameter, Public :: TTM_DE_METAL=0, TTM_DE_CONST=1, TTM_DE_RECIP=2, &
       TTM_DE_TABULATED=3
  Integer, Parameter, Public :: TTM_EPVAR_NULL=0, TTM_EPVAR_HOMO=1, TTM_EPVAR_HETERO=2
  Integer, Parameter, Public :: TTM_SDEPO_NULL=0, TTM_SDEPO_GAUSS=1, TTM_SDEPO_FLAT=2, TTM_SDEPO_EXP=3
  Integer, Parameter, Public :: TTM_TDEPO_GAUSS=1, TTM_TDEPO_EXP=2, TTM_TDEPO_DELTA=3, TTM_TDEPO_PULSE=4
  Integer, Parameter, Public :: TTM_BC_PERIODIC=1, TTM_BC_DIRICHLET=2, TTM_BC_NEUMANN=3, &
       TTM_BC_DIRICHLET_XY=4, TTM_BC_ROBIN=5, TTM_BC_ROBIN_XY=6

  Public :: allocate_ttm_arrays, deallocate_ttm_arrays
  Public :: eltemp_min, eltemp_max, &
            eltemp_maxKe, eltemp_minKe, eltemp_mean, eltemp_sum
  Public :: ttm_system_init, ttm_system_revive, ttm_table_read, &
            ttm_table_scan, boundaryHalo, boundaryCond, depoinit
  Public :: ttm_setup_bounds
Contains

  Subroutine ttm_setup_bounds(ttm, config, domain, megatm, padding_margin)

    !-----------------------------------------------------------------------
    !
    ! dl_poly_4 subroutine to finalise ttm for set_bounds
    !
    ! copyright - daresbury laboratory
    !
    !-----------------------------------------------------------------------

    Type(ttm_type), Intent( InOut ) :: ttm
    Type(configuration_type), Intent( In    ) :: config
    Type(domains_type), Intent( In    ) :: domain
    Integer, Intent( In    ) :: megatm
    Real(kind=wp),  Intent(   Out ) :: padding_margin
    Real(kind=wp) :: tol, test
    Integer :: i

    padding_margin = 0.0_wp
    ! two-temperature model: determine number of CITs
    ! in x- and y-directions based on number in z-direction
    ! and system size

    ttm%delz     = config%cell(9) / Real(ttm%ntsys(3), wp)
    ttm%ntsys(1) = Nint(config%cell(1) / ttm%delz)
    ttm%ntsys(2) = Nint(config%cell(5) / ttm%delz)
    ttm%delx     = config%cell(1) / Real(ttm%ntsys(1), wp)
    ttm%dely     = config%cell(5) / Real(ttm%ntsys(2), wp)
    ttm%volume   = ttm%delx * ttm%dely * ttm%delz
    ttm%rvolume  = 1.0_wp / ttm%volume

    ! Check number of electronic temperature cells is
    ! >= to number of ionic temperature cells

    If (Any(ttm%eltsys < ttm%ntsys)) Call error(670)

    ! Check rpad does not go too far for determining ionic temperatures
    tol = Min(ttm%delx, ttm%dely, ttm%delz)
    Do i = 1, domain%nx - 1
      test = Real(i, wp) * config%cell(1) * domain%nx_recip
      test = test - ttm%delx * Floor(test / ttm%delx)
      If (test > zero_plus) tol = Min(tol, test)
    End Do
    Do i = 1, domain%ny - 1
      test = Real(i, wp) * config%cell(5) * domain%ny_recip
      test = test - ttm%dely * Floor(test / ttm%dely)
      If (test > zero_plus) tol = Min(tol, test)
    End Do
    Do i = 1, domain%nz - 1
      test = Real(i, wp) * config%cell(9) * domain%nz_recip
      test = test - ttm%delz * Floor(test / ttm%delz)
      If (test > zero_plus) tol = Min(tol, test)
    End Do
    padding_margin = tol

    ! If ttm%redistribute option selected, check for sufficient electronic temperature
    ! cells to redistribute energy when ionic temperature cells are switched off:
    ! if not available, switch off this option

    If (ttm%redistribute .and. (ttm%eltsys(1) < ttm%ntsys(1) + 2 .or. &
         ttm%eltsys(2) < ttm%ntsys(2) + 2 .or. &
         ttm%eltsys(3) < ttm%ntsys(3) + 2)) Then
      Call warning(500, 0.0_wp, 0.0_wp, 0.0_wp)
      ttm%redistribute = .false.
    End If

    ! Calculate average atomic density: if not overridden by
    ! 'ttm atomdens' directive in CONTROL file, will be used
    ! to convert specific heat capacities to ttm%volumetric
    ! heat capacity etc.

    ttm%sysrho = Real(megatm, Kind=wp) / Product(config%cell(1:9:4))

  end Subroutine ttm_setup_bounds

  Subroutine allocate_ttm_arrays(ttm, domain, config, comm)

    Type(ttm_type),           Intent(InOut) :: ttm
    Type(domains_type),       Intent(In   ) :: domain
    Type(configuration_type), Intent(InOut) :: config
    Type(comms_type),         Intent(In   ) :: comm

    Integer                        :: basicslice, bbasicslice, boneslicex, boneslicey, boneslicez, &
                                      i, ierr, numbc, numbcmap, oneslicex, oneslicey, oneslicez
    Integer(Kind=MPI_ADDRESS_KIND) :: bxlth, bylth, dbleth, inleth, lb1, lb2, xlth, ylth
    Integer, Dimension(1:7)        :: fail
    Real(Kind=wp)                  :: finish, start, rcell(1:9), det, lx, ly, lz, lxx, lyy, lzz
    Real(Kind=wp)                  :: lxlxx, lylyy, lzlzz

    fail = 0

    ttm%depostart = 0.0_wp
    ttm%depoend = 0.0_wp

    ! Setup constants based on fundamental values (found in
    ! setup.f90)

    If (ttm%l_ttm) Then

      ! Calculate grid inversion parameters for ion temperature
      ! determination

      Call invert(config%cell, rcell, det)

      ttm%grcell(1) = rcell(1) * Real(ttm%ntsys(1), wp)
      ttm%grcell(4) = rcell(4) * Real(ttm%ntsys(1), wp)
      ttm%grcell(7) = rcell(7) * Real(ttm%ntsys(1), wp)
      ttm%grcell(2) = rcell(2) * Real(ttm%ntsys(2), wp)
      ttm%grcell(5) = rcell(5) * Real(ttm%ntsys(2), wp)
      ttm%grcell(8) = rcell(8) * Real(ttm%ntsys(2), wp)
      ttm%grcell(3) = rcell(3) * Real(ttm%ntsys(3), wp)
      ttm%grcell(6) = rcell(6) * Real(ttm%ntsys(3), wp)
      ttm%grcell(9) = rcell(9) * Real(ttm%ntsys(3), wp)

      ! Work out weighting coefficients for thermal diffusion equation
      ! (mainly needed for parallelepiped systems)

      If (config%imcon==3) Then
        lxx = config%cell(1)*config%cell(1) + config%cell(2)*config%cell(2) + config%cell(3)*config%cell(3)
        lyy = config%cell(4)*config%cell(4) + config%cell(5)*config%cell(5) + config%cell(6)*config%cell(6)
        lzz = config%cell(7)*config%cell(7) + config%cell(8)*config%cell(8) + config%cell(9)*config%cell(9)
        lx = Sqrt(lxx)
        ly = Sqrt(lyy)
        lz = Sqrt(lzz)
        lxlxx = lxx/config%cell(1)/config%cell(1)
        lylyy = lyy/config%cell(5)/config%cell(5)
        lzlzz = lzz/config%cell(9)/config%cell(9)
        ttm%tdiffw(1) = lxlxx + lylyy * config%cell(2)*config%cell(2)/lxx + lzlzz * config%cell(3)*config%cell(3)/lxx
        ttm%tdiffw(2) = lylyy + lxlxx * config%cell(4)*config%cell(4)/lyy + lzlzz * config%cell(6)*config%cell(6)/lyy
        ttm%tdiffw(3) = lzlzz + lxlxx * config%cell(7)*config%cell(7)/lzz + lylyy * config%cell(8)*config%cell(8)/lzz
        ttm%tdiffw(4) = lzlzz * config%cell(3)*config%cell(6)/lx/ly - lxlxx * config%cell(4)/ly - lylyy * config%cell(2)/lx
        ttm%tdiffw(5) = lylyy * config%cell(2)*config%cell(8)/lx/lz - lxlxx * config%cell(7)/lz - lzlzz * config%cell(3)/lx
        ttm%tdiffw(6) = lxlxx * config%cell(4)*config%cell(7)/ly/lz - lylyy * config%cell(8)/lz - lzlzz * config%cell(6)/ly
        ttm%delu = (ttm%delx*config%cell(1) + ttm%dely*config%cell(2) + ttm%delz*config%cell(3))/lx
        ttm%delv = (ttm%delx*config%cell(4) + ttm%dely*config%cell(5) + ttm%delz*config%cell(6))/ly
        ttm%delw = (ttm%delx*config%cell(7) + ttm%dely*config%cell(8) + ttm%delz*config%cell(9))/lz
      Else
        ttm%tdiffw(1) = 1.0_wp
        ttm%tdiffw(2) = 1.0_wp
        ttm%tdiffw(3) = 1.0_wp
        ttm%tdiffw(4) = 0.0_wp
        ttm%tdiffw(5) = 0.0_wp
        ttm%tdiffw(6) = 0.0_wp
        ttm%delu = ttm%delx
        ttm%delv = ttm%dely
        ttm%delw = ttm%delz
      End If

      ! Determine number of ion temperature ttm%cells for domain and
      ! offsets for ion temperature determination

      start = config%cell(1) * Real(domain%idx, wp) * domain%nx_recip
      finish = config%cell(1) * Real(domain%idx + 1, wp) * domain%nx_recip
      ttm%ntcell(1) = Ceiling(finish / ttm%delx) - Ceiling(start / ttm%delx)
      ttm%ntcelloff(1) = Ceiling(start / ttm%delx)
      ttm%zerocell(1) = 0.5_wp * Real(ttm%ntsys(1), wp) - Real(Ceiling(start / ttm%delx), wp)

      start = config%cell(5) * Real(domain%idy, wp) * domain%ny_recip
      finish = config%cell(5) * Real(domain%idy + 1, wp) * domain%ny_recip
      ttm%ntcell(2) = Ceiling(finish / ttm%dely) - Ceiling(start / ttm%dely)
      ttm%ntcelloff(2) = Ceiling(start / ttm%dely)
      ttm%zerocell(2) = 0.5_wp * Real(ttm%ntsys(2), wp) - Real(Ceiling(start / ttm%dely), wp)

      start = config%cell(9) * Real(domain%idz, wp) * domain%nz_recip
      finish = config%cell(9) * Real(domain%idz + 1, wp) * domain%nz_recip
      ttm%ntcell(3) = Ceiling(finish / ttm%delz) - Ceiling(start / ttm%delz)
      ttm%ntcelloff(3) = Ceiling(start / ttm%delz)
      ttm%zerocell(3) = 0.5_wp * Real(ttm%ntsys(3), wp) - Real(Ceiling(start / ttm%delz), wp)

      ttm%numcell = (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2) * (ttm%ntcell(3) + 2)

      ! Determine mid-values for ion and electronic temperature grid

      ttm%midI(:) = Int((ttm%ntsys(:) + 1) / 2)
      ttm%midE(:) = Int((ttm%eltsys(:) + 1) / 2)

      ! Determine number of multiple ion temperature grids for electronic
      ! temperature grids

      ttm%eltcell(1) = Ceiling(Real(ttm%eltsys(1) - ttm%ntsys(1), Kind=wp) / Real(2 * ttm%ntsys(1), Kind=wp))
      ttm%eltcell(2) = Ceiling(Real(ttm%eltsys(2) - ttm%ntsys(2), Kind=wp) / Real(2 * ttm%ntsys(2), Kind=wp))
      ttm%eltcell(3) = Ceiling(Real(ttm%eltsys(3) - ttm%ntsys(3), Kind=wp) / Real(2 * ttm%ntsys(3), Kind=wp))

      ! Determine positions of boundaries for electronic temperature grids

      !   -x boundary
      numbc = -(ttm%eltsys(1) - ttm%ntsys(1)) / 2
      numbc = Mod(numbc + ttm%ntsys(1) * (ttm%eltcell(1) + 1), ttm%ntsys(1)) + 1
      ttm%zeroE(1) = numbc - 1
      numbcmap = (ttm%eltsys(1) - ttm%ntsys(1)) / 2
      numbcmap = Mod(numbcmap + ttm%ntsys(1) * (ttm%eltcell(1) + 1) - 1, ttm%ntsys(1)) + 1
      If (numbc > ttm%ntcelloff(1) .and. numbc <= (ttm%ntcelloff(1) + ttm%ntcell(1))) Then
        ttm%ttmbc(1) = numbc - ttm%ntcelloff(1)
        Do i = 0, domain%nx - 1
          start = config%cell(1) * Real(i, wp) * domain%nx_recip
          finish = config%cell(1) * Real(i + 1, wp) * domain%nx_recip
          If (numbcmap > Ceiling(start / ttm%delx) .and. numbcmap <= Ceiling(finish / ttm%delx)) Then
            ttm%ttmbcmap(1) = idcube(i, domain%idy, domain%idz, domain)
          End If
        End Do
      Else
        ttm%ttmbc(1) = 0
        ttm%ttmbcmap(1) = -1
      End If

      !   +x boundary
      numbc = (ttm%eltsys(1) - ttm%ntsys(1)) / 2
      numbc = Mod(numbc + ttm%ntsys(1) * (ttm%eltcell(1) + 1) - 1, ttm%ntsys(1)) + 1
      numbcmap = -(ttm%eltsys(1) - ttm%ntsys(1)) / 2
      numbcmap = Mod(numbcmap + ttm%ntsys(1) * (ttm%eltcell(1) + 1), ttm%ntsys(1)) + 1
      If (numbc > ttm%ntcelloff(1) .and. numbc <= (ttm%ntcelloff(1) + ttm%ntcell(1))) Then
        ttm%ttmbc(2) = numbc - ttm%ntcelloff(1)
        Do i = 0, domain%nx - 1
          start = config%cell(1) * Real(i, wp) * domain%nx_recip
          finish = config%cell(1) * Real(i + 1, wp) * domain%nx_recip
          If (numbcmap > Ceiling(start / ttm%delx) .and. numbcmap <= Ceiling(finish / ttm%delx)) Then
            ttm%ttmbcmap(2) = idcube(i, domain%idy, domain%idz, domain)
          End If
        End Do
      Else
        ttm%ttmbc(2) = 0
        ttm%ttmbcmap(2) = -1
      End If

      !   -y boundary
      numbc = -(ttm%eltsys(2) - ttm%ntsys(2)) / 2
      numbc = Mod(numbc + ttm%ntsys(2) * (ttm%eltcell(2) + 1), ttm%ntsys(2)) + 1
      ttm%zeroE(2) = numbc - 1
      numbcmap = (ttm%eltsys(2) - ttm%ntsys(2)) / 2
      numbcmap = Mod(numbcmap + ttm%ntsys(2) * (ttm%eltcell(2) + 1) - 1, ttm%ntsys(2)) + 1
      If (numbc > ttm%ntcelloff(2) .and. numbc <= (ttm%ntcelloff(2) + ttm%ntcell(2))) Then
        ttm%ttmbc(3) = numbc - ttm%ntcelloff(2)
        Do i = 0, domain%ny - 1
          start = config%cell(5) * Real(i, wp) * domain%ny_recip
          finish = config%cell(5) * Real(i + 1, wp) * domain%ny_recip
          If (numbcmap > Ceiling(start / ttm%dely) .and. numbcmap <= Ceiling(finish / ttm%dely)) Then
            ttm%ttmbcmap(3) = idcube(domain%idx, i, domain%idz, domain)
          End If
        End Do
      Else
        ttm%ttmbc(3) = 0
        ttm%ttmbcmap(3) = -1
      End If

      !   +y boundary
      numbc = (ttm%eltsys(2) - ttm%ntsys(2)) / 2
      numbc = Mod(numbc + ttm%ntsys(2) * (ttm%eltcell(2) + 1) - 1, ttm%ntsys(2)) + 1
      numbcmap = -(ttm%eltsys(2) - ttm%ntsys(2)) / 2
      numbcmap = Mod(numbcmap + ttm%ntsys(2) * (ttm%eltcell(2) + 1), ttm%ntsys(2)) + 1
      If (numbc > ttm%ntcelloff(2) .and. numbc <= (ttm%ntcelloff(2) + ttm%ntcell(2))) Then
        ttm%ttmbc(4) = numbc - ttm%ntcelloff(2)
        Do i = 0, domain%ny - 1
          start = config%cell(5) * Real(i, wp) * domain%ny_recip
          finish = config%cell(5) * Real(i + 1, wp) * domain%ny_recip
          If (numbcmap > Ceiling(start / ttm%dely) .and. numbcmap <= Ceiling(finish / ttm%dely)) Then
            ttm%ttmbcmap(4) = idcube(domain%idx, i, domain%idz, domain)
          End If
        End Do
      Else
        ttm%ttmbc(4) = 0
        ttm%ttmbcmap(4) = -1
      End If

      !   -z boundary
      numbc = -(ttm%eltsys(3) - ttm%ntsys(3)) / 2
      numbc = Mod(numbc + ttm%ntsys(3) * (ttm%eltcell(3) + 1), ttm%ntsys(3)) + 1
      ttm%zeroE(3) = numbc - 1
      numbcmap = (ttm%eltsys(3) - ttm%ntsys(3)) / 2
      numbcmap = Mod(numbcmap + ttm%ntsys(3) * (ttm%eltcell(3) + 1) - 1, ttm%ntsys(3)) + 1
      If (numbc > ttm%ntcelloff(3) .and. numbc <= (ttm%ntcelloff(3) + ttm%ntcell(3))) Then
        ttm%ttmbc(5) = numbc - ttm%ntcelloff(3)
        Do i = 0, domain%nz - 1
          start = config%cell(9) * Real(i, wp) * domain%nz_recip
          finish = config%cell(9) * Real(i + 1, wp) * domain%nz_recip
          If (numbcmap > Ceiling(start / ttm%delz) .and. numbcmap <= Ceiling(finish / ttm%delz)) Then
            ttm%ttmbcmap(5) = idcube(domain%idx, domain%idy, i, domain)
          End If
        End Do
      Else
        ttm%ttmbc(5) = 0
        ttm%ttmbcmap(5) = -1
      End If

      !   +z boundary
      numbc = (ttm%eltsys(3) - ttm%ntsys(3)) / 2
      numbc = Mod(numbc + ttm%ntsys(3) * (ttm%eltcell(3) + 1) - 1, ttm%ntsys(3)) + 1
      numbcmap = -(ttm%eltsys(3) - ttm%ntsys(3)) / 2
      numbcmap = Mod(numbcmap + ttm%ntsys(3) * (ttm%eltcell(3) + 1), ttm%ntsys(3)) + 1
      If (numbc > ttm%ntcelloff(3) .and. numbc <= (ttm%ntcelloff(3) + ttm%ntcell(3))) Then
        ttm%ttmbc(6) = numbc - ttm%ntcelloff(3)
        Do i = 0, domain%nz - 1
          start = config%cell(9) * Real(i, wp) * domain%nz_recip
          finish = config%cell(9) * Real(i + 1, wp) * domain%nz_recip
          If (numbcmap > Ceiling(start / ttm%delz) .and. numbcmap <= Ceiling(finish / ttm%delz)) Then
            ttm%ttmbcmap(6) = idcube(domain%idx, domain%idy, i, domain)
          End If
        End Do
      Else
        ttm%ttmbc(6) = 0
        ttm%ttmbcmap(6) = -1
      End If

      ! Derived MPI datatypes for communication of temperatures (MPI 2.x+)

      If (comm%mxnode > 1) Then
        Call MPI_TYPE_GET_EXTENT(wp_mpi, lb1, dbleth, ierr)
        Call MPI_TYPE_GET_EXTENT(MPI_INTEGER, lb2, inleth, ierr)
        Call MPI_TYPE_VECTOR(1, 1, 1, wp_mpi, basicslice, ierr)
        Call MPI_TYPE_VECTOR(1, 2, 2, MPI_INTEGER, bbasicslice, ierr)
        xlth = (ttm%ntcell(1) + 2) * dbleth
        ylth = (ttm%ntcell(2) + 2) * xlth
        Call MPI_TYPE_CREATE_HVECTOR(ttm%ntcell(2), 1, xlth, basicslice, oneslicex, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(ttm%ntcell(3), 1, ylth, oneslicex, ttm%tmpmsgx, ierr)
        Call MPI_TYPE_COMMIT(ttm%tmpmsgx, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(1, ttm%ntcell(1) + 2, xlth, basicslice, oneslicey, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(ttm%ntcell(3), 1, ylth, oneslicey, ttm%tmpmsgy, ierr)
        Call MPI_TYPE_COMMIT(ttm%tmpmsgy, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(1, ttm%ntcell(1) + 2, xlth, basicslice, oneslicez, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(1, ttm%ntcell(2) + 2, ylth, oneslicez, ttm%tmpmsgz, ierr)
        Call MPI_TYPE_COMMIT(ttm%tmpmsgz, ierr)
        bxlth = 2 * (ttm%ntcell(1) + 2) * inleth
        bylth = (ttm%ntcell(2) + 2) * bxlth
        Call MPI_TYPE_CREATE_HVECTOR(ttm%ntcell(2), 1, bxlth, bbasicslice, boneslicex, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(ttm%ntcell(3), 1, bylth, boneslicex, ttm%nummsgx, ierr)
        Call MPI_TYPE_COMMIT(ttm%nummsgx, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(1, ttm%ntcell(1) + 2, bxlth, bbasicslice, boneslicey, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(ttm%ntcell(3), 1, bylth, boneslicey, ttm%nummsgy, ierr)
        Call MPI_TYPE_COMMIT(ttm%nummsgy, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(1, ttm%ntcell(1) + 2, bxlth, bbasicslice, boneslicez, ierr)
        Call MPI_TYPE_CREATE_HVECTOR(1, ttm%ntcell(2) + 2, bylth, boneslicez, ttm%nummsgz, ierr)
        Call MPI_TYPE_COMMIT(ttm%nummsgz, ierr)
      Else
        ttm%tmpmsgx = 0; ttm%tmpmsgy = 0; ttm%tmpmsgz = 0
        ttm%nummsgx = 0; ttm%nummsgy = 0; ttm%nummsgz = 0
      End If

      ! Array allocation and initialization

      Allocate (ttm%eltemp(1:ttm%numcell, -ttm%eltcell(1):ttm%eltcell(1), -ttm%eltcell(2):ttm%eltcell(2), &
                           -ttm%eltcell(3):ttm%eltcell(3)), Stat=fail(1))
      Allocate (ttm%asource(1:ttm%numcell), ttm%tempion(1:ttm%numcell), ttm%gsource(1:ttm%numcell) &
                , Stat=fail(2))
      Allocate (ttm%ttmvom(1:ttm%numcell, 1:4), Stat=fail(3))
      Allocate (ttm%eltemp_adj(1:ttm%numcell, -ttm%eltcell(1):ttm%eltcell(1), &
                               -ttm%eltcell(2):ttm%eltcell(2), -ttm%eltcell(3):ttm%eltcell(3)), Stat=fail(4))
      Allocate (ttm%act_ele_cell(1:ttm%numcell, -1:1, -1:1, -1:1), &
                ttm%old_ele_cell(1:ttm%numcell, -1:1, -1:1, -1:1), Stat=fail(5))
      Allocate (ttm%adjust(1:ttm%numcell, -1:1, -1:1, -1:1), Stat=fail(6))

      If (Any(fail > 0)) Call error(1083)

      ttm%eltemp(:, :, :, :) = 0.0_wp
      ttm%eltemp_adj(:, :, :, :) = 0.0_wp
      ttm%gsource(:) = 0.0_wp
      ttm%asource(:) = 0.0_wp
      ttm%tempion(:) = 0.0_wp
      ttm%ttmvom(:, :) = 0.0_wp
      ttm%act_ele_cell(:, :, :, :) = 1.0_wp
      ttm%old_ele_cell(:, :, :, :) = 1.0_wp
      ttm%acell = ttm%ntsys(1) * ttm%ntsys(2) * ttm%ntsys(3)
      ttm%acell_old = ttm%acell
      ttm%adjust = .false.
      ttm%findepo = .false.
      ttm%cel = 0
      ttm%kel = 0
      ttm%del = 0
      ttm%gel = 0

    End If

    ttm%keyres0 = RESTART_KEY_OLD

  End Subroutine allocate_ttm_arrays

  Subroutine deallocate_ttm_arrays(ttm)

    Type(ttm_type), Intent(InOut) :: ttm

    Integer, Dimension(1:5) :: fail

    fail = 0

    Deallocate (ttm%eltemp, ttm%eltemp_adj, ttm%asource, ttm%tempion, ttm%gsource, &
                ttm%ttmvom, ttm%act_ele_cell, ttm%old_ele_cell, ttm%adjust, Stat=fail(1))
    If (ttm%kel > 0) Deallocate (ttm%ketable, Stat=fail(2))
    If (ttm%cel > 0) Deallocate (ttm%cetable, Stat=fail(3))
    If (ttm%del > 0) Deallocate (ttm%detable, Stat=fail(4))
    If (ttm%gel > 0) Deallocate (ttm%gtable, Stat=fail(5))

    If (Any(fail > 0)) Call error(1084)

  End Subroutine deallocate_ttm_arrays

  Subroutine eltemp_sum(eltempsum, ttm, comm)

    ! Find sum of electronic temperatures over all active CET voxels

    Real(Kind=wp),    Intent(  Out) :: eltempsum
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Integer       :: i, ii, ijk, imax, imin, j, jj, jmax, jmin, k, kk, kmax, kmin, lx, ly, lz
    Logical       :: lcentre, lrange
    Real(Kind=wp) :: tmp

    eltempsum = 0.0_wp

    Do kk = -ttm%eltcell(3), ttm%eltcell(3)
      If (ttm%eltcell(3) > 0 .and. kk == -ttm%eltcell(3) .and. ttm%ttmbcmap(5) >= 0) Then
        kmin = ttm%ttmbc(5)
      Else
        kmin = 1
      End If
      If (ttm%eltcell(3) > 0 .and. kk == ttm%eltcell(3) .and. ttm%ttmbcmap(6) >= 0) Then
        kmax = ttm%ttmbc(6)
      Else
        kmax = ttm%ntcell(3)
      End If

      Do jj = -ttm%eltcell(2), ttm%eltcell(2)
        If (ttm%eltcell(2) > 0 .and. jj == -ttm%eltcell(2) .and. ttm%ttmbcmap(3) >= 0) Then
          jmin = ttm%ttmbc(3)
        Else
          jmin = 1
        End If
        If (ttm%eltcell(2) > 0 .and. jj == ttm%eltcell(2) .and. ttm%ttmbcmap(4) >= 0) Then
          jmax = ttm%ttmbc(4)
        Else
          jmax = ttm%ntcell(2)
        End If

        Do ii = -ttm%eltcell(1), ttm%eltcell(1)
          If (ttm%eltcell(1) > 0 .and. ii == -ttm%eltcell(1) .and. ttm%ttmbcmap(1) >= 0) Then
            imin = ttm%ttmbc(1)
          Else
            imin = 1
          End If
          If (ttm%eltcell(1) > 0 .and. ii == ttm%eltcell(1) .and. ttm%ttmbcmap(2) >= 0) Then
            imax = ttm%ttmbc(2)
          Else
            imax = ttm%ntcell(1)
          End If
          lcentre = (ii == 0 .and. jj == 0 .and. kk == 0)
          Do k = kmin, kmax
            lz = k + ttm%ntcelloff(3) + (kk + ttm%eltcell(3)) * ttm%ntsys(3) - ttm%zeroE(3)
            Do j = jmin, jmax
              ly = j + ttm%ntcelloff(2) + (jj + ttm%eltcell(2)) * ttm%ntsys(2) - ttm%zeroE(2)
              Do i = imin, imax
                lx = i + ttm%ntcelloff(1) + (ii + ttm%eltcell(1)) * ttm%ntsys(1) - ttm%zeroE(1)
           lrange = (lx > 0 .and. lx <= ttm%eltsys(1) .and. ly > 0 .and. ly <= ttm%eltsys(2) .and. lz > 0 .and. lz <= ttm%eltsys(3))
                ijk = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
                tmp = ttm%eltemp(ijk, ii, jj, kk) * Merge(ttm%act_ele_cell(ijk, 0, 0, 0), 1.0_wp, lcentre)
                If (lrange) eltempsum = eltempsum + tmp
              End Do
            End Do
          End Do

        End Do
      End Do
    End Do

    Call gsum(comm, eltempsum)

  End Subroutine eltemp_sum

  Subroutine eltemp_mean(eltempav, ttm, comm)

    ! Find mean electronic temperature over all active CET voxels

    Real(Kind=wp),    Intent(  Out) :: eltempav
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Integer       :: i, ii, ijk, imax, imin, j, jj, jmax, jmin, k, kk, kmax, kmin, lx, ly, lz
    Logical       :: lcentre, lrange
    Real(Kind=wp) :: acl, tmp

    eltempav = 0.0_wp
    acl = 0.0_wp

    Do kk = -ttm%eltcell(3), ttm%eltcell(3)
      If (ttm%eltcell(3) > 0 .and. kk == -ttm%eltcell(3) .and. ttm%ttmbcmap(5) >= 0) Then
        kmin = ttm%ttmbc(5)
      Else
        kmin = 1
      End If
      If (ttm%eltcell(3) > 0 .and. kk == ttm%eltcell(3) .and. ttm%ttmbcmap(6) >= 0) Then
        kmax = ttm%ttmbc(6)
      Else
        kmax = ttm%ntcell(3)
      End If

      Do jj = -ttm%eltcell(2), ttm%eltcell(2)
        If (ttm%eltcell(2) > 0 .and. jj == -ttm%eltcell(2) .and. ttm%ttmbcmap(3) >= 0) Then
          jmin = ttm%ttmbc(3)
        Else
          jmin = 1
        End If
        If (ttm%eltcell(2) > 0 .and. jj == ttm%eltcell(2) .and. ttm%ttmbcmap(4) >= 0) Then
          jmax = ttm%ttmbc(4)
        Else
          jmax = ttm%ntcell(2)
        End If

        Do ii = -ttm%eltcell(1), ttm%eltcell(1)
          If (ttm%eltcell(1) > 0 .and. ii == -ttm%eltcell(1) .and. ttm%ttmbcmap(1) >= 0) Then
            imin = ttm%ttmbc(1)
          Else
            imin = 1
          End If
          If (ttm%eltcell(1) > 0 .and. ii == ttm%eltcell(1) .and. ttm%ttmbcmap(2) >= 0) Then
            imax = ttm%ttmbc(2)
          Else
            imax = ttm%ntcell(1)
          End If
          lcentre = (ii == 0 .and. jj == 0 .and. kk == 0)
          Do k = kmin, kmax
            lz = k + ttm%ntcelloff(3) + (kk + ttm%eltcell(3)) * ttm%ntsys(3) - ttm%zeroE(3)
            Do j = jmin, jmax
              ly = j + ttm%ntcelloff(2) + (jj + ttm%eltcell(2)) * ttm%ntsys(2) - ttm%zeroE(2)
              Do i = imin, imax
                lx = i + ttm%ntcelloff(1) + (ii + ttm%eltcell(1)) * ttm%ntsys(1) - ttm%zeroE(1)
           lrange = (lx > 0 .and. lx <= ttm%eltsys(1) .and. ly > 0 .and. ly <= ttm%eltsys(2) .and. lz > 0 .and. lz <= ttm%eltsys(3))
                ijk = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
                tmp = Merge(ttm%act_ele_cell(ijk, 0, 0, 0), 1.0_wp, lcentre)
                If (lrange) Then
                  eltempav = eltempav + ttm%eltemp(ijk, ii, jj, kk) * tmp
                  acl = acl + tmp
                End If
              End Do
            End Do
          End Do

        End Do
      End Do
    End Do

    Call gsum(comm, eltempav)
    Call gsum(comm, acl)

    If (acl > zero_plus) eltempav = eltempav / acl

  End Subroutine eltemp_mean

  Subroutine eltemp_maxKe(temp, eltempmax, ttm, comm)

    ! Find maximum temperature for calculating tabulated
    ! thermal conductivities (ionic or system) over all
    ! active CET voxels (note that system temperature
    ! applies over all CET voxels that do not overlap
    ! CIT voxels)

    Real(Kind=wp),    Intent(In   ) :: temp
    Real(Kind=wp),    Intent(  Out) :: eltempmax
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Integer       :: i, ijk, j, k
    Real(Kind=wp) :: eltempKe

    eltempmax = 0.0_wp

    Do k = 1, ttm%ntcell(3)
      Do j = 1, ttm%ntcell(2)
        Do i = 1, ttm%ntcell(1)
          ijk = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
          eltempKe = ttm%tempion(ijk)
          If (ttm%act_ele_cell(ijk, 0, 0, 0) > zero_plus) eltempmax = Max(eltempmax, eltempKe)
        End Do
      End Do
    End Do
    eltempmax = Max(eltempmax, temp)

    Call gmax(comm, eltempmax)

  End Subroutine eltemp_maxKe

  Subroutine eltemp_max(eltempmax, ttm, comm)

    ! Find maximum electronic temperature over all
    ! active CET ttm%cells
    Real(Kind=wp),    Intent(  Out) :: eltempmax
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Integer :: i, ii, ijk, imax, imin, j, jj, jmax, jmin, k, kk, kmax, kmin, lx, ly, lz
    Logical :: lcentre, lrange

    eltempmax = 0.0_wp

    Do kk = -ttm%eltcell(3), ttm%eltcell(3)
      If (ttm%eltcell(3) > 0 .and. kk == -ttm%eltcell(3) .and. ttm%ttmbcmap(5) >= 0) Then
        kmin = ttm%ttmbc(5)
      Else
        kmin = 1
      End If
      If (ttm%eltcell(3) > 0 .and. kk == ttm%eltcell(3) .and. ttm%ttmbcmap(6) >= 0) Then
        kmax = ttm%ttmbc(6)
      Else
        kmax = ttm%ntcell(3)
      End If

      Do jj = -ttm%eltcell(2), ttm%eltcell(2)
        If (ttm%eltcell(2) > 0 .and. jj == -ttm%eltcell(2) .and. ttm%ttmbcmap(3) >= 0) Then
          jmin = ttm%ttmbc(3)
        Else
          jmin = 1
        End If
        If (ttm%eltcell(2) > 0 .and. jj == ttm%eltcell(2) .and. ttm%ttmbcmap(4) >= 0) Then
          jmax = ttm%ttmbc(4)
        Else
          jmax = ttm%ntcell(2)
        End If

        Do ii = -ttm%eltcell(1), ttm%eltcell(1)
          If (ttm%eltcell(1) > 0 .and. ii == -ttm%eltcell(1) .and. ttm%ttmbcmap(1) >= 0) Then
            imin = ttm%ttmbc(1)
          Else
            imin = 1
          End If
          If (ttm%eltcell(1) > 0 .and. ii == ttm%eltcell(1) .and. ttm%ttmbcmap(2) >= 0) Then
            imax = ttm%ttmbc(2)
          Else
            imax = ttm%ntcell(1)
          End If
          lcentre = (ii == 0 .and. jj == 0 .and. kk == 0)
          Do k = kmin, kmax
            lz = k + ttm%ntcelloff(3) + (kk + ttm%eltcell(3)) * ttm%ntsys(3) - ttm%zeroE(3)
            Do j = jmin, jmax
              ly = j + ttm%ntcelloff(2) + (jj + ttm%eltcell(2)) * ttm%ntsys(2) - ttm%zeroE(2)
              Do i = imin, imax
                lx = i + ttm%ntcelloff(1) + (ii + ttm%eltcell(1)) * ttm%ntsys(1) - ttm%zeroE(1)
                ijk = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
           lrange = (lx > 0 .and. lx <= ttm%eltsys(1) .and. ly > 0 .and. ly <= ttm%eltsys(2) .and. lz > 0 .and. lz <= ttm%eltsys(3))
                If (lcentre) lrange = (lrange .and. (ttm%act_ele_cell(ijk, 0, 0, 0) > zero_plus))
                If (lrange) eltempmax = Max(eltempmax, ttm%eltemp(ijk, ii, jj, kk))
              End Do
            End Do
          End Do

        End Do
      End Do
    End Do

    Call gmax(comm, eltempmax)

  End Subroutine eltemp_max

  Subroutine eltemp_minKe(temp, eltempmin, ttm, comm)

    ! Find minimum temperature for calculating tabulated
    ! thermal conductivities (ionic or system) over all
    ! active CET voxels (note that system temperature
    ! applies over all CET voxels that do not overlap
    ! CIT voxels)
    Real(Kind=wp),    Intent(In   ) :: temp
    Real(Kind=wp),    Intent(  Out) :: eltempmin
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Integer       :: i, ijk, j, k
    Real(Kind=wp) :: eltempKe

    eltempmin = 1.0e30_wp

    Do k = 1, ttm%ntcell(3)
      Do j = 1, ttm%ntcell(2)
        Do i = 1, ttm%ntcell(1)
          ijk = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
          eltempKe = ttm%tempion(ijk)
          If (ttm%act_ele_cell(ijk, 0, 0, 0) > zero_plus) eltempmin = Min(eltempmin, eltempKe)
        End Do
      End Do
    End Do

    eltempmin = Min(eltempmin, temp)

    Call gmin(comm, eltempmin)

  End Subroutine eltemp_minKe

  Subroutine eltemp_min(eltempmin, ttm, comm)

    ! Find minimum electronic temperature over all
    ! active CET ttm%cells

    Real(Kind=wp),    Intent(  Out) :: eltempmin
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Integer :: i, ii, ijk, imax, imin, j, jj, jmax, jmin, k, kk, kmax, kmin, lx, ly, lz
    Logical :: lrange

    eltempmin = 1.0e30_wp

    Do kk = -ttm%eltcell(3), ttm%eltcell(3)
      If (ttm%eltcell(3) > 0 .and. kk == -ttm%eltcell(3) .and. ttm%ttmbcmap(5) >= 0) Then
        kmin = ttm%ttmbc(5)
      Else
        kmin = 1
      End If
      If (ttm%eltcell(3) > 0 .and. kk == ttm%eltcell(3) .and. ttm%ttmbcmap(6) >= 0) Then
        kmax = ttm%ttmbc(6)
      Else
        kmax = ttm%ntcell(3)
      End If

      Do jj = -ttm%eltcell(2), ttm%eltcell(2)
        If (ttm%eltcell(2) > 0 .and. jj == -ttm%eltcell(2) .and. ttm%ttmbcmap(3) >= 0) Then
          jmin = ttm%ttmbc(3)
        Else
          jmin = 1
        End If
        If (ttm%eltcell(2) > 0 .and. jj == ttm%eltcell(2) .and. ttm%ttmbcmap(4) >= 0) Then
          jmax = ttm%ttmbc(4)
        Else
          jmax = ttm%ntcell(2)
        End If

        Do ii = -ttm%eltcell(1), ttm%eltcell(1)
          If (ttm%eltcell(1) > 0 .and. ii == -ttm%eltcell(1) .and. ttm%ttmbcmap(1) >= 0) Then
            imin = ttm%ttmbc(1)
          Else
            imin = 1
          End If
          If (ttm%eltcell(1) > 0 .and. ii == ttm%eltcell(1) .and. ttm%ttmbcmap(2) >= 0) Then
            imax = ttm%ttmbc(2)
          Else
            imax = ttm%ntcell(1)
          End If
          Do k = kmin, kmax
            lz = k + ttm%ntcelloff(3) + (kk + ttm%eltcell(3)) * ttm%ntsys(3) - ttm%zeroE(3)
            Do j = jmin, jmax
              ly = j + ttm%ntcelloff(2) + (jj + ttm%eltcell(2)) * ttm%ntsys(2) - ttm%zeroE(2)
              Do i = imin, imax
                lx = i + ttm%ntcelloff(1) + (ii + ttm%eltcell(1)) * ttm%ntsys(1) - ttm%zeroE(1)
                ijk = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
           lrange = (lx > 0 .and. lx <= ttm%eltsys(1) .and. ly > 0 .and. ly <= ttm%eltsys(2) .and. lz > 0 .and. lz <= ttm%eltsys(3))
                If (ii == 0 .and. jj == 0 .and. kk == 0) lrange = (lrange .and. (ttm%act_ele_cell(ijk, 0, 0, 0) > zero_plus))
                If (lrange) eltempmin = Min(eltempmin, ttm%eltemp(ijk, ii, jj, kk))
              End Do
            End Do
          End Do

        End Do
      End Do
    End Do

    Call gmin(comm, eltempmin)

  End Subroutine eltemp_min

  Subroutine depoinit(time, ttm, comm)

    ! determine initial energy deposition to electronic system,
    ! both temporally and spatially
    Real(Kind=wp),    Intent(In   ) :: time
    Type(ttm_type),   Intent(inOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Character(Len=14)       :: number
    Character(Len=256)      :: messages(2)
    Integer, Dimension(1:3) :: fail

    fail = 0

    Allocate (ttm%lat_U(0:ttm%ntcell(1) + 1, 0:ttm%ntcell(2) + 1, 0:ttm%ntcell(3) + 1), Stat=fail(1))
    Allocate (ttm%lat_B(0:ttm%ntcell(1) + 1, 0:ttm%ntcell(2) + 1, 0:ttm%ntcell(3) + 1), Stat=fail(2))
    Allocate (ttm%lat_I(0:ttm%ntcell(1) + 1, 0:ttm%ntcell(2) + 1, 0:ttm%ntcell(3) + 1), Stat=fail(3))

    If (Any(fail > 0)) Call error(1089)

    ttm%lat_U(:, :, :) = 0.0_wp ! spatial deposition (eV)
    ttm%lat_B(:, :, :) = 0.0_wp ! temporal deposition of ttm%lat_U (eV)
    ttm%lat_I(:, :, :) = 0.0_wp ! sum of temporal deposition of ttm%lat_B (eV)

    ! spatial distribution of track

    Select Case (ttm%sdepoType)
    Case (TTM_SDEPO_GAUSS)
      ! Gaussian spatial deposition
      Call gaussianTrack(ttm%lat_U, ttm, comm)
    Case (TTM_SDEPO_FLAT)
      ! Constant (flat) spatial deposition
      Call uniformDist(ttm%lat_U, ttm)
    Case (TTM_SDEPO_EXP)
      ! xy-flat, z-exp spatial deposition
      Call uniformDistZexp(ttm%lat_U, ttm)
    End Select

    ttm%trackInit = .true. ! switch on flag indicating track initialisation is in progress
    If (ttm%depostart <= zero_plus) ttm%depostart = time ! time (ps) when deposition starts, i.e. current time

    ! temporal deposition of track: calculate time ttm%normalisation factor

    Select Case (ttm%tdepoType)
      !   type=1: gauss(t)
    Case (TTM_TDEPO_GAUSS)
      ! Gaussian temporal deposition
      ttm%norm = 1.0_wp / (sqrpi * rt2 * ttm%tdepo)
      ttm%depoend = ttm%depostart + 2.0_wp * ttm%tcdepo * ttm%tdepo
    Case (TTM_TDEPO_EXP)
      ! decaying exponential temporal deposition
      ttm%norm = 1.0_wp / (1.0_wp - Exp(-ttm%tcdepo))
      ttm%depoend = ttm%depostart + 2.0_wp * ttm%tcdepo * ttm%tdepo
    Case (TTM_TDEPO_DELTA)
      ! delta temporal deposition
      ttm%norm = 1.0_wp
      ttm%depoend = ttm%depostart
    Case (TTM_TDEPO_PULSE)
      ! pulse temporal deposition
      ttm%norm = 1.0_wp / ttm%tdepo
      ttm%depoend = ttm%depostart + ttm%tdepo
    End Select

    ! report start of energy deposition

    Write (number, '(f14.5)') ttm%depostart
    Write (messages(1), "(a,a,a)") &
      'electronic energy deposition starting at time = ', Trim(Adjustl(number)), ' ps'
    Write (messages(2), "(130('-'))")
    Call info(messages, 2, .true.)

  End Subroutine depoinit

  Subroutine ttm_system_init(nstep, nsteql, keyres, dumpfile, time, temp, domain, ttm, comm)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !
    ! dl_poly_4 subroutine for writing electronic temperature restart files
    ! at job termination or selected intervals in simulation
    !
    ! copyright - daresbury laboratory
    ! authors   - s.l.daraszewicz & m.a.seaton september 2017
    ! refactoring:
    !           - a.m.elena march-october 2018
    !           - j.madge march-october 2018
    !           - a.b.g.chalk march-october 2018
    !           - i.scivetti march-october 2018
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    Integer,            Intent(In   ) :: nstep, nsteql, keyres
    Character(Len=*),   Intent(In   ) :: dumpfile
    Real(Kind=wp),      Intent(In   ) :: time, temp
    Type(domains_type), Intent(In   ) :: domain
    Type(ttm_type),     Intent(InOut) :: ttm
    Type(comms_type),   Intent(InOut) :: comm

    Character(Len=200) :: record
    Character(Len=256) :: message
    Character(Len=40)  :: word
    Integer            :: i, ii, ijk, iounit, ipos(3), ix, iy, iz, jj, kk, nstp, nxx, nyy, &
                          nzz
    Logical            :: l_tmp = .true., safe
    Real(Kind=wp)      :: eltmp, lat_max, lat_min, lat_sum, tme

    ! check existence of readable restart file (DUMP_E)

    If (comm%idnode == 0) Inquire (File=dumpfile, Exist=l_tmp)
    Call gcheck(comm, l_tmp)
    If ((.not. l_tmp) .and. keyres == ttm%keyres0) Call error(694)

    ! if restarting simulation, read restart file

    If (l_tmp .and. keyres == ttm%keyres0) Then

      If (comm%idnode == 0) Open (Newunit=iounit, File=Trim(dumpfile))
      Call get_line(safe, iounit, record, comm); If (.not. safe) Goto 100
      Call get_word(record, word); nxx = Nint(word_2_real(word, 0.0_wp))
      Call get_word(record, word); nyy = Nint(word_2_real(word, 0.0_wp))
      Call get_word(record, word); nzz = Nint(word_2_real(word, 0.0_wp))
      Call get_line(safe, iounit, record, comm); If (.not. safe) Goto 100
      Call get_word(record, word); nstp = Nint(word_2_real(word, 0.0_wp))
      Call get_word(record, word); tme = word_2_real(word, 0.0_wp)
      Call get_word(record, word); ttm%depostart = word_2_real(word, 0.0_wp)
      Call get_word(record, word); ttm%depoend = word_2_real(word, 0.0_wp)
      ! check size of electronic temperature grid matches with size given in CONTROL file
      If (nxx /= ttm%eltsys(1) .or. nyy /= ttm%eltsys(2) .or. nzz /= ttm%eltsys(3)) Call error(695)
      ! check restart file is at same timestep as restart
      ! (can proceed if not, but need to warn user)
      If (nstp /= nstep .or. Abs(tme - time) > zero_plus) Call warning(520, 0.0_wp, 0.0_wp, 0.0_wp)
      ! read in each line, find appropriate grid config%cell and assign
      ! electronic temperature if processor has that config%cell
      Do i = 1, ttm%eltsys(1) * ttm%eltsys(2) * ttm%eltsys(3)
        Call get_line(safe, iounit, record, comm); If (.not. safe) Goto 100
        Call get_word(record, word); ipos(1) = Nint(word_2_real(word, 0.0_wp))
        Call get_word(record, word); ipos(2) = Nint(word_2_real(word, 0.0_wp))
        Call get_word(record, word); ipos(3) = Nint(word_2_real(word, 0.0_wp))
        Call get_word(record, word); eltmp = word_2_real(word, 0.0_wp)
        ix = ipos(1) + ttm%midI(1) - 1
        iy = ipos(2) + ttm%midI(2) - 1
        iz = ipos(3) + ttm%midI(3) - 1
        ii = Floor(Real(ix, Kind=wp) / Real(ttm%ntsys(1), Kind=wp))
        jj = Floor(Real(iy, Kind=wp) / Real(ttm%ntsys(2), Kind=wp))
        kk = Floor(Real(iz, Kind=wp) / Real(ttm%ntsys(3), Kind=wp))
        ix = Mod(ix + ttm%ntsys(1) * ttm%eltcell(1), ttm%ntsys(1)) + 1 - ttm%ntcelloff(1)
        iy = Mod(iy + ttm%ntsys(2) * ttm%eltcell(2), ttm%ntsys(2)) + 1 - ttm%ntcelloff(2)
        iz = Mod(iz + ttm%ntsys(3) * ttm%eltcell(3), ttm%ntsys(3)) + 1 - ttm%ntcelloff(3)
        If (ix > 0 .and. ix <= ttm%ntcell(1) .and. iy > 0 .and. iy <= ttm%ntcell(2) .and. iz > 0 .and. iz <= ttm%ntcell(3)) Then
          ijk = 1 + ix + (ttm%ntcell(1) + 2) * (iy + (ttm%ntcell(2) + 2) * iz)
          ttm%eltemp(ijk, ii, jj, kk) = eltmp
        End If
      End Do
      ! fill boundary halo values and deal with required boundary conditions
      Call boundaryHalo(ttm, domain, comm)
      Call boundaryCond(ttm%bcTypeE, temp, ttm, comm)
      ! check whether or not energy deposition has happened yet
      If (nstep > nsteql .and. time >= ttm%depostart .and. time < ttm%depoend) Then
        Call depoinit(time, ttm, comm)
      Else If (time >= ttm%depoend) Then
        ttm%findepo = .true.
      End If

      ! report successful reading and minimum, maximum and sums of
      ! electronic temperatures
      Call eltemp_sum(lat_sum, ttm, comm)
      Call eltemp_max(lat_max, ttm, comm)
      Call eltemp_min(lat_min, ttm, comm)
      Write (message, '(a)') 'electronic temperatures read from DUMP_E file for two-temperature model'
      Call info(message, .true.)
      Write (message, '(1x,"minimum temperature (K) = ",ES11.4,/,1x,"maximum temperature (K) = ",' &
             //'ES11.4,/,1x,"sum of temperatures (K) = ",ES11.4)') &
        lat_min, lat_max, lat_sum
      Call info(message, .true.)
      If (comm%idnode == 0) Then
        Close (iounit)
      End If

    Else

      ! if not restarting simulation, set electronic temperature grid
      ! to system temperature

      ttm%eltemp = temp

    End If

    Return

    ! Abttm%normal exit from electronic temperature dump file read

    100 Continue

    Write (message, "(a)") dumpfile, ' data mishmash detected'
    Call error(696, message, .true.)
    Return

  End Subroutine ttm_system_init

  Subroutine ttm_system_revive &
    (dumpfile, nstep, time, freq, nstrun, ttm, comm)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !
    ! dl_poly_4 subroutine for writing electronic temperature restart files
    ! at job termination or selected intervals in simulation
    !
    ! copyright - daresbury laboratory
    ! authors   - s.l.daraszewicz & m.a.seaton september 2015
    ! refactoring:
    !           - a.m.elena march-october 2018
    !           - j.madge march-october 2018
    !           - a.b.g.chalk march-october 2018
    !           - i.scivetti march-october 2018
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    Character(Len=*), Intent(In   ) :: dumpfile
    Integer,          Intent(In   ) :: nstep
    Real(Kind=wp),    Intent(In   ) :: time
    Integer,          Intent(In   ) :: freq, nstrun
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Integer :: i, id, ii, ijk, imax, imin, iounit, ix, iy, iz, j, jj, jmax, jmin, k, kk, &
               kmax, kmin
    Logical :: lrange

    If (freq /= 0) Then
      If (Mod(nstep, freq) == 0 .or. nstep == nstrun) Then

        If (comm%idnode == 0) Then
          Open (Newunit=iounit, File=Trim(dumpfile), Status='replace')
          Write (iounit, '(3i8)') ttm%eltsys(1), ttm%eltsys(2), ttm%eltsys(3)
          Write (iounit, '(i12,3(2x,es24.15))') nstep, time, ttm%depostart, ttm%depoend
          Close (iounit)
        End If
        Call gsync(comm)

        Do id = 0, comm%mxnode - 1
          If (comm%idnode == id) Then
            Open (Newunit=iounit, File=dumpfile, Status='old', Position='append')
            Do kk = -ttm%eltcell(3), ttm%eltcell(3)
              If (ttm%eltcell(3) > 0 .and. kk == -ttm%eltcell(3) .and. ttm%ttmbcmap(5) >= 0) Then
                kmin = ttm%ttmbc(5)
              Else
                kmin = 1
              End If
              If (ttm%eltcell(3) > 0 .and. kk == ttm%eltcell(3) .and. ttm%ttmbcmap(6) >= 0) Then
                kmax = ttm%ttmbc(6)
              Else
                kmax = ttm%ntcell(3)
              End If
              Do jj = -ttm%eltcell(2), ttm%eltcell(2)
                If (ttm%eltcell(2) > 0 .and. jj == -ttm%eltcell(2) .and. ttm%ttmbcmap(3) >= 0) Then
                  jmin = ttm%ttmbc(3)
                Else
                  jmin = 1
                End If
                If (ttm%eltcell(2) > 0 .and. jj == ttm%eltcell(2) .and. ttm%ttmbcmap(4) >= 0) Then
                  jmax = ttm%ttmbc(4)
                Else
                  jmax = ttm%ntcell(2)
                End If
                Do ii = -ttm%eltcell(1), ttm%eltcell(1)
                  If (ttm%eltcell(1) > 0 .and. ii == -ttm%eltcell(1) .and. ttm%ttmbcmap(1) >= 0) Then
                    imin = ttm%ttmbc(1)
                  Else
                    imin = 1
                  End If
                  If (ttm%eltcell(1) > 0 .and. ii == ttm%eltcell(1) .and. ttm%ttmbcmap(2) >= 0) Then
                    imax = ttm%ttmbc(2)
                  Else
                    imax = ttm%ntcell(1)
                  End If
                  Do k = kmin, kmax
                    iz = k + ttm%ntcelloff(3) + (kk + ttm%eltcell(3)) * ttm%ntsys(3) - ttm%zeroE(3)
                    Do j = jmin, jmax
                      iy = j + ttm%ntcelloff(2) + (jj + ttm%eltcell(2)) * ttm%ntsys(2) - ttm%zeroE(2)
                      Do i = imin, imax
                        ix = i + ttm%ntcelloff(1) + (ii + ttm%eltcell(1)) * ttm%ntsys(1) - ttm%zeroE(1)
           lrange = (ix > 0 .and. ix <= ttm%eltsys(1) .and. iy > 0 .and. iy <= ttm%eltsys(2) .and. iz > 0 .and. iz <= ttm%eltsys(3))
                        ijk = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
                        If (lrange) Write (iounit, '(3i8,2x,es24.15)') ix - ttm%midE(1), iy - ttm%midE(2), iz - ttm%midE(3), &
                          ttm%eltemp(ijk, ii, jj, kk)
                      End Do
                    End Do
                  End Do
                End Do
              End Do
            End Do
            Close (iounit)
          End If
          Call gsync(comm)
        End Do

      End If
    End If

  End Subroutine ttm_system_revive

  Subroutine ttm_table_read(ttm, comm)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !
    ! dl_poly_4 subroutine for reading specific heat capacity and coupling
    ! constant table files
    !
    ! copyright - daresbury laboratory
    ! author    - m.a.seaton may 2012
    ! contrib   - g.khara may 2016
    ! contrib   - m.a.seaton february 2017
    ! refactoring:
    !           - a.m.elena march-october 2018
    !           - j.madge march-october 2018
    !           - a.b.g.chalk march-october 2018
    !           - i.scivetti march-october 2018
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Character(Len=200) :: record
    Character(Len=256) :: message
    Character(Len=40)  :: word
    Integer            :: i, ntable
    Logical            :: safe
    Real(Kind=wp)      :: vk1, vk2

    ! read thermal conductivity data

    If (ttm%KeType == TTM_KE_TABULATED) Then

      If (comm%idnode == 0) Open (Newunit=ntable, File='Ke.dat', Status='old')

      i = 0
      Do While (i < ttm%kel)

        Call get_line(safe, ntable, record, comm)
        If (.not. safe) Then
          Go To 100
        Else
          Call get_word(record, word)
          vk1 = word_2_real(word)
          Call get_word(record, word)
          vk2 = word_2_real(word)
          If (vk1 >= zero_plus) Then
            i = i + 1
            ttm%ketable(i, 1) = vk1
            ttm%ketable(i, 2) = vk2
          End If
        End If

      End Do

      If (comm%idnode == 0) Then
        Close (Unit=ntable)
      End If
      Write (message, '(a)') 'thermal conductivity table read from Ke.dat file for two-temperature model'
      Call info(message, .true.)
      Write (message, '(1x,"minimum temperature            (K) = ",ES12.4,/,1x,"maximum temperature            (K) = ",' &
             //'ES12.4,/,1x,"minimum t.c. value   (W m^-1 K^-1) = ",ES12.4,/,1x,' &
             //'"maximum t.c. value   (W m^-1 K^-1) = ",ES12.4)') &
        Minval(ttm%ketable(:, 1)), Maxval(ttm%ketable(:, 1)), Minval(ttm%ketable(:, 2)), Maxval(ttm%ketable(:, 2))
      Call info(message, .true.)

      ! convert thermal conductivity values from W m^-1 K^-1 to kB A^-1 ps^-1

      ttm%ketable(1:ttm%kel, 2) = ttm%ketable(1:ttm%kel, 2) * JKms_to_kBAps

    End If

    ! read volumetric heat capacity data

    If (ttm%CeType == TTM_CE_TABULATED .or. ttm%CeType == TTM_CE_TABULATED_DYN) Then

      If (comm%idnode == 0) Open (Newunit=ntable, File='Ce.dat', Status='old')

      i = 0

      Do While (i < ttm%cel)
        Call get_line(safe, ntable, record, comm)
        If (.not. safe) Then
          Go To 100
        Else
          Call get_word(record, word)
          vk1 = word_2_real(word)
          Call get_word(record, word)
          vk2 = word_2_real(word)
          If (vk1 >= zero_plus) Then
            i = i + 1
            ttm%cetable(i, 1) = vk1
            ttm%cetable(i, 2) = vk2
          End If
        End If

      End Do

      If (comm%idnode == 0) Then
        Close (Unit=ntable)
      End If
      Write (message, '(a)') 'electronic volumetric heat capacity table read from Ce.dat file for two-temperature model'
      Call info(message, .true.)
      Write (message, '(1x,"minimum temperature            (K) = ",ES12.4)') Minval(ttm%cetable(:, 1))
      Call info(message, .true.)
      Write (message, '(1x,"maximum temperature            (K) = ",ES12.4)') Maxval(ttm%cetable(:, 1))
      Call info(message, .true.)
      Write (message, '(1x,"minimum v.h.c. value (J m^-3 K^-1) = ",ES12.4)') Minval(ttm%cetable(:, 2))
      Call info(message, .true.)
      Write (message, '(1x,"maximum v.h.c. value (J m^-3 K^-1) = ",ES12.4)') Maxval(ttm%cetable(:, 2))
      Call info(message, .true.)

      ! convert volumetric heat capacity values from J m^-3 K^-1 to kB A^-3

      ttm%cetable(1:ttm%cel, 2) = ttm%cetable(1:ttm%cel, 2) * Jm3K_to_kBA3

    End If

    ! read thermal diffusivity data

    If (ttm%DeType == TTM_DE_TABULATED) Then

      If (comm%idnode == 0) Open (Newunit=ntable, File='De.dat', Status='old')

      i = 0
      Do While (i < ttm%del)

        Call get_line(safe, ntable, record, comm)
        If (.not. safe) Then
          Go To 100
        Else
          Call get_word(record, word)
          vk1 = word_2_real(word)
          Call get_word(record, word)
          vk2 = word_2_real(word)
          If (vk1 >= zero_plus) Then
            i = i + 1
            ttm%detable(i, 1) = vk1
            ttm%detable(i, 2) = vk2
          End If
        End If

      End Do

      If (comm%idnode == 0) Then
        Close (Unit=ntable)
      End If
      Write (message, '(a)') 'thermal diffusivity table read from De.dat file for two-temperature model'
      Call info(message, .true.)
      Write (message, '(1x,"minimum temperature            (K) = ",ES12.4,/,1x,"maximum temperature            (K) = ",ES12.4,' &
             //'/,1x,"minimum diffusivity value  (m^2/s) = ",ES12.4,/,1x,"maximum diffusivity value  (m^2/s) = ",ES12.4)') &
        Minval(ttm%detable(:, 1)), Maxval(ttm%detable(:, 1)), Minval(ttm%detable(:, 2)), Maxval(ttm%detable(:, 2))
      Call info(message, .true.)

      ! convert thermal diffusivity values from m^2 s^-1 to A^2 ps^-1

      ttm%detable(1:ttm%del, 2) = ttm%detable(1:ttm%del, 2) * 1e8_wp

    End If

    ! read coupling constant data

    If (ttm%gvar /= TTM_EPVAR_NULL) Then

      If (comm%idnode == 0) Open (Newunit=ntable, File='g.dat', Status='old')

      i = 0
      Do While (i < ttm%gel)

        Call get_line(safe, ntable, record, comm)
        If (.not. safe) Then
          Go To 100
        Else
          Call get_word(record, word)
          vk1 = word_2_real(word)
          Call get_word(record, word)
          vk2 = word_2_real(word)

          If (vk1 >= zero_plus) Then
            i = i + 1
            ttm%gtable(i, 1) = vk1
            ttm%gtable(i, 2) = vk2
          End If
        End If

      End Do

      If (comm%idnode == 0) Then
        Close (Unit=ntable)
      End If
      Write (message, '(a)') 'electron-phonon coupling table read from g.dat file for two-temperature model'
      Call info(message, .true.)
      Write (message, '(1x,"minimum temperature            (K) = ",ES12.4)') Minval(ttm%gtable(:, 1))
      Call info(message, .true.)
      Write (message, '(1x,"maximum temperature            (K) = ",ES12.4)') Maxval(ttm%gtable(:, 1))
      Call info(message, .true.)
      Write (message, '(1x,"minimum e-p value    (W m^-3 K^-1) = ",ES12.4)') Minval(ttm%gtable(:, 2))
      Call info(message, .true.)
      Write (message, '(1x,"maximum e-p value    (W m^-3 K^-1) = ",ES12.4)') Maxval(ttm%gtable(:, 2))
      Call info(message, .true.)

      ! convert electron-phonon coupling values from W m^-3 K^-1 to ps^-1

      ttm%gtable(1:ttm%gel, 2) = ttm%gtable(1:ttm%gel, 2) * ttm%epc_to_chi

    End If

    Return

    ! end of file error exit

    100 Continue

    If (comm%idnode == 0) Then
      Close (Unit=ntable)
    End If
    Call error(692, master_only=.true.)

  End Subroutine ttm_table_read

  Subroutine ttm_table_scan(mxbuff, ttm, comm)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !
    ! dl_poly_4 subroutine for scanning specific heat capacity,
    ! thermal conductivity and electron-phonon coupling
    ! constant table files to determine numbers of data points
    !
    ! copyright - daresbury laboratory
    ! author    - m.a.seaton may 2012
    ! contrib   - g.khara    may 2016
    ! refactoring:
    !           - a.m.elena march-october 2018
    !           - j.madge march-october 2018
    !           - a.b.g.chalk march-october 2018
    !           - i.scivetti march-october 2018
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    Integer,          Intent(InOut) :: mxbuff
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(InOut) :: comm

    Character(Len=200)                       :: record
    Character(Len=256)                       :: message
    Character(Len=40)                        :: word
    Integer                                  :: fail, ntable
    Logical                                  :: lexist, safe
    Real(Kind=wp)                            :: vk1, vk2
    Real(Kind=wp), Allocatable, Dimension(:) :: buffer

    If (ttm%l_ttm) Then

      fail = 0
      Allocate (buffer(1:mxbuff), Stat=fail)
      If (fail > 0) Then
        Write (message, '(a)') 'ttm_table_scan allocation failure'
        Call error(0, message)
      End If

      ! check existence of thermal conductivity table file

      If (ttm%KeType == TTM_KE_TABULATED) Then

        Inquire (File='Ke.dat', Exist=lexist)
        Call gcheck(comm, lexist)

        If (.not. lexist) Then
          Go To 100
        Else
          If (comm%idnode == 0) Open (Newunit=ntable, File='Ke.dat', Status='old')
        End If

        ! determine number of lines of data to read

        ttm%kel = 0
        Do While (.true.)

          Call get_line(safe, ntable, record, comm)
          If (.not. safe) Then
            Go To 5
          Else
            Call get_word(record, word)
            vk1 = word_2_real(word)
            Call get_word(record, word)
            vk2 = word_2_real(word)
            If (vk1 >= zero_plus) ttm%kel = ttm%kel + 1
          End If

        End Do
        5 Continue

        If (comm%idnode == 0) Close (Unit=ntable)

        ! check number of data lines and allocate array

        safe = (ttm%kel > 0)
        Call gcheck(comm, safe)
        If (.not. safe) Then
          Call error(685)
        Else
          Allocate (ttm%ketable(1:ttm%kel, 2), Stat=fail)
          If (fail > 0) Then
            Write (message, '(a)') 'ttm_table_scan allocation failure'
            Call error(0, message)
          End If
          ttm%ketable(:, :) = 0.0_wp
        End If

      End If

      ! check existence of specific heat capacity table file

      If (ttm%CeType == TTM_CE_TABULATED .or. ttm%CeType == TTM_CE_TABULATED_DYN) Then

        Inquire (File='Ce.dat', Exist=lexist)
        Call gcheck(comm, lexist)
        If (.not. lexist) Then
          Go To 200
        Else
          If (comm%idnode == 0) Open (Newunit=ntable, File='Ce.dat', Status='old')
        End If

        ! determine number of lines of data to read

        ttm%cel = 0
        Do While (.true.)

          Call get_line(safe, ntable, record, comm)
          If (.not. safe) Then
            Go To 10
          Else
            Call get_word(record, word)
            vk1 = word_2_real(word)
            Call get_word(record, word)
            vk2 = word_2_real(word)
            If (vk1 >= zero_plus) ttm%cel = ttm%cel + 1
          End If

        End Do
        10 Continue

        If (comm%idnode == 0) Close (Unit=ntable)

        ! check number of data lines and allocate array

        safe = (ttm%cel > 0)
        Call gcheck(comm, safe)
        If (.not. safe) Then
          Call error(687)
        Else
          Allocate (ttm%cetable(1:ttm%cel, 2), Stat=fail)
          If (fail > 0) Then
            Write (message, '(a)') 'ttm_table_scan allocation failure'
            Call error(0, message)
          End If
          ttm%cetable(:, :) = 0.0_wp
        End If

      End If

      ! check existence of thermal diffusivity table file

      If (ttm%DeType == TTM_DE_TABULATED) Then

        Inquire (File='De.dat', Exist=lexist)
        Call gcheck(comm, lexist)

        If (.not. lexist) Then
          Go To 300
        Else
          If (comm%idnode == 0) Open (Newunit=ntable, File='De.dat', Status='old')
        End If

        ! determine number of lines of data to read

        ttm%del = 0
        Do While (.true.)

          Call get_line(safe, ntable, record, comm)
          If (.not. safe) Then
            Go To 15
          Else
            Call get_word(record, word)
            vk1 = word_2_real(word)
            Call get_word(record, word)
            vk2 = word_2_real(word)
            If (vk1 >= zero_plus) ttm%del = ttm%del + 1
          End If

        End Do
        15 Continue

        If (comm%idnode == 0) Close (Unit=ntable)

        ! check number of data lines and allocate array

        safe = (ttm%del > 0)
        Call gcheck(comm, safe)
        If (.not. safe) Then
          Call error(689)
        Else
          Allocate (ttm%detable(1:ttm%del, 2), Stat=fail)
          If (fail > 0) Then
            Write (message, '(a)') 'ttm_table_scan allocation failure'
            Call error(0, message)
          End If
          ttm%detable(:, :) = 0.0_wp
        End If

      End If

      ! check existence of coupling constant table file

      If (ttm%gvar /= TTM_EPVAR_NULL) Then

        Inquire (File='g.dat', Exist=lexist)
        Call gcheck(comm, lexist)

        If (.not. lexist) Then
          Go To 400
        Else
          If (comm%idnode == 0) Open (Newunit=ntable, File='g.dat', Status='old')
        End If

        ! determine number of lines of data to read

        ttm%gel = 0
        Do While (.true.)

          Call get_line(safe, ntable, record, comm)
          If (.not. safe) Then
            Go To 20
          Else
            Call get_word(record, word)
            vk1 = word_2_real(word)
            Call get_word(record, word)
            vk2 = word_2_real(word)
            If (vk1 >= zero_plus) ttm%gel = ttm%gel + 1
          End If

        End Do
        20 Continue

        If (comm%idnode == 0) Close (Unit=ntable)

        ! check number of data lines and allocate array

        safe = (ttm%gel > 0)
        Call gcheck(comm, safe)
        If (.not. safe) Then
          Call error(691)
        Else
          Allocate (ttm%gtable(1:ttm%gel, 2), Stat=fail) ! [GK] array length corrected
          If (fail > 0) Then
            Write (message, '(a)') 'ttm_table_scan allocation failure'
            Call error(0, message)
          End If
          ttm%gtable(:, :) = 0.0_wp
        End If

      End If

      Deallocate (buffer, Stat=fail)
      If (fail > 0) Then
        Write (message, '(a)') 'ttm_table_scan deallocation failure'
        Call error(0, message)
      End If

    End If

    Return

    ! end of Ke.dat file error exit

    100 Continue

    If (comm%idnode == 0) Close (Unit=ntable)
    Call error(684)

    ! end of Ce.dat file error exit

    200 Continue

    If (comm%idnode == 0) Close (Unit=ntable)
    Call error(686)

    ! end of g.dat file error exit

    300 Continue

    If (comm%idnode == 0) Close (Unit=ntable)
    Call error(688)

    400 Continue

    If (comm%idnode == 0) Close (Unit=ntable)
    Call error(690)

  End Subroutine ttm_table_scan

! fills halo regions of electronic temperature lattice from neighbouring sections
! (periodic boundary conditions)
  Subroutine boundaryHalo(ttm, domain, comm)

    Type(ttm_type),     Intent(InOut) :: ttm
    Type(domains_type), Intent(In   ) :: domain
    Type(comms_type),   Intent(In   ) :: comm

    Integer               :: i, ierr, ii, iii1, iii2, ijk1, ijk2, j, jj, jjj1, jjj2, k, kk, kkk1, &
                             kkk2
    Integer, Allocatable  :: stats(:, :)
    Integer, Dimension(4) :: req

    Allocate (stats(1:MPI_STATUS_SIZE, 1:4))
    If (comm%mxnode > 1) Then

      Do kk = -ttm%eltcell(3), ttm%eltcell(3)
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            If (domain%idx == domain%nx - 1) Then
              iii1 = Mod(ii + 3 * ttm%eltcell(1), (ttm%eltcell(1) * 2 + 1)) - ttm%eltcell(1)
            Else
              iii1 = ii
            End If
            If (domain%idx == 0) Then
              iii2 = Mod(ii + ttm%eltcell(1) + 1, (ttm%eltcell(1) * 2 + 1)) - ttm%eltcell(1)
            Else
              iii2 = ii
            End If
            ijk1 = 2 + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
            ijk2 = 1 + (ttm%ntcell(1) + 1) + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
            Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgx, domain%map(1), Grid1_tag, MPI_COMM_WORLD, req(1), ierr)
            Call MPI_IRECV(ttm%eltemp(ijk2, iii1, jj, kk), 1, ttm%tmpmsgx, domain%map(2), Grid1_tag, MPI_COMM_WORLD, req(2), ierr)
            ijk1 = 1 + (ttm%ntcell(1)) + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
            ijk2 = 1 + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
            Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgx, domain%map(2), Grid2_tag, MPI_COMM_WORLD, req(3), ierr)
            Call MPI_IRECV(ttm%eltemp(ijk2, iii2, jj, kk), 1, ttm%tmpmsgx, domain%map(1), Grid2_tag, MPI_COMM_WORLD, req(4), ierr)
            Call MPI_WAITALL(4, req, stats, ierr)
          End Do
        End Do
      End Do

      Do kk = -ttm%eltcell(3), ttm%eltcell(3)
        Do ii = -ttm%eltcell(1), ttm%eltcell(1)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            If (domain%idy == domain%ny - 1) Then
              jjj1 = Mod(jj + 3 * ttm%eltcell(2), (ttm%eltcell(2) * 2 + 1)) - ttm%eltcell(2)
            Else
              jjj1 = jj
            End If
            If (domain%idy == 0) Then
              jjj2 = Mod(jj + ttm%eltcell(2) + 1, (ttm%eltcell(2) * 2 + 1)) - ttm%eltcell(2)
            Else
              jjj2 = jj
            End If
            ijk1 = 1 + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
            ijk2 = 1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 1 + (ttm%ntcell(2) + 2))
            Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgy, domain%map(3), Grid1_tag, MPI_COMM_WORLD, req(1), ierr)
            Call MPI_IRECV(ttm%eltemp(ijk2, ii, jjj1, kk), 1, ttm%tmpmsgy, domain%map(4), Grid1_tag, MPI_COMM_WORLD, req(2), ierr)
            ijk1 = 1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + (ttm%ntcell(2) + 2))
            ijk2 = 1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
            Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgy, domain%map(4), Grid2_tag, MPI_COMM_WORLD, req(3), ierr)
            Call MPI_IRECV(ttm%eltemp(ijk2, ii, jjj2, kk), 1, ttm%tmpmsgy, domain%map(3), Grid2_tag, MPI_COMM_WORLD, req(4), ierr)
            Call MPI_WAITALL(4, req, stats, ierr)
          End Do
        End Do
      End Do

      Do jj = -ttm%eltcell(2), ttm%eltcell(2)
        Do ii = -ttm%eltcell(1), ttm%eltcell(1)
          Do kk = -ttm%eltcell(3), ttm%eltcell(3)
            If (domain%idz == domain%nz - 1) Then
              kkk1 = Mod(kk + 3 * ttm%eltcell(3), (ttm%eltcell(3) * 2 + 1)) - ttm%eltcell(3)
            Else
              kkk1 = kk
            End If
            If (domain%idz == 0) Then
              kkk2 = Mod(kk + ttm%eltcell(3) + 1, (ttm%eltcell(3) * 2 + 1)) - ttm%eltcell(3)
            Else
              kkk2 = kk
            End If
            ijk1 = 1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
            ijk2 = 1 + (ttm%ntcell(1) + 2) * ((ttm%ntcell(2) + 2) * (ttm%ntcell(3) + 1))
            Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgz, domain%map(5), Grid1_tag, MPI_COMM_WORLD, req(1), ierr)
            Call MPI_IRECV(ttm%eltemp(ijk2, ii, jj, kkk1), 1, ttm%tmpmsgz, domain%map(6), Grid1_tag, MPI_COMM_WORLD, req(2), ierr)
            ijk1 = 1 + (ttm%ntcell(1) + 2) * ((ttm%ntcell(2) + 2) * ttm%ntcell(3))
            ijk2 = 1
            Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgz, domain%map(6), Grid2_tag, MPI_COMM_WORLD, req(3), ierr)
            Call MPI_IRECV(ttm%eltemp(ijk2, ii, jj, kkk2), 1, ttm%tmpmsgz, domain%map(5), Grid2_tag, MPI_COMM_WORLD, req(4), ierr)
            Call MPI_WAITALL(4, req, stats, ierr)
          End Do
        End Do
      End Do

    Else

      Do kk = -ttm%eltcell(3), ttm%eltcell(3)
        kkk1 = Mod(kk + 3 * ttm%eltcell(3), (ttm%eltcell(3) * 2 + 1)) - ttm%eltcell(3)
        kkk2 = Mod(kk + ttm%eltcell(3) + 1, (ttm%eltcell(3) * 2 + 1)) - ttm%eltcell(3)
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          jjj1 = Mod(jj + 3 * ttm%eltcell(2), (ttm%eltcell(2) * 2 + 1)) - ttm%eltcell(2)
          jjj2 = Mod(jj + ttm%eltcell(2) + 1, (ttm%eltcell(2) * 2 + 1)) - ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            iii1 = Mod(ii + 3 * ttm%eltcell(1), (ttm%eltcell(1) * 2 + 1)) - ttm%eltcell(1)
            iii2 = Mod(ii + ttm%eltcell(1) + 1, (ttm%eltcell(1) * 2 + 1)) - ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk1 = 2 + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
                ijk2 = 2 + ttm%ntcell(1) + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
                ttm%eltemp(ijk2, iii1, jj, kk) = ttm%eltemp(ijk1, ii, jj, kk)
                ijk1 = 1 + ttm%ntcell(1) + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
                ijk2 = 1 + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * k)
                ttm%eltemp(ijk2, iii2, jj, kk) = ttm%eltemp(ijk1, ii, jj, kk)
              End Do
            End Do
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2) * k)
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 1 + (ttm%ntcell(2) + 2) * k)
                ttm%eltemp(ijk2, ii, jjj1, kk) = ttm%eltemp(ijk1, ii, jj, kk)
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + (ttm%ntcell(2) + 2) * k)
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2) * k
                ttm%eltemp(ijk2, ii, jjj2, kk) = ttm%eltemp(ijk1, ii, jj, kk)
              End Do
            End Do
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2))
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * (ttm%ntcell(3) + 1))
                ttm%eltemp(ijk2, ii, jj, kkk1) = ttm%eltemp(ijk1, ii, jj, kk)
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ntcell(2) + 2) * ttm%ntcell(3))
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * j
                ttm%eltemp(ijk2, ii, jj, kkk2) = ttm%eltemp(ijk1, ii, jj, kk)
              End Do
            End Do
          End Do
        End Do
      End Do

    End If

    Deallocate (stats)

  End Subroutine boundaryHalo

  Subroutine boundaryCond(key, temp, ttm, comm)

    ! appends halo regions of entire electronic temperature lattice with appropriate boundary conditions

    Integer,          Intent(In   ) :: key
    Real(Kind=wp),    Intent(In   ) :: temp
    Type(ttm_type),   Intent(InOut) :: ttm
    Type(comms_type), Intent(In   ) :: comm

    Integer                                :: i, ierr, ii, ijk1, ijk2, j, jj, k, kk
    Integer, Dimension(4)                  :: req
    Integer, Dimension(MPI_STATUS_SIZE, 4) :: stat

    Select Case (key)
      ! Periodic boundary conditions
    Case (TTM_BC_PERIODIC)
      If (ttm%ttmbcmap(1) >= 0 .or. ttm%ttmbcmap(2) >= 0) Then
        If (comm%mxnode > 1) Then
          Do kk = -ttm%eltcell(3), ttm%eltcell(3)
            Do jj = -ttm%eltcell(2), ttm%eltcell(2)
              If (ttm%ttmbcmap(1) >= 0) Then
                ijk1 = 1 + ttm%ttmbc(1) + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
                ijk2 = 1 + (ttm%ttmbc(1) - 1) + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
                ii = -ttm%eltcell(1)
              Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgx, ttm%ttmbcmap(1), Grid1_tag, MPI_COMM_WORLD, req(1), ierr)
             Call MPI_IRECV(ttm%eltemp(ijk2, -ii, jj, kk), 1, ttm%tmpmsgx, ttm%ttmbcmap(1), Grid2_tag, MPI_COMM_WORLD, req(2), ierr)
              End If
              If (ttm%ttmbcmap(2) >= 0) Then
                ijk1 = 1 + ttm%ttmbc(2) + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
                ijk2 = 1 + (ttm%ttmbc(2) + 1) + (ttm%ntcell(1) + 2) * (1 + (ttm%ntcell(2) + 2))
                ii = ttm%eltcell(1)
              Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgx, ttm%ttmbcmap(2), Grid2_tag, MPI_COMM_WORLD, req(3), ierr)
             Call MPI_IRECV(ttm%eltemp(ijk2, -ii, jj, kk), 1, ttm%tmpmsgx, ttm%ttmbcmap(2), Grid1_tag, MPI_COMM_WORLD, req(4), ierr)
              End If
              Call MPI_WAITALL(4, req, stat, ierr)
            End Do
          End Do
        Else
          Do kk = -ttm%eltcell(3), ttm%eltcell(3)
            Do jj = -ttm%eltcell(2), ttm%eltcell(2)
              Do k = 1, ttm%ntcell(3)
                Do j = 1, ttm%ntcell(2)
                  ijk1 = 1 + ttm%ttmbc(1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                  ijk2 = 1 + (ttm%ttmbc(2) + 1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                  ttm%eltemp(ijk2, ttm%eltcell(1), jj, kk) = ttm%eltemp(ijk1, -ttm%eltcell(1), jj, kk)
                  ijk1 = 1 + ttm%ttmbc(2) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                  ijk2 = 1 + (ttm%ttmbc(1) - 1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                  ttm%eltemp(ijk2, -ttm%eltcell(1), jj, kk) = ttm%eltemp(ijk1, ttm%eltcell(1), jj, kk)
                End Do
              End Do
            End Do
          End Do
        End If
      End If

      If (ttm%ttmbcmap(3) >= 0 .or. ttm%ttmbcmap(4) >= 0) Then
        If (comm%mxnode > 1) Then
          Do kk = -ttm%eltcell(3), ttm%eltcell(3)
            Do ii = -ttm%eltcell(1), ttm%eltcell(1)
              If (ttm%ttmbcmap(3) >= 0) Then
                ijk1 = 1 + (ttm%ntcell(1) + 2) * (ttm%ttmbc(3) + (ttm%ntcell(2) + 2))
                ijk2 = 1 + (ttm%ntcell(1) + 2) * (ttm%ttmbc(3) - 1 + (ttm%ntcell(2) + 2))
                jj = -ttm%eltcell(2)
              Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgy, ttm%ttmbcmap(3), Grid1_tag, MPI_COMM_WORLD, req(1), ierr)
             Call MPI_IRECV(ttm%eltemp(ijk2, ii, -jj, kk), 1, ttm%tmpmsgy, ttm%ttmbcmap(3), Grid2_tag, MPI_COMM_WORLD, req(2), ierr)
              End If
              If (ttm%ttmbcmap(4) >= 0) Then
                ijk1 = 1 + (ttm%ntcell(1) + 2) * (ttm%ttmbc(4) + (ttm%ntcell(2) + 2))
                ijk2 = 1 + (ttm%ntcell(1) + 2) * (ttm%ttmbc(4) + 1 + (ttm%ntcell(2) + 2))
                jj = ttm%eltcell(2)
              Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgy, ttm%ttmbcmap(4), Grid2_tag, MPI_COMM_WORLD, req(3), ierr)
             Call MPI_IRECV(ttm%eltemp(ijk2, ii, -jj, kk), 1, ttm%tmpmsgy, ttm%ttmbcmap(4), Grid1_tag, MPI_COMM_WORLD, req(4), ierr)
              End If
              Call MPI_WAITALL(4, req, stat, ierr)
            End Do
          End Do
        Else
          Do kk = -ttm%eltcell(3), ttm%eltcell(3)
            Do ii = -ttm%eltcell(1), ttm%eltcell(1)
              Do k = 1, ttm%ntcell(3)
                Do i = 0, ttm%ntcell(1) + 1
                  ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(3) + k * (ttm%ntcell(2) + 2))
                  ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(4) + 1 + k * (ttm%ntcell(2) + 2))
                  ttm%eltemp(ijk2, ii, ttm%eltcell(2), kk) = ttm%eltemp(ijk1, ii, -ttm%eltcell(2), kk)
                  ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(4) + k * (ttm%ntcell(2) + 2))
                  ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(3) - 1 + k * (ttm%ntcell(2) + 2))
                  ttm%eltemp(ijk2, ii, -ttm%eltcell(2), kk) = ttm%eltemp(ijk1, ii, ttm%eltcell(2), kk)
                End Do
              End Do
            End Do
          End Do
        End If
      End If

      If (ttm%ttmbcmap(5) >= 0 .or. ttm%ttmbcmap(6) >= 0) Then
        If (comm%mxnode > 1) Then
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do ii = -ttm%eltcell(1), ttm%eltcell(1)
              If (ttm%ttmbcmap(5) >= 0) Then
                ijk1 = 1 + (ttm%ntcell(1) + 2) * (ttm%ttmbc(5) * (ttm%ntcell(2) + 2))
                ijk2 = 1 + (ttm%ntcell(1) + 2) * ((ttm%ttmbc(5) - 1) * (ttm%ntcell(2) + 2))
                kk = -ttm%eltcell(3)
              Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgz, ttm%ttmbcmap(5), Grid1_tag, MPI_COMM_WORLD, req(1), ierr)
             Call MPI_IRECV(ttm%eltemp(ijk2, ii, jj, -kk), 1, ttm%tmpmsgz, ttm%ttmbcmap(5), Grid2_tag, MPI_COMM_WORLD, req(2), ierr)
              End If
              If (ttm%ttmbcmap(6) >= 0) Then
                ijk1 = 1 + (ttm%ntcell(1) + 2) * (ttm%ttmbc(6) * (ttm%ntcell(2) + 2))
                ijk2 = 1 + (ttm%ntcell(1) + 2) * ((ttm%ttmbc(6) + 1) * (ttm%ntcell(2) + 2))
                kk = ttm%eltcell(3)
              Call MPI_ISEND(ttm%eltemp(ijk1, ii, jj, kk), 1, ttm%tmpmsgz, ttm%ttmbcmap(6), Grid2_tag, MPI_COMM_WORLD, req(3), ierr)
             Call MPI_IRECV(ttm%eltemp(ijk2, ii, jj, -kk), 1, ttm%tmpmsgz, ttm%ttmbcmap(6), Grid1_tag, MPI_COMM_WORLD, req(4), ierr)
              End If
              Call MPI_WAITALL(4, req, stat, ierr)
            End Do
          End Do
        Else
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do ii = -ttm%eltcell(1), ttm%eltcell(1)
              Do j = 0, ttm%ntcell(2) + 1
                Do i = 0, ttm%ntcell(1) + 1
                  ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(5) * (ttm%ntcell(2) + 2))
                  ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ttmbc(6) + 1) * (ttm%ntcell(2) + 2))
                  ttm%eltemp(ijk2, ii, jj, ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, -ttm%eltcell(3))
                  ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(6) * (ttm%ntcell(2) + 2))
                  ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ttmbc(5) - 1) * (ttm%ntcell(2) + 2))
                  ttm%eltemp(ijk2, ii, jj, -ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, ttm%eltcell(3))
                End Do
              End Do
            End Do
          End Do
        End If

      End If

      ! Infinite sink/source (Dirichlet) boundary conditions
    Case (TTM_BC_DIRICHLET)
      If (ttm%ttmbcmap(1) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk2 = 1 + (ttm%ttmbc(1) - 1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, -ttm%eltcell(1), jj, kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(2) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk2 = 1 + (ttm%ttmbc(2) + 1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ttm%eltcell(1), jj, kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(3) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * ((ttm%ttmbc(3) - 1) + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ii, -ttm%eltcell(2), kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(4) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * ((ttm%ttmbc(4) + 1) + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ii, ttm%eltcell(2), kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(5) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ttmbc(5) - 1) * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ii, jj, -ttm%eltcell(3)) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(6) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * (j + (ttm%ttmbc(6) + 1) * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ii, jj, ttm%eltcell(3)) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      ! 'Confined' (von Neumann) boundary conditions
    Case (TTM_BC_NEUMANN)
      If (ttm%ttmbcmap(1) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk1 = 1 + ttm%ttmbc(1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - 1
                ttm%eltemp(ijk2, -ttm%eltcell(1), jj, kk) = ttm%eltemp(ijk1, -ttm%eltcell(1), jj, kk)
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(2) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk1 = 1 + ttm%ttmbc(2) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + 1
                ttm%eltemp(ijk2, ttm%eltcell(1), jj, kk) = ttm%eltemp(ijk1, ttm%eltcell(1), jj, kk)
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(3) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(3) + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - (ttm%ntcell(1) + 2)
                ttm%eltemp(ijk2, ii, -ttm%eltcell(2), kk) = ttm%eltemp(ijk1, ii, -ttm%eltcell(2), kk)
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(4) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(4) + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + (ttm%ntcell(1) + 2)
                ttm%eltemp(ijk2, ii, ttm%eltcell(2), kk) = ttm%eltemp(ijk1, ii, ttm%eltcell(2), kk)
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(5) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(5) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, -ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, -ttm%eltcell(3))
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(6) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(6) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, ttm%eltcell(3))
              End Do
            End Do
          End Do
        End Do
      End If

      ! Mixed case: Infinite sink/source (Dirichlet) boundaries in x/y-directions
      !             'Confined' (von Neumann) boundary in z-direction
    Case (TTM_BC_DIRICHLET_XY)
      If (ttm%ttmbcmap(1) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk2 = 1 + (ttm%ttmbc(1) - 1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, -ttm%eltcell(1), jj, kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(2) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk2 = 1 + (ttm%ttmbc(2) + 1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ttm%eltcell(1), jj, kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(3) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * ((ttm%ttmbc(3) - 1) + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ii, -ttm%eltcell(2), kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(4) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk2 = 1 + i + (ttm%ntcell(1) + 2) * ((ttm%ttmbc(4) + 1) + k * (ttm%ntcell(2) + 2))
                ttm%eltemp(ijk2, ii, ttm%eltcell(2), kk) = temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(5) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(5) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, -ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, -ttm%eltcell(3))
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(6) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(6) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, ttm%eltcell(3))
              End Do
            End Do
          End Do
        End Do
      End If

      ! Robin boundary conditions
    Case (TTM_BC_ROBIN)
      If (ttm%ttmbcmap(1) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk1 = 1 + ttm%ttmbc(1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - 1
                ttm%eltemp(ijk2, -ttm%eltcell(1), jj, kk) = ttm%fluxout * (ttm%eltemp(ijk1, -ttm%eltcell(1), jj, kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(2) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk1 = 1 + ttm%ttmbc(2) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + 1
                ttm%eltemp(ijk2, ttm%eltcell(1), jj, kk) = ttm%fluxout * (ttm%eltemp(ijk1, ttm%eltcell(1), jj, kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(3) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(3) + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - (ttm%ntcell(1) + 2)
                ttm%eltemp(ijk2, ii, -ttm%eltcell(2), kk) = ttm%fluxout * (ttm%eltemp(ijk1, ii, -ttm%eltcell(2), kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(4) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(4) + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + (ttm%ntcell(1) + 2)
                ttm%eltemp(ijk2, ii, ttm%eltcell(2), kk) = ttm%fluxout * (ttm%eltemp(ijk1, ii, ttm%eltcell(2), kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(5) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(5) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, -ttm%eltcell(3)) = ttm%fluxout * (ttm%eltemp(ijk1, ii, jj, -ttm%eltcell(3)) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(6) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(6) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, ttm%eltcell(3)) = ttm%fluxout * (ttm%eltemp(ijk1, ii, jj, ttm%eltcell(3)) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      ! Mixed case: Robin boundaries in x/y-directions
      !             'Confined' (von Neumann) boundary in z-direction
    Case (TTM_BC_ROBIN_XY)
      If (ttm%ttmbcmap(1) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk1 = 1 + ttm%ttmbc(1) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - 1
                ttm%eltemp(ijk2, -ttm%eltcell(1), jj, kk) = ttm%fluxout * (ttm%eltemp(ijk1, -ttm%eltcell(1), jj, kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(2) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do jj = -ttm%eltcell(2), ttm%eltcell(2)
            Do k = 1, ttm%ntcell(3)
              Do j = 1, ttm%ntcell(2)
                ijk1 = 1 + ttm%ttmbc(2) + (ttm%ntcell(1) + 2) * (j + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + 1
                ttm%eltemp(ijk2, ttm%eltcell(1), jj, kk) = ttm%fluxout * (ttm%eltemp(ijk1, ttm%eltcell(1), jj, kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(3) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(3) + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - (ttm%ntcell(1) + 2)
                ttm%eltemp(ijk2, ii, -ttm%eltcell(2), kk) = ttm%fluxout * (ttm%eltemp(ijk1, ii, -ttm%eltcell(2), kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(4) >= 0) Then
        Do kk = -ttm%eltcell(3), ttm%eltcell(3)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do k = 1, ttm%ntcell(3)
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (ttm%ttmbc(4) + k * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + (ttm%ntcell(1) + 2)
                ttm%eltemp(ijk2, ii, ttm%eltcell(2), kk) = ttm%fluxout * (ttm%eltemp(ijk1, ii, ttm%eltcell(2), kk) - temp) + temp
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(5) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(5) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 - (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, -ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, -ttm%eltcell(3))
              End Do
            End Do
          End Do
        End Do
      End If

      If (ttm%ttmbcmap(6) >= 0) Then
        Do jj = -ttm%eltcell(2), ttm%eltcell(2)
          Do ii = -ttm%eltcell(1), ttm%eltcell(1)
            Do j = 0, ttm%ntcell(2) + 1
              Do i = 0, ttm%ntcell(1) + 1
                ijk1 = 1 + i + (ttm%ntcell(1) + 2) * (j + ttm%ttmbc(6) * (ttm%ntcell(2) + 2))
                ijk2 = ijk1 + (ttm%ntcell(1) + 2) * (ttm%ntcell(2) + 2)
                ttm%eltemp(ijk2, ii, jj, ttm%eltcell(3)) = ttm%eltemp(ijk1, ii, jj, ttm%eltcell(3))
              End Do
            End Do
          End Do
        End Do
      End If

    End Select

  End Subroutine boundaryCond

  Subroutine uniformDist(lat_in, ttm)

    ! implement constant (homogeneous) spatial deposition

    Type(ttm_type),                                                                    Intent(InOut) :: ttm
    Real(Kind=wp), Dimension(0:ttm%ntcell(1)+1, 0:ttm%ntcell(2)+1, 0:ttm%ntcell(3)+1), Intent(Inout) :: lat_in

    Real(Kind=wp) :: dEdV

    ! express deposition energy per unit volume (eV/A^3):
    ! note penetration depth will be non-zero if laser is
    ! in use, otherwise use dE/dX value

    If (ttm%pdepth > zero_plus) Then
      dEdV = ttm%fluence / ttm%pdepth
    Else
      dEdV = ttm%dEdX / (Real(ttm%ntsys(1), Kind=wp) * Real(ttm%ntsys(2), Kind=wp) * ttm%delx * ttm%dely)
    End If

    ! homogeneous excitation: each temperature config%cell receives
    ! the same energy

    lat_in(1:ttm%ntcell(1), 1:ttm%ntcell(2), 1:ttm%ntcell(3)) = dEdV * ttm%volume

  End Subroutine uniformDist

  Subroutine uniformDistZexp(lat_in, ttm)

    ! implement constant (homogeneous) spatial deposition
    ! in x and y-directions, exponential decay of fluence
    ! in z-direction (only with laser)

    Type(ttm_type),                                                                    Intent(InOut) :: ttm
    Real(Kind=wp), Dimension(0:ttm%ntcell(1)+1, 0:ttm%ntcell(2)+1, 0:ttm%ntcell(3)+1), Intent(Inout) :: lat_in

    Integer       :: k
    Real(Kind=wp) :: dEdV, dEdVmax, rpdepth, zz

    ! express maximum deposition energy per unit volume (eV/A^3)

    If (ttm%pdepth > zero_plus) Then
      rpdepth = 1.0_wp / ttm%pdepth
    Else
      rpdepth = 0.0_wp
    End If
    dEdVmax = ttm%fluence * rpdepth

    ! loop through z-config%cells: calculate stopping power per
    ! config%cell based on z-position (maximum at z=0, grid centre)
    ! and assign across all x and y points in plane

    Do k = 1, ttm%ntcell(3)
      zz = Abs(Real(k + ttm%ntcelloff(3) - ttm%midI(3), Kind=wp))
      dEdV = dEdVmax * Exp(-zz * ttm%delz * rpdepth)
      lat_in(1:ttm%ntcell(1), 1:ttm%ntcell(2), k) = dEdV * ttm%volume
    End Do

  End Subroutine uniformDistZexp

  Subroutine gaussianTrack(lat_in, ttm, comm)

    ! implement gaussian spatial deposition

    Type(ttm_type),                                                                    Intent(InOut) :: ttm
    Real(Kind=wp), Dimension(0:ttm%ntcell(1)+1, 0:ttm%ntcell(2)+1, 0:ttm%ntcell(3)+1), Intent(Inout) :: lat_in
    Type(comms_type),                                                                  Intent(InOut) :: comm

    Integer       :: i, j, sgmx, sgmy
    Logical       :: cutwarn = .false.
    Real(Kind=wp) :: ii, ii2, iim2, iip2, jj, jj2, jjm2, jjp2, normdEdX, realdEdx, sig2x, sig2y, &
                     sigcellx, sigcelly, sigmamx, sigmamy

    lat_in(:, :, :) = 0.0_wp

    ! converting stopping power to a value per config%cell (in z-direction)

    normdEdX = ttm%dEdX * ttm%delz

    ! find extents of gaussian in x and y directions

    sigcellx = ttm%sig / ttm%delx
    sigcelly = ttm%sig / ttm%dely
    sig2x = 2.0_wp * sigcellx * sigcellx
    sig2y = 2.0_wp * sigcelly * sigcelly
    sigmamx = ttm%sigmax * sigcellx
    sigmamy = ttm%sigmax * sigcelly

    ! if cutoff larger than ionic temperature grid,
    ! warn of deposition errors

    If (sigmamx > Ceiling(ttm%ntsys(1) / 2.0_wp)) Then
      sigmamx = Ceiling(ttm%ntsys(1) / 2.0_wp)
      cutwarn = .true.
    End If
    If (sigmamy > Ceiling(ttm%ntsys(2) / 2.0_wp)) Then
      sigmamy = Ceiling(ttm%ntsys(2) / 2.0_wp)
      cutwarn = .true.
    End If

    If (comm%idnode == 0 .and. cutwarn) Then
      Call warning(535, 0.0_wp, 0.0_wp, 0.0_wp)
    End If

    sgmx = Nint(sigmamx)
    sgmy = Nint(sigmamy)

    ! apply five-point linear stencil for gaussian track:
    ! stencil modified to (hopefully!) deposit correct overall energy

    Do j = 1, ttm%ntcell(2)
      jj = Real(j + ttm%ntcelloff(2) - ttm%midI(2), Kind=wp)
      jj2 = -jj * jj / sig2y
      jjp2 = -(jj + 0.5_wp) * (jj + 0.5_wp) / sig2y
      jjm2 = -(jj - 0.5_wp) * (jj - 0.5_wp) / sig2y
      Do i = 1, ttm%ntcell(1)
        ii = Real(i + ttm%ntcelloff(1) - ttm%midI(1), Kind=wp)
        ii2 = -ii * ii / sig2x
        iip2 = -(ii + 0.5_wp) * (ii + 0.5_wp) / sig2x
        iim2 = -(ii - 0.5_wp) * (ii - 0.5_wp) / sig2x
        If (Abs(ii) <= sgmx .and. Abs(jj) <= sgmy) Then
          lat_in(i, j, 1:ttm%ntcell(3)) = 0.2_wp * normdEdX / (2.0_wp * pi * sigcellx * sigcelly) * &
                                          (Exp(ii2 + jj2) + Exp(iim2 + jj2) + Exp(iim2 + jj2) + Exp(ii2 + jjp2) + Exp(ii2 + jjm2))
        End If
      End Do
    End Do

    ! calculate deposited energy for comparison with specified value
    ! (note that stopping power is in z-direction)

    realdEdx = Sum(lat_in(1:ttm%ntcell(1), 1:ttm%ntcell(2), 1:ttm%ntcell(3)))
    Call gsum(comm, realdEdx)
    realdEdx = realdEdx / (Real(ttm%ntsys(3), Kind=wp) * ttm%delz)

    ! check if lattice sum equals the expected value

    If (comm%idnode == 0 .and. Abs((realdEdx - ttm%dEdX) / ttm%dEdX) > 0.01_wp) Then
      Call warning(540, Abs(realdEdx - ttm%dEdX) / ttm%dEdX * 100_wp, 0.0_wp, 0.0_wp)
    End If
  End Subroutine gaussianTrack

End Module ttm
